﻿Option Strict On
Option Explicit On

' alle D2D1 Built-in Effekte, entsprechende Parameter und Default-Werte werden hier beschrieben:
' https://docs.microsoft.com/en-us/windows/win32/direct2d/built-in-effects
' Die meisten Effekte können ab Windows 8 und/oder dem "Platform Update for Windows 7" genutzt werden!
' Ein paar Effekte stehen erst ab Windows 10 zur Verfügung. Siehe Doku zu den Built-in Effekten!
' Ich habe hier keine Extra-Funktion dafür eingebaut, um zu überprüfen auf welchen Windows dieser Code ausgeführt wird!

Imports System.IO
Imports System.Drawing.Imaging
Imports System.Runtime.InteropServices

Public Class D2D1Effects
    Implements IDisposable

#Region "Const"
    Private Const S_OK As Integer = 0
    Private Const D3D_FEATURE_LEVEL_9_1 As Integer = 37120
    Private Const D3D_FEATURE_LEVEL_10_0 As Integer = 40960
    Private Const WINCODEC_SDK_VERSION2 As Integer = 567

    Private Const GUID_WICPixelFormat32bppPBGRA As String = "6fddc324-4e03-4bfe-b185-3d77768dc910"

    Private Const IID_ID2D1Factory As String = "06152247-6f50-465a-9245-118bfd3b6007"
    Private Const IID_ID2D1DeviceContext2 As String = "394ea6a3-0c34-4321-950b-6ca20f0be6c7"

    ' alle D2D1 Built-in Effekte
    Private Const CLSID_D2D1Tint As String = "36312b17-f7dd-4014-915d-ffca768cf211"
    Private Const CLSID_D2D1Crop As String = "e23f7110-0e9a-4324-af47-6a2c0c46f35b"
    Private Const CLSID_D2D1Tile As String = "b0784138-3b76-4bc5-b13b-0fa2ad02659f"
    Private Const CLSID_D2D1YCbCr As String = "99503cc1-66c7-45c9-a875-8ad8a7914401"
    Private Const CLSID_D2D1Atlas As String = "913e2be4-fdcf-4fe2-a5f0-2454f14ff408"
    Private Const CLSID_D2D1Blend As String = "81c5b77b-13f8-4cdd-ad20-c890547ac65d"
    Private Const CLSID_D2D1Flood As String = "61c23c20-ae69-4d8e-94cf-50078df638f2"
    Private Const CLSID_D2D1Sepia As String = "3a1af410-5f1d-4dbe-84df-915da79b7153"
    Private Const CLSID_D2D1Scale As String = "9daf9369-3846-4d0e-a44e-0c607934a5d7"
    Private Const CLSID_D2D1Shadow As String = "c67ea361-1863-4e69-89db-695d3e9a5b6b"
    Private Const CLSID_D2D1Border As String = "2a2d49c0-4acf-43c7-8c6a-7c4a27874d27"
    Private Const CLSID_D2D1Emboss As String = "b1c5eb2b-0348-43f0-8107-4957cacba2ae"
    Private Const CLSID_D2D1Invert As String = "e0c3784d-cb39-4e84-b6fd-6b72f0810263"
    Private Const CLSID_D2D1Sharpen As String = "c9b887cb-c5ff-4dc5-9779-273dcf417c7d"
    Private Const CLSID_D2D1Opacity As String = "811d79a4-de28-4454-8094-c64685f8bd4c"
    Private Const CLSID_D2D1Exposure As String = "b56c8cfa-f634-41ee-bee0-ffa617106004"
    Private Const CLSID_D2D1Vignette As String = "c00c40be-5e67-4ca3-95b4-f4b02c115135"
    Private Const CLSID_D2D1Contrast As String = "b648a78a-0ed5-4f80-a94a-8e825aca6b77"
    Private Const CLSID_D2D1RgbToHue As String = "23f3e5ec-91e8-4d3d-ad0a-afadc1004aa1"
    Private Const CLSID_D2D1HueToRgb As String = "7b78a6bd-0141-4def-8a52-6356ee0cbdd5"
    Private Const CLSID_D2D1CrossFade As String = "12f575e8-4db1-485f-9a84-03a07dd3829f"
    Private Const CLSID_D2D1AlphaMask As String = "c80ecff0-3fd5-4f05-8328-c5d1724b4f0a"
    Private Const CLSID_D2D1Posterize As String = "2188945e-33a3-4366-b7bc-086bd02d0884"
    Private Const CLSID_D2D1Grayscale As String = "36dde0eb-3725-42e0-836d-52fb20aee644"
    Private Const CLSID_D2D1ChromaKey As String = "74c01f5b-2a0d-408c-88e2-c7a3c7197742"
    Private Const CLSID_D2D1Composite As String = "48fc9f51-f6ac-48f1-8b58-3b28ac46f76d"
    Private Const CLSID_D2D1Histogram As String = "881db7d0-f7ee-4d4d-a6d2-4697acc66ee8"
    Private Const CLSID_D2D1HdrToneMap As String = "7b0b748d-4610-4486-a90c-999d9a2e2b11"
    Private Const CLSID_D2D1Saturation As String = "5cb2d9cf-327d-459f-a0ce-40c0b2086bf7"
    Private Const CLSID_D2D1Straighten As String = "4da47b12-79a3-4fb0-8237-bbc3b2a4de08"
    Private Const CLSID_D2D1Brightness As String = "8cea8d1e-77b0-4986-b3b9-2f0c0eae7887"
    Private Const CLSID_D2D1Morphology As String = "eae6c40d-626a-4c2d-bfcb-391001abe202"
    Private Const CLSID_D2D1Turbulence As String = "cf2bb6ae-889a-4ad7-ba29-a2fd732c9fc9"
    Private Const CLSID_D2D1Premultiply As String = "06eab419-deed-4018-80d2-3e1d471adeb2"
    Private Const CLSID_D2D1HueRotation As String = "0f4458ec-4b32-491b-9e85-bd73f44d3eb6"
    Private Const CLSID_D2D1SpotDiffuse As String = "818a1105-7932-44f4-aa86-08ae7b2f2c93"
    Private Const CLSID_D2D1ColorMatrix As String = "921f03d6-641c-47df-852d-b4bb6153ae11"
    Private Const CLSID_D2D13DTransform As String = "e8467b04-ec61-4b8a-b5de-d4d73debea5a"
    Private Const CLSID_D2D1GaussianBlur As String = "1feb6d69-2fe6-4ac9-8c58-1d7f93e7a6a5"
    Private Const CLSID_D2D1BitmapSource As String = "5fb6c24d-c6dd-4231-9404-50f4d5c3252d"
    Private Const CLSID_D2D1PointDiffuse As String = "b9e303c3-c08c-4f91-8b7b-38656bc48c20"
    Private Const CLSID_D2D1SpotSpecular As String = "edae421e-7654-4a37-9db8-71acc1beb3c1"
    Private Const CLSID_D2D1EdgeDetection As String = "eff583ca-cb07-4aa9-ac5d-2cc44c76460f"
    Private Const CLSID_D2D1TableTransfer As String = "5bf818c3-5e43-48cb-b631-868396d6a1d4"
    Private Const CLSID_D2D1GammaTransfer As String = "409444c4-c419-41a0-b0c1-8cd0c0a18e42"
    Private Const CLSID_D2D1UnPremultiply As String = "fb9ac489-ad8d-41ed-9999-bb6347d110f7"
    Private Const CLSID_D2D1PointSpecular As String = "09c3ca26-3ae2-4f09-9ebc-ed3865d53f22"
    Private Const CLSID_D2D1LookupTable3D As String = "349e0eda-0088-4a79-9ca3-c7e300202020"
    Private Const CLSID_D2D1LinearTransfer As String = "ad47c8fd-63ef-4acc-9b51-67979c036c06"
    Private Const CLSID_D2D1ConvolveMatrix As String = "407f8c08-5533-4331-a341-23cc3877843e"
    Private Const CLSID_D2D1DistantDiffuse As String = "3e7efd62-a32d-46d4-a83c-5278889ac954"
    Private Const CLSID_D2D1DpiCompensation As String = "6c26c5c7-34e0-46fc-9cfd-e5823706e228"
    Private Const CLSID_D2D1DirectionalBlur As String = "174319a6-58e9-49b2-bb63-caf2c811a3db"
    Private Const CLSID_D2D1TemperatureTint As String = "89176087-8af9-4a08-aeb1-895f38db1766"
    Private Const CLSID_D2D1DisplacementMap As String = "edc48364-0417-4111-9450-43845fa9f890"
    Private Const CLSID_D2D1DistantSpecular As String = "428c1ee5-77b8-4450-8ab5-72219c21abda"
    Private Const CLSID_D2D1ColorManagement As String = "1a28524c-fdd6-4aa4-ae8f-837eb8267b37"
    Private Const CLSID_D2D1OpacityMetadata As String = "6c53006a-4450-4199-aa5b-ad1656fece5e"
    Private Const CLSID_D2D1DiscreteTransfer As String = "90866fcd-488e-454b-af06-e5041b66c36c"
    Private Const CLSID_D2D1LuminanceToAlpha As String = "41251ab7-0beb-46f8-9da7-59e93fcce5de"
    Private Const CLSID_D2D1HighlightsShadows As String = "cadc8384-323f-4c7e-a361-2e2b24df6ee4"
    Private Const CLSID_D2D12DAffineTransform As String = "6aa97485-6354-4cfc-908c-e4a74f62c96c"
    Private Const CLSID_D2D1ArithmeticComposite As String = "fc151437-049a-4784-a24a-f1c4daf20987"
    Private Const CLSID_D2D1WhiteLevelAdjustment As String = "44a1cadb-6cdd-4818-8ff4-26c1cfe95bdb"
    Private Const CLSID_D2D13DPerspectiveTransform As String = "c2844d0b-3d86-46e7-85ba-526c9240f3fb"
#End Region

#Region "Enum"
    Private Enum VTable_Interfaces As Integer

        '----==== IWICImagingFactory ====----
        CreateBitmapFromHBITMAP = 21

        '----==== ID2D1Factory ====----
        CreateWicBitmapRenderTarget = 13

        ' ----==== ID2D1DeviceContext ====----
        Clear = 47
        BeginDraw = 48
        EndDraw = 49
        CreateColorContext = 59
        CreateEffect = 63
        GetImageLocalBounds = 70
        DrawImage = 83
        CreateLookupTable3D = 99

        ' ----==== ID2D1Effect ====----
        SetValue = 9
        GetValue = 11
        SetInput = 14
        GetOutput = 18

    End Enum

    Private Enum WICAccessOption As Integer
        GENERIC_READ = &H80000000
        GENERIC_WRITE = &H40000000
    End Enum

    Private Enum WICDecodeOptions As Integer
        WICDecodeMetadataCacheOnDemand = 0
        WICDecodeMetadataCacheOnLoad = 1
    End Enum

    Private Enum WICBitmapCreateCacheOption As Integer
        WICBitmapNoCache = 0
        WICBitmapCacheOnDemand = 1
        WICBitmapCacheOnLoad = 2
    End Enum

    Private Enum WICBitmapLockFlags As Integer
        WICBitmapLockRead = 1
        WICBitmapLockWrite = 2
    End Enum

    Private Enum WICBitmapAlphaChannelOption As Integer
        WICBitmapUseAlpha = 0
        WICBitmapUsePremultipliedAlpha = 1
        WICBitmapIgnoreAlpha = 2
    End Enum

    Private Enum DXGI_FORMAT As Integer
        DXGI_FORMAT_UNKNOWN = 0
        DXGI_FORMAT_R32G32B32A32_TYPELESS = 1
        DXGI_FORMAT_R32G32B32A32_FLOAT = 2
        DXGI_FORMAT_R32G32B32A32_UINT = 3
        DXGI_FORMAT_R32G32B32A32_SINT = 4
        DXGI_FORMAT_R32G32B32_TYPELESS = 5
        DXGI_FORMAT_R32G32B32_FLOAT = 6
        DXGI_FORMAT_R32G32B32_UINT = 7
        DXGI_FORMAT_R32G32B32_SINT = 8
        DXGI_FORMAT_R16G16B16A16_TYPELESS = 9
        DXGI_FORMAT_R16G16B16A16_FLOAT = 10
        DXGI_FORMAT_R16G16B16A16_UNORM = 11
        DXGI_FORMAT_R16G16B16A16_UINT = 12
        DXGI_FORMAT_R16G16B16A16_SNORM = 13
        DXGI_FORMAT_R16G16B16A16_SINT = 14
        DXGI_FORMAT_R32G32_TYPELESS = 15
        DXGI_FORMAT_R32G32_FLOAT = 16
        DXGI_FORMAT_R32G32_UINT = 17
        DXGI_FORMAT_R32G32_SINT = 18
        DXGI_FORMAT_R32G8X24_TYPELESS = 19
        DXGI_FORMAT_D32_FLOAT_S8X24_UINT = 20
        DXGI_FORMAT_R32_FLOAT_X8X24_TYPELESS = 21
        DXGI_FORMAT_X32_TYPELESS_G8X24_UINT = 22
        DXGI_FORMAT_R10G10B10A2_TYPELESS = 23
        DXGI_FORMAT_R10G10B10A2_UNORM = 24
        DXGI_FORMAT_R10G10B10A2_UINT = 25
        DXGI_FORMAT_R11G11B10_FLOAT = 26
        DXGI_FORMAT_R8G8B8A8_TYPELESS = 27
        DXGI_FORMAT_R8G8B8A8_UNORM = 28
        DXGI_FORMAT_R8G8B8A8_UNORM_SRGB = 29
        DXGI_FORMAT_R8G8B8A8_UINT = 30
        DXGI_FORMAT_R8G8B8A8_SNORM = 31
        DXGI_FORMAT_R8G8B8A8_SINT = 32
        DXGI_FORMAT_R16G16_TYPELESS = 33
        DXGI_FORMAT_R16G16_FLOAT = 34
        DXGI_FORMAT_R16G16_UNORM = 35
        DXGI_FORMAT_R16G16_UINT = 36
        DXGI_FORMAT_R16G16_SNORM = 37
        DXGI_FORMAT_R16G16_SINT = 38
        DXGI_FORMAT_R32_TYPELESS = 39
        DXGI_FORMAT_D32_FLOAT = 40
        DXGI_FORMAT_R32_FLOAT = 41
        DXGI_FORMAT_R32_UINT = 42
        DXGI_FORMAT_R32_SINT = 43
        DXGI_FORMAT_R24G8_TYPELESS = 44
        DXGI_FORMAT_D24_UNORM_S8_UINT = 45
        DXGI_FORMAT_R24_UNORM_X8_TYPELESS = 46
        DXGI_FORMAT_X24_TYPELESS_G8_UINT = 47
        DXGI_FORMAT_R8G8_TYPELESS = 48
        DXGI_FORMAT_R8G8_UNORM = 49
        DXGI_FORMAT_R8G8_UINT = 50
        DXGI_FORMAT_R8G8_SNORM = 51
        DXGI_FORMAT_R8G8_SINT = 52
        DXGI_FORMAT_R16_TYPELESS = 53
        DXGI_FORMAT_R16_FLOAT = 54
        DXGI_FORMAT_D16_UNORM = 55
        DXGI_FORMAT_R16_UNORM = 56
        DXGI_FORMAT_R16_UINT = 57
        DXGI_FORMAT_R16_SNORM = 58
        DXGI_FORMAT_R16_SINT = 59
        DXGI_FORMAT_R8_TYPELESS = 60
        DXGI_FORMAT_R8_UNORM = 61
        DXGI_FORMAT_R8_UINT = 62
        DXGI_FORMAT_R8_SNORM = 63
        DXGI_FORMAT_R8_SINT = 64
        DXGI_FORMAT_A8_UNORM = 65
        DXGI_FORMAT_R1_UNORM = 66
        DXGI_FORMAT_R9G9B9E5_SHAREDEXP = 67
        DXGI_FORMAT_R8G8_B8G8_UNORM = 68
        DXGI_FORMAT_G8R8_G8B8_UNORM = 69
        DXGI_FORMAT_BC1_TYPELESS = 70
        DXGI_FORMAT_BC1_UNORM = 71
        DXGI_FORMAT_BC1_UNORM_SRGB = 72
        DXGI_FORMAT_BC2_TYPELESS = 73
        DXGI_FORMAT_BC2_UNORM = 74
        DXGI_FORMAT_BC2_UNORM_SRGB = 75
        DXGI_FORMAT_BC3_TYPELESS = 76
        DXGI_FORMAT_BC3_UNORM = 77
        DXGI_FORMAT_BC3_UNORM_SRGB = 78
        DXGI_FORMAT_BC4_TYPELESS = 79
        DXGI_FORMAT_BC4_UNORM = 80
        DXGI_FORMAT_BC4_SNORM = 81
        DXGI_FORMAT_BC5_TYPELESS = 82
        DXGI_FORMAT_BC5_UNORM = 83
        DXGI_FORMAT_BC5_SNORM = 84
        DXGI_FORMAT_B5G6R5_UNORM = 85
        DXGI_FORMAT_B5G5R5A1_UNORM = 86
        DXGI_FORMAT_B8G8R8A8_UNORM = 87
        DXGI_FORMAT_B8G8R8X8_UNORM = 88
        DXGI_FORMAT_R10G10B10_XR_BIAS_A2_UNORM = 89
        DXGI_FORMAT_B8G8R8A8_TYPELESS = 90
        DXGI_FORMAT_B8G8R8A8_UNORM_SRGB = 91
        DXGI_FORMAT_B8G8R8X8_TYPELESS = 92
        DXGI_FORMAT_B8G8R8X8_UNORM_SRGB = 93
        DXGI_FORMAT_BC6H_TYPELESS = 94
        DXGI_FORMAT_BC6H_UF16 = 95
        DXGI_FORMAT_BC6H_SF16 = 96
        DXGI_FORMAT_BC7_TYPELESS = 97
        DXGI_FORMAT_BC7_UNORM = 98
        DXGI_FORMAT_BC7_UNORM_SRGB = 99
        DXGI_FORMAT_AYUV = 100
        DXGI_FORMAT_Y410 = 101
        DXGI_FORMAT_Y416 = 102
        DXGI_FORMAT_NV12 = 103
        DXGI_FORMAT_P010 = 104
        DXGI_FORMAT_P016 = 105
        DXGI_FORMAT_420_OPAQUE = 106
        DXGI_FORMAT_YUY2 = 107
        DXGI_FORMAT_Y210 = 108
        DXGI_FORMAT_Y216 = 109
        DXGI_FORMAT_NV11 = 110
        DXGI_FORMAT_AI44 = 111
        DXGI_FORMAT_IA44 = 112
        DXGI_FORMAT_P8 = 113
        DXGI_FORMAT_A8P8 = 114
        DXGI_FORMAT_B4G4R4A4_UNORM = 115
        DXGI_FORMAT_P208 = 130
        DXGI_FORMAT_V208 = 131
        DXGI_FORMAT_V408 = 132
    End Enum

    Private Enum D2D1_FACTORY_TYPE As Integer
        D2D1_FACTORY_TYPE_SINGLE_THREADED = 0
        D2D1_FACTORY_TYPE_MULTI_THREADED = 1
    End Enum

    Private Enum D2D1_DEBUG_LEVEL As Integer
        D2D1_DEBUG_LEVEL_NONE = 0
        D2D1_DEBUG_LEVEL_ERROR = 1
        D2D1_DEBUG_LEVEL_WARNING = 2
        D2D1_DEBUG_LEVEL_INFORMATION = 3
    End Enum

    Private Enum D2D1_RENDER_TARGET_TYPE As Integer
        D2D1_RENDER_TARGET_TYPE_DEFAULT = 0
        D2D1_RENDER_TARGET_TYPE_SOFTWARE = 1
        D2D1_RENDER_TARGET_TYPE_HARDWARE = 2
    End Enum
    Private Enum D2D1_RENDER_TARGET_USAGE As Integer
        D2D1_RENDER_TARGET_USAGE_NONE = 0
        D2D1_RENDER_TARGET_USAGE_FORCE_BITMAP_REMOTING = 1
        D2D1_RENDER_TARGET_USAGE_GDI_COMPATIBLE = 2
    End Enum

    Private Enum D2D1_FEATURE_LEVEL As Integer
        D2D1_FEATURE_LEVEL_DEFAULT = 0
        D2D1_FEATURE_LEVEL_9 = D3D_FEATURE_LEVEL_9_1
        D2D1_FEATURE_LEVEL_10 = D3D_FEATURE_LEVEL_10_0
    End Enum

    Public Enum D2D1_COLOR_SPACE As Integer
        'D2D1_COLOR_SPACE_CUSTOM = 0
        D2D1_COLOR_SPACE_SRGB = 1
        D2D1_COLOR_SPACE_SCRGB = 2
    End Enum

    Private Enum D2D1_BUFFER_PRECISION As Integer
        D2D1_BUFFER_PRECISION_UNKNOWN = 0
        D2D1_BUFFER_PRECISION_8BPC_UNORM = 1
        D2D1_BUFFER_PRECISION_8BPC_UNORM_SRGB = 2
        D2D1_BUFFER_PRECISION_16BPC_UNORM = 3
        D2D1_BUFFER_PRECISION_16BPC_FLOAT = 4
        D2D1_BUFFER_PRECISION_32BPC_FLOAT = 5
    End Enum

    Public Enum D2D1_ALPHA_MODE As Integer
        D2D1_ALPHA_MODE_UNKNOWN = 0
        D2D1_ALPHA_MODE_PREMULTIPLIED = 1
        D2D1_ALPHA_MODE_STRAIGHT = 2
        D2D1_ALPHA_MODE_IGNORE = 3
    End Enum

    Private Enum D2D1_PROPERTY_TYPE As Integer
        D2D1_PROPERTY_TYPE_UNKNOWN = 0
        D2D1_PROPERTY_TYPE_STRING = 1
        D2D1_PROPERTY_TYPE_BOOL = 2
        D2D1_PROPERTY_TYPE_UINT32 = 3
        D2D1_PROPERTY_TYPE_INT32 = 4
        D2D1_PROPERTY_TYPE_FLOAT = 5
        D2D1_PROPERTY_TYPE_VECTOR2 = 6
        D2D1_PROPERTY_TYPE_VECTOR3 = 7
        D2D1_PROPERTY_TYPE_VECTOR4 = 8
        D2D1_PROPERTY_TYPE_BLOB = 9
        D2D1_PROPERTY_TYPE_IUNKNOWN = 10
        D2D1_PROPERTY_TYPE_ENUM = 11
        D2D1_PROPERTY_TYPE_ARRAY = 12
        D2D1_PROPERTY_TYPE_CLSID = 13
        D2D1_PROPERTY_TYPE_MATRIX_3X2 = 14
        D2D1_PROPERTY_TYPE_MATRIX_4X3 = 15
        D2D1_PROPERTY_TYPE_MATRIX_4X4 = 16
        D2D1_PROPERTY_TYPE_MATRIX_5X4 = 17
        D2D1_PROPERTY_TYPE_COLOR_CONTEXT = 18
    End Enum

    Private Enum D2D1_INTERPOLATION_MODE As Integer
        D2D1_INTERPOLATION_MODE_NEAREST_NEIGHBOR = 0
        D2D1_INTERPOLATION_MODE_LINEAR = 1
        D2D1_INTERPOLATION_MODE_CUBIC = 2
        D2D1_INTERPOLATION_MODE_MULTI_SAMPLE_LINEAR = 3
        D2D1_INTERPOLATION_MODE_ANISOTROPIC = 4
        D2D1_INTERPOLATION_MODE_HIGH_QUALITY_CUBIC = 5
    End Enum

    Public Enum D2D1_COMPOSITE_MODE As Integer
        D2D1_COMPOSITE_MODE_SOURCE_OVER = 0
        D2D1_COMPOSITE_MODE_DESTINATION_OVER = 1
        D2D1_COMPOSITE_MODE_SOURCE_IN = 2
        D2D1_COMPOSITE_MODE_DESTINATION_IN = 3
        D2D1_COMPOSITE_MODE_SOURCE_OUT = 4
        D2D1_COMPOSITE_MODE_DESTINATION_OUT = 5
        D2D1_COMPOSITE_MODE_SOURCE_ATOP = 6
        D2D1_COMPOSITE_MODE_DESTINATION_ATOP = 7
        D2D1_COMPOSITE_MODE_XOR = 8
        D2D1_COMPOSITE_MODE_PLUS = 9
        D2D1_COMPOSITE_MODE_SOURCE_COPY = 10
        D2D1_COMPOSITE_MODE_BOUNDED_SOURCE_COPY = 11
        D2D1_COMPOSITE_MODE_MASK_INVERT = 12
    End Enum

    Public Enum D2D1_BORDER_MODE As Integer
        D2D1_BORDER_MODE_SOFT = 0
        D2D1_BORDER_MODE_HARD = 1
    End Enum

    Public Enum D2D1_CHANNEL_SELECTOR As Integer
        D2D1_CHANNEL_SELECTOR_R = 0
        D2D1_CHANNEL_SELECTOR_G = 1
        D2D1_CHANNEL_SELECTOR_B = 2
        D2D1_CHANNEL_SELECTOR_A = 3
    End Enum

    Public Enum D2D1_BITMAPSOURCE_ORIENTATION As Integer
        D2D1_BITMAPSOURCE_ORIENTATION_DEFAULT = 1
        D2D1_BITMAPSOURCE_ORIENTATION_FLIP_HORIZONTAL = 2
        D2D1_BITMAPSOURCE_ORIENTATION_ROTATE_CLOCKWISE180 = 3
        D2D1_BITMAPSOURCE_ORIENTATION_ROTATE_CLOCKWISE180_FLIP_HORIZONTAL = 4
        D2D1_BITMAPSOURCE_ORIENTATION_ROTATE_CLOCKWISE270_FLIP_HORIZONTAL = 5
        D2D1_BITMAPSOURCE_ORIENTATION_ROTATE_CLOCKWISE90 = 6
        D2D1_BITMAPSOURCE_ORIENTATION_ROTATE_CLOCKWISE90_FLIP_HORIZONTAL = 7
        D2D1_BITMAPSOURCE_ORIENTATION_ROTATE_CLOCKWISE270 = 8
    End Enum

    Private Enum D2D1_GAUSSIANBLUR_PROP As Integer
        D2D1_GAUSSIANBLUR_PROP_STANDARD_DEVIATION = 0
        D2D1_GAUSSIANBLUR_PROP_OPTIMIZATION = 1
        D2D1_GAUSSIANBLUR_PROP_BORDER_MODE = 2
    End Enum

    Public Enum D2D1_GAUSSIANBLUR_OPTIMIZATION As Integer
        D2D1_GAUSSIANBLUR_OPTIMIZATION_SPEED = 0
        D2D1_GAUSSIANBLUR_OPTIMIZATION_BALANCED = 1
        D2D1_GAUSSIANBLUR_OPTIMIZATION_QUALITY = 2
    End Enum

    Private Enum D2D1_DIRECTIONALBLUR_PROP As Integer
        D2D1_DIRECTIONALBLUR_PROP_STANDARD_DEVIATION = 0
        D2D1_DIRECTIONALBLUR_PROP_ANGLE = 1
        D2D1_DIRECTIONALBLUR_PROP_OPTIMIZATION = 2
        D2D1_DIRECTIONALBLUR_PROP_BORDER_MODE = 3
    End Enum

    Public Enum D2D1_DIRECTIONALBLUR_OPTIMIZATION As Integer
        D2D1_DIRECTIONALBLUR_OPTIMIZATION_SPEED = 0
        D2D1_DIRECTIONALBLUR_OPTIMIZATION_BALANCED = 1
        D2D1_DIRECTIONALBLUR_OPTIMIZATION_QUALITY = 2
    End Enum

    Private Enum D2D1_HDRTONEMAP_PROP As Integer
        D2D1_HDRTONEMAP_PROP_INPUT_MAX_LUMINANCE = 0
        D2D1_HDRTONEMAP_PROP_OUTPUT_MAX_LUMINANCE = 1
        D2D1_HDRTONEMAP_PROP_DISPLAY_MODE = 2
    End Enum

    Public Enum D2D1_HDRTONEMAP_DISPLAY_MODE As Integer
        D2D1_HDRTONEMAP_DISPLAY_MODE_SDR = 0
        D2D1_HDRTONEMAP_DISPLAY_MODE_HDR = 1
    End Enum

    Private Enum D2D1_SHADOW_PROP As Integer
        D2D1_SHADOW_PROP_BLUR_STANDARD_DEVIATION = 0
        D2D1_SHADOW_PROP_COLOR = 1
        D2D1_SHADOW_PROP_OPTIMIZATION = 2
    End Enum

    Public Enum D2D1_SHADOW_OPTIMIZATION As Integer
        D2D1_SHADOW_OPTIMIZATION_SPEED = 0
        D2D1_SHADOW_OPTIMIZATION_BALANCED = 1
        D2D1_SHADOW_OPTIMIZATION_QUALITY = 2
    End Enum

    Private Enum D2D1_BLEND_PROP As Integer
        D2D1_BLEND_PROP_MODE = 0
    End Enum

    Private Enum D2D1_BLEND_MODE As Integer
        D2D1_BLEND_MODE_MULTIPLY = 0
        D2D1_BLEND_MODE_SCREEN = 1
        D2D1_BLEND_MODE_DARKEN = 2
        D2D1_BLEND_MODE_LIGHTEN = 3
        D2D1_BLEND_MODE_DISSOLVE = 4
        D2D1_BLEND_MODE_COLOR_BURN = 5
        D2D1_BLEND_MODE_LINEAR_BURN = 6
        D2D1_BLEND_MODE_DARKER_COLOR = 7
        D2D1_BLEND_MODE_LIGHTER_COLOR = 8
        D2D1_BLEND_MODE_COLOR_DODGE = 9
        D2D1_BLEND_MODE_LINEAR_DODGE = 10
        D2D1_BLEND_MODE_OVERLAY = 11
        D2D1_BLEND_MODE_SOFT_LIGHT = 12
        D2D1_BLEND_MODE_HARD_LIGHT = 13
        D2D1_BLEND_MODE_VIVID_LIGHT = 14
        D2D1_BLEND_MODE_LINEAR_LIGHT = 15
        D2D1_BLEND_MODE_PIN_LIGHT = 16
        D2D1_BLEND_MODE_HARD_MIX = 17
        D2D1_BLEND_MODE_DIFFERENCE = 18
        D2D1_BLEND_MODE_EXCLUSION = 19
        D2D1_BLEND_MODE_HUE = 20
        D2D1_BLEND_MODE_SATURATION = 21
        D2D1_BLEND_MODE_COLOR = 22
        D2D1_BLEND_MODE_LUMINOSITY = 23
        D2D1_BLEND_MODE_SUBTRACT = 24
        D2D1_BLEND_MODE_DIVISION = 25
    End Enum

    Private Enum D2D1_SATURATION_PROP As Integer
        D2D1_SATURATION_PROP_SATURATION = 0
    End Enum

    Private Enum D2D1_HUEROTATION_PROP As Integer
        D2D1_HUEROTATION_PROP_ANGLE = 0
    End Enum

    Private Enum D2D1_COLORMATRIX_PROP As Integer
        D2D1_COLORMATRIX_PROP_COLOR_MATRIX = 0
        D2D1_COLORMATRIX_PROP_ALPHA_MODE = 1
        D2D1_COLORMATRIX_PROP_CLAMP_OUTPUT = 2
    End Enum

    Public Enum D2D1_COLORMATRIX_ALPHA_MODE As Integer
        D2D1_COLORMATRIX_ALPHA_MODE_PREMULTIPLIED = 1
        D2D1_COLORMATRIX_ALPHA_MODE_STRAIGHT = 2
    End Enum

    Private Enum D2D1_BITMAPSOURCE_PROP As Integer
        D2D1_BITMAPSOURCE_PROP_WIC_BITMAP_SOURCE = 0
        D2D1_BITMAPSOURCE_PROP_SCALE = 1
        D2D1_BITMAPSOURCE_PROP_INTERPOLATION_MODE = 2
        D2D1_BITMAPSOURCE_PROP_ENABLE_DPI_CORRECTION = 3
        D2D1_BITMAPSOURCE_PROP_ALPHA_MODE = 4
        D2D1_BITMAPSOURCE_PROP_ORIENTATION = 5
    End Enum

    Public Enum D2D1_BITMAPSOURCE_INTERPOLATION_MODE As Integer
        D2D1_BITMAPSOURCE_INTERPOLATION_MODE_NEAREST_NEIGHBOR = 0
        D2D1_BITMAPSOURCE_INTERPOLATION_MODE_LINEAR = 1
        D2D1_BITMAPSOURCE_INTERPOLATION_MODE_CUBIC = 2
        D2D1_BITMAPSOURCE_INTERPOLATION_MODE_FANT = 6
        D2D1_BITMAPSOURCE_INTERPOLATION_MODE_MIPMAP_LINEAR = 7
    End Enum

    Public Enum D2D1_BITMAPSOURCE_ALPHA_MODE As Integer
        D2D1_BITMAPSOURCE_ALPHA_MODE_PREMULTIPLIED = 1
        D2D1_BITMAPSOURCE_ALPHA_MODE_STRAIGHT = 2
    End Enum

    Private Enum D2D1_COMPOSITE_PROP As Integer
        D2D1_COMPOSITE_PROP_MODE = 0
    End Enum

    Private Enum D2D1_3DTRANSFORM_PROP As Integer
        D2D1_3DTRANSFORM_PROP_INTERPOLATION_MODE = 0
        D2D1_3DTRANSFORM_PROP_BORDER_MODE = 1
        D2D1_3DTRANSFORM_PROP_TRANSFORM_MATRIX = 2
    End Enum

    Public Enum D2D1_3DTRANSFORM_INTERPOLATION_MODE As Integer
        D2D1_3DTRANSFORM_INTERPOLATION_MODE_NEAREST_NEIGHBOR = 0
        D2D1_3DTRANSFORM_INTERPOLATION_MODE_LINEAR = 1
        D2D1_3DTRANSFORM_INTERPOLATION_MODE_CUBIC = 2
        D2D1_3DTRANSFORM_INTERPOLATION_MODE_MULTI_SAMPLE_LINEAR = 3
        D2D1_3DTRANSFORM_INTERPOLATION_MODE_ANISOTROPIC = 4
    End Enum

    Private Enum D2D1_3DPERSPECTIVETRANSFORM_PROP As Integer
        D2D1_3DPERSPECTIVETRANSFORM_PROP_INTERPOLATION_MODE = 0
        D2D1_3DPERSPECTIVETRANSFORM_PROP_BORDER_MODE = 1
        D2D1_3DPERSPECTIVETRANSFORM_PROP_DEPTH = 2
        D2D1_3DPERSPECTIVETRANSFORM_PROP_PERSPECTIVE_ORIGIN = 3
        D2D1_3DPERSPECTIVETRANSFORM_PROP_LOCAL_OFFSET = 4
        D2D1_3DPERSPECTIVETRANSFORM_PROP_GLOBAL_OFFSET = 5
        D2D1_3DPERSPECTIVETRANSFORM_PROP_ROTATION_ORIGIN = 6
        D2D1_3DPERSPECTIVETRANSFORM_PROP_ROTATION = 7
    End Enum

    Public Enum D2D1_3DPERSPECTIVETRANSFORM_INTERPOLATION_MODE As Integer
        D2D1_3DPERSPECTIVETRANSFORM_INTERPOLATION_MODE_NEAREST_NEIGHBOR = 0
        D2D1_3DPERSPECTIVETRANSFORM_INTERPOLATION_MODE_LINEAR = 1
        D2D1_3DPERSPECTIVETRANSFORM_INTERPOLATION_MODE_CUBIC = 2
        D2D1_3DPERSPECTIVETRANSFORM_INTERPOLATION_MODE_MULTI_SAMPLE_LINEAR = 3
        D2D1_3DPERSPECTIVETRANSFORM_INTERPOLATION_MODE_ANISOTROPIC = 4
    End Enum

    Private Enum D2D1_2DAFFINETRANSFORM_PROP As Integer
        D2D1_2DAFFINETRANSFORM_PROP_INTERPOLATION_MODE = 0
        D2D1_2DAFFINETRANSFORM_PROP_BORDER_MODE = 1
        D2D1_2DAFFINETRANSFORM_PROP_TRANSFORM_MATRIX = 2
        D2D1_2DAFFINETRANSFORM_PROP_SHARPNESS = 3
    End Enum

    Public Enum D2D1_2DAFFINETRANSFORM_INTERPOLATION_MODE As Integer
        D2D1_2DAFFINETRANSFORM_INTERPOLATION_MODE_NEAREST_NEIGHBOR = 0
        D2D1_2DAFFINETRANSFORM_INTERPOLATION_MODE_LINEAR = 1
        D2D1_2DAFFINETRANSFORM_INTERPOLATION_MODE_CUBIC = 2
        D2D1_2DAFFINETRANSFORM_INTERPOLATION_MODE_MULTI_SAMPLE_LINEAR = 3
        D2D1_2DAFFINETRANSFORM_INTERPOLATION_MODE_ANISOTROPIC = 4
        D2D1_2DAFFINETRANSFORM_INTERPOLATION_MODE_HIGH_QUALITY_CUBIC = 5
    End Enum

    Private Enum D2D1_DPICOMPENSATION_PROP As Integer
        D2D1_DPICOMPENSATION_PROP_INTERPOLATION_MODE = 0
        D2D1_DPICOMPENSATION_PROP_BORDER_MODE = 1
        D2D1_DPICOMPENSATION_PROP_INPUT_DPI = 2
    End Enum

    Public Enum D2D1_DPICOMPENSATION_INTERPOLATION_MODE As Integer
        D2D1_DPICOMPENSATION_INTERPOLATION_MODE_NEAREST_NEIGHBOR = 0
        D2D1_DPICOMPENSATION_INTERPOLATION_MODE_LINEAR = 1
        D2D1_DPICOMPENSATION_INTERPOLATION_MODE_CUBIC = 2
        D2D1_DPICOMPENSATION_INTERPOLATION_MODE_MULTI_SAMPLE_LINEAR = 3
        D2D1_DPICOMPENSATION_INTERPOLATION_MODE_ANISOTROPIC = 4
        D2D1_DPICOMPENSATION_INTERPOLATION_MODE_HIGH_QUALITY_CUBIC = 5
    End Enum

    Private Enum D2D1_SCALE_PROP As Integer
        D2D1_SCALE_PROP_SCALE = 0
        D2D1_SCALE_PROP_CENTER_POINT = 1
        D2D1_SCALE_PROP_INTERPOLATION_MODE = 2
        D2D1_SCALE_PROP_BORDER_MODE = 3
        D2D1_SCALE_PROP_SHARPNESS = 4
    End Enum

    Public Enum D2D1_SCALE_INTERPOLATION_MODE As Integer
        D2D1_SCALE_INTERPOLATION_MODE_NEAREST_NEIGHBOR = 0
        D2D1_SCALE_INTERPOLATION_MODE_LINEAR = 1
        D2D1_SCALE_INTERPOLATION_MODE_CUBIC = 2
        D2D1_SCALE_INTERPOLATION_MODE_MULTI_SAMPLE_LINEAR = 3
        D2D1_SCALE_INTERPOLATION_MODE_ANISOTROPIC = 4
        D2D1_SCALE_INTERPOLATION_MODE_HIGH_QUALITY_CUBIC = 5
    End Enum

    Private Enum D2D1_TURBULENCE_PROP As Integer
        D2D1_TURBULENCE_PROP_OFFSET = 0
        D2D1_TURBULENCE_PROP_SIZE = 1
        D2D1_TURBULENCE_PROP_BASE_FREQUENCY = 2
        D2D1_TURBULENCE_PROP_NUM_OCTAVES = 3
        D2D1_TURBULENCE_PROP_SEED = 4
        D2D1_TURBULENCE_PROP_NOISE = 5
        D2D1_TURBULENCE_PROP_STITCHABLE = 6
    End Enum

    Public Enum D2D1_TURBULENCE_NOISE As Integer
        D2D1_TURBULENCE_NOISE_FRACTAL_SUM = 0
        D2D1_TURBULENCE_NOISE_TURBULENCE = 1
    End Enum

    Private Enum D2D1_DISPLACEMENTMAP_PROP As Integer
        D2D1_DISPLACEMENTMAP_PROP_SCALE = 0
        D2D1_DISPLACEMENTMAP_PROP_X_CHANNEL_SELECT = 1
        D2D1_DISPLACEMENTMAP_PROP_Y_CHANNEL_SELECT = 2
    End Enum

    Private Enum D2D1_COLORMANAGEMENT_PROP As Integer
        D2D1_COLORMANAGEMENT_PROP_SOURCE_COLOR_CONTEXT = 0
        D2D1_COLORMANAGEMENT_PROP_SOURCE_RENDERING_INTENT = 1
        D2D1_COLORMANAGEMENT_PROP_DESTINATION_COLOR_CONTEXT = 2
        D2D1_COLORMANAGEMENT_PROP_DESTINATION_RENDERING_INTENT = 3
        D2D1_COLORMANAGEMENT_PROP_ALPHA_MODE = 4
        D2D1_COLORMANAGEMENT_PROP_QUALITY = 5
    End Enum

    Public Enum D2D1_COLORMANAGEMENT_ALPHA_MODE As Integer
        D2D1_COLORMANAGEMENT_ALPHA_MODE_PREMULTIPLIED = 1
        D2D1_COLORMANAGEMENT_ALPHA_MODE_STRAIGHT = 2
    End Enum

    Public Enum D2D1_COLORMANAGEMENT_QUALITY As Integer
        D2D1_COLORMANAGEMENT_QUALITY_PROOF = 0
        D2D1_COLORMANAGEMENT_QUALITY_NORMAL = 1
        D2D1_COLORMANAGEMENT_QUALITY_BEST = 2
    End Enum

    Public Enum D2D1_COLORMANAGEMENT_RENDERING_INTENT As Integer
        D2D1_COLORMANAGEMENT_RENDERING_INTENT_PERCEPTUAL = 0
        D2D1_COLORMANAGEMENT_RENDERING_INTENT_RELATIVE_COLORIMETRIC = 1
        D2D1_COLORMANAGEMENT_RENDERING_INTENT_SATURATION = 2
        D2D1_COLORMANAGEMENT_RENDERING_INTENT_ABSOLUTE_COLORIMETRIC = 3
    End Enum

    Private Enum D2D1_HISTOGRAM_PROP As Integer
        D2D1_HISTOGRAM_PROP_NUM_BINS = 0
        D2D1_HISTOGRAM_PROP_CHANNEL_SELECT = 1
        D2D1_HISTOGRAM_PROP_HISTOGRAM_OUTPUT = 2
    End Enum

    Private Enum D2D1_POINTSPECULAR_PROP As Integer
        D2D1_POINTSPECULAR_PROP_LIGHT_POSITION = 0
        D2D1_POINTSPECULAR_PROP_SPECULAR_EXPONENT = 1
        D2D1_POINTSPECULAR_PROP_SPECULAR_CONSTANT = 2
        D2D1_POINTSPECULAR_PROP_SURFACE_SCALE = 3
        D2D1_POINTSPECULAR_PROP_COLOR = 4
        D2D1_POINTSPECULAR_PROP_KERNEL_UNIT_LENGTH = 5
        D2D1_POINTSPECULAR_PROP_SCALE_MODE = 6
    End Enum

    Public Enum D2D1_POINTSPECULAR_SCALE_MODE As Integer
        D2D1_POINTSPECULAR_SCALE_MODE_NEAREST_NEIGHBOR = 0
        D2D1_POINTSPECULAR_SCALE_MODE_LINEAR = 1
        D2D1_POINTSPECULAR_SCALE_MODE_CUBIC = 2
        D2D1_POINTSPECULAR_SCALE_MODE_MULTI_SAMPLE_LINEAR = 3
        D2D1_POINTSPECULAR_SCALE_MODE_ANISOTROPIC = 4
        D2D1_POINTSPECULAR_SCALE_MODE_HIGH_QUALITY_CUBIC = 5
    End Enum

    Private Enum D2D1_SPOTSPECULAR_PROP As Integer
        D2D1_SPOTSPECULAR_PROP_LIGHT_POSITION = 0
        D2D1_SPOTSPECULAR_PROP_POINTS_AT = 1
        D2D1_SPOTSPECULAR_PROP_FOCUS = 2
        D2D1_SPOTSPECULAR_PROP_LIMITING_CONE_ANGLE = 3
        D2D1_SPOTSPECULAR_PROP_SPECULAR_EXPONENT = 4
        D2D1_SPOTSPECULAR_PROP_SPECULAR_CONSTANT = 5
        D2D1_SPOTSPECULAR_PROP_SURFACE_SCALE = 6
        D2D1_SPOTSPECULAR_PROP_COLOR = 7
        D2D1_SPOTSPECULAR_PROP_KERNEL_UNIT_LENGTH = 8
        D2D1_SPOTSPECULAR_PROP_SCALE_MODE = 9
    End Enum

    Public Enum D2D1_SPOTSPECULAR_SCALE_MODE As Integer
        D2D1_SPOTSPECULAR_SCALE_MODE_NEAREST_NEIGHBOR = 0
        D2D1_SPOTSPECULAR_SCALE_MODE_LINEAR = 1
        D2D1_SPOTSPECULAR_SCALE_MODE_CUBIC = 2
        D2D1_SPOTSPECULAR_SCALE_MODE_MULTI_SAMPLE_LINEAR = 3
        D2D1_SPOTSPECULAR_SCALE_MODE_ANISOTROPIC = 4
        D2D1_SPOTSPECULAR_SCALE_MODE_HIGH_QUALITY_CUBIC = 5
    End Enum

    Private Enum D2D1_DISTANTSPECULAR_PROP As Integer
        D2D1_DISTANTSPECULAR_PROP_AZIMUTH = 0
        D2D1_DISTANTSPECULAR_PROP_ELEVATION = 1
        D2D1_DISTANTSPECULAR_PROP_SPECULAR_EXPONENT = 2
        D2D1_DISTANTSPECULAR_PROP_SPECULAR_CONSTANT = 3
        D2D1_DISTANTSPECULAR_PROP_SURFACE_SCALE = 4
        D2D1_DISTANTSPECULAR_PROP_COLOR = 5
        D2D1_DISTANTSPECULAR_PROP_KERNEL_UNIT_LENGTH = 6
        D2D1_DISTANTSPECULAR_PROP_SCALE_MODE = 7
    End Enum

    Public Enum D2D1_DISTANTSPECULAR_SCALE_MODE As Integer
        D2D1_DISTANTSPECULAR_SCALE_MODE_NEAREST_NEIGHBOR = 0
        D2D1_DISTANTSPECULAR_SCALE_MODE_LINEAR = 1
        D2D1_DISTANTSPECULAR_SCALE_MODE_CUBIC = 2
        D2D1_DISTANTSPECULAR_SCALE_MODE_MULTI_SAMPLE_LINEAR = 3
        D2D1_DISTANTSPECULAR_SCALE_MODE_ANISOTROPIC = 4
        D2D1_DISTANTSPECULAR_SCALE_MODE_HIGH_QUALITY_CUBIC = 5
    End Enum

    Private Enum D2D1_POINTDIFFUSE_PROP As Integer
        D2D1_POINTDIFFUSE_PROP_LIGHT_POSITION = 0
        D2D1_POINTDIFFUSE_PROP_DIFFUSE_CONSTANT = 1
        D2D1_POINTDIFFUSE_PROP_SURFACE_SCALE = 2
        D2D1_POINTDIFFUSE_PROP_COLOR = 3
        D2D1_POINTDIFFUSE_PROP_KERNEL_UNIT_LENGTH = 4
        D2D1_POINTDIFFUSE_PROP_SCALE_MODE = 5
    End Enum

    Public Enum D2D1_POINTDIFFUSE_SCALE_MODE As Integer
        D2D1_POINTDIFFUSE_SCALE_MODE_NEAREST_NEIGHBOR = 0
        D2D1_POINTDIFFUSE_SCALE_MODE_LINEAR = 1
        D2D1_POINTDIFFUSE_SCALE_MODE_CUBIC = 2
        D2D1_POINTDIFFUSE_SCALE_MODE_MULTI_SAMPLE_LINEAR = 3
        D2D1_POINTDIFFUSE_SCALE_MODE_ANISOTROPIC = 4
        D2D1_POINTDIFFUSE_SCALE_MODE_HIGH_QUALITY_CUBIC = 5
    End Enum

    Private Enum D2D1_WHITELEVELADJUSTMENT_PROP As Integer
        D2D1_WHITELEVELADJUSTMENT_PROP_INPUT_WHITE_LEVEL = 0
        D2D1_WHITELEVELADJUSTMENT_PROP_OUTPUT_WHITE_LEVEL = 1
    End Enum

    Private Enum D2D1_SPOTDIFFUSE_PROP As Integer
        D2D1_SPOTDIFFUSE_PROP_LIGHT_POSITION = 0
        D2D1_SPOTDIFFUSE_PROP_POINTS_AT = 1
        D2D1_SPOTDIFFUSE_PROP_FOCUS = 2
        D2D1_SPOTDIFFUSE_PROP_LIMITING_CONE_ANGLE = 3
        D2D1_SPOTDIFFUSE_PROP_DIFFUSE_CONSTANT = 4
        D2D1_SPOTDIFFUSE_PROP_SURFACE_SCALE = 5
        D2D1_SPOTDIFFUSE_PROP_COLOR = 6
        D2D1_SPOTDIFFUSE_PROP_KERNEL_UNIT_LENGTH = 7
        D2D1_SPOTDIFFUSE_PROP_SCALE_MODE = 8
    End Enum

    Public Enum D2D1_SPOTDIFFUSE_SCALE_MODE As Integer
        D2D1_SPOTDIFFUSE_SCALE_MODE_NEAREST_NEIGHBOR = 0
        D2D1_SPOTDIFFUSE_SCALE_MODE_LINEAR = 1
        D2D1_SPOTDIFFUSE_SCALE_MODE_CUBIC = 2
        D2D1_SPOTDIFFUSE_SCALE_MODE_MULTI_SAMPLE_LINEAR = 3
        D2D1_SPOTDIFFUSE_SCALE_MODE_ANISOTROPIC = 4
        D2D1_SPOTDIFFUSE_SCALE_MODE_HIGH_QUALITY_CUBIC = 5
    End Enum

    Private Enum D2D1_DISTANTDIFFUSE_PROP As Integer
        D2D1_DISTANTDIFFUSE_PROP_AZIMUTH = 0
        D2D1_DISTANTDIFFUSE_PROP_ELEVATION = 1
        D2D1_DISTANTDIFFUSE_PROP_DIFFUSE_CONSTANT = 2
        D2D1_DISTANTDIFFUSE_PROP_SURFACE_SCALE = 3
        D2D1_DISTANTDIFFUSE_PROP_COLOR = 4
        D2D1_DISTANTDIFFUSE_PROP_KERNEL_UNIT_LENGTH = 5
        D2D1_DISTANTDIFFUSE_PROP_SCALE_MODE = 6
    End Enum

    Public Enum D2D1_DISTANTDIFFUSE_SCALE_MODE As Integer
        D2D1_DISTANTDIFFUSE_SCALE_MODE_NEAREST_NEIGHBOR = 0
        D2D1_DISTANTDIFFUSE_SCALE_MODE_LINEAR = 1
        D2D1_DISTANTDIFFUSE_SCALE_MODE_CUBIC = 2
        D2D1_DISTANTDIFFUSE_SCALE_MODE_MULTI_SAMPLE_LINEAR = 3
        D2D1_DISTANTDIFFUSE_SCALE_MODE_ANISOTROPIC = 4
        D2D1_DISTANTDIFFUSE_SCALE_MODE_HIGH_QUALITY_CUBIC = 5
    End Enum

    Private Enum D2D1_FLOOD_PROP As Integer
        D2D1_FLOOD_PROP_COLOR = 0
    End Enum

    Private Enum D2D1_LINEARTRANSFER_PROP As Integer
        D2D1_LINEARTRANSFER_PROP_RED_Y_INTERCEPT = 0
        D2D1_LINEARTRANSFER_PROP_RED_SLOPE = 1
        D2D1_LINEARTRANSFER_PROP_RED_DISABLE = 2
        D2D1_LINEARTRANSFER_PROP_GREEN_Y_INTERCEPT = 3
        D2D1_LINEARTRANSFER_PROP_GREEN_SLOPE = 4
        D2D1_LINEARTRANSFER_PROP_GREEN_DISABLE = 5
        D2D1_LINEARTRANSFER_PROP_BLUE_Y_INTERCEPT = 6
        D2D1_LINEARTRANSFER_PROP_BLUE_SLOPE = 7
        D2D1_LINEARTRANSFER_PROP_BLUE_DISABLE = 8
        D2D1_LINEARTRANSFER_PROP_ALPHA_Y_INTERCEPT = 9
        D2D1_LINEARTRANSFER_PROP_ALPHA_SLOPE = 10
        D2D1_LINEARTRANSFER_PROP_ALPHA_DISABLE = 11
        D2D1_LINEARTRANSFER_PROP_CLAMP_OUTPUT = 12
    End Enum

    Private Enum D2D1_GAMMATRANSFER_PROP As Integer
        D2D1_GAMMATRANSFER_PROP_RED_AMPLITUDE = 0
        D2D1_GAMMATRANSFER_PROP_RED_EXPONENT = 1
        D2D1_GAMMATRANSFER_PROP_RED_OFFSET = 2
        D2D1_GAMMATRANSFER_PROP_RED_DISABLE = 3
        D2D1_GAMMATRANSFER_PROP_GREEN_AMPLITUDE = 4
        D2D1_GAMMATRANSFER_PROP_GREEN_EXPONENT = 5
        D2D1_GAMMATRANSFER_PROP_GREEN_OFFSET = 6
        D2D1_GAMMATRANSFER_PROP_GREEN_DISABLE = 7
        D2D1_GAMMATRANSFER_PROP_BLUE_AMPLITUDE = 8
        D2D1_GAMMATRANSFER_PROP_BLUE_EXPONENT = 9
        D2D1_GAMMATRANSFER_PROP_BLUE_OFFSET = 10
        D2D1_GAMMATRANSFER_PROP_BLUE_DISABLE = 11
        D2D1_GAMMATRANSFER_PROP_ALPHA_AMPLITUDE = 12
        D2D1_GAMMATRANSFER_PROP_ALPHA_EXPONENT = 13
        D2D1_GAMMATRANSFER_PROP_ALPHA_OFFSET = 14
        D2D1_GAMMATRANSFER_PROP_ALPHA_DISABLE = 15
        D2D1_GAMMATRANSFER_PROP_CLAMP_OUTPUT = 16
    End Enum

    Private Enum D2D1_TABLETRANSFER_PROP As Integer
        D2D1_TABLETRANSFER_PROP_RED_TABLE = 0
        D2D1_TABLETRANSFER_PROP_RED_DISABLE = 1
        D2D1_TABLETRANSFER_PROP_GREEN_TABLE = 2
        D2D1_TABLETRANSFER_PROP_GREEN_DISABLE = 3
        D2D1_TABLETRANSFER_PROP_BLUE_TABLE = 4
        D2D1_TABLETRANSFER_PROP_BLUE_DISABLE = 5
        D2D1_TABLETRANSFER_PROP_ALPHA_TABLE = 6
        D2D1_TABLETRANSFER_PROP_ALPHA_DISABLE = 7
        D2D1_TABLETRANSFER_PROP_CLAMP_OUTPUT = 8
    End Enum

    Private Enum D2D1_DISCRETETRANSFER_PROP As Integer
        D2D1_DISCRETETRANSFER_PROP_RED_TABLE = 0
        D2D1_DISCRETETRANSFER_PROP_RED_DISABLE = 1
        D2D1_DISCRETETRANSFER_PROP_GREEN_TABLE = 2
        D2D1_DISCRETETRANSFER_PROP_GREEN_DISABLE = 3
        D2D1_DISCRETETRANSFER_PROP_BLUE_TABLE = 4
        D2D1_DISCRETETRANSFER_PROP_BLUE_DISABLE = 5
        D2D1_DISCRETETRANSFER_PROP_ALPHA_TABLE = 6
        D2D1_DISCRETETRANSFER_PROP_ALPHA_DISABLE = 7
        D2D1_DISCRETETRANSFER_PROP_CLAMP_OUTPUT = 8
    End Enum

    Private Enum D2D1_CONVOLVEMATRIX_PROP As Integer
        D2D1_CONVOLVEMATRIX_PROP_KERNEL_UNIT_LENGTH = 0
        D2D1_CONVOLVEMATRIX_PROP_SCALE_MODE = 1
        D2D1_CONVOLVEMATRIX_PROP_KERNEL_SIZE_X = 2
        D2D1_CONVOLVEMATRIX_PROP_KERNEL_SIZE_Y = 3
        D2D1_CONVOLVEMATRIX_PROP_KERNEL_MATRIX = 4
        D2D1_CONVOLVEMATRIX_PROP_DIVISOR = 5
        D2D1_CONVOLVEMATRIX_PROP_BIAS = 6
        D2D1_CONVOLVEMATRIX_PROP_KERNEL_OFFSET = 7
        D2D1_CONVOLVEMATRIX_PROP_PRESERVE_ALPHA = 8
        D2D1_CONVOLVEMATRIX_PROP_BORDER_MODE = 9
        D2D1_CONVOLVEMATRIX_PROP_CLAMP_OUTPUT = 10
    End Enum

    Public Enum D2D1_CONVOLVEMATRIX_SCALE_MODE As Integer
        D2D1_CONVOLVEMATRIX_SCALE_MODE_NEAREST_NEIGHBOR = 0
        D2D1_CONVOLVEMATRIX_SCALE_MODE_LINEAR = 1
        D2D1_CONVOLVEMATRIX_SCALE_MODE_CUBIC = 2
        D2D1_CONVOLVEMATRIX_SCALE_MODE_MULTI_SAMPLE_LINEAR = 3
        D2D1_CONVOLVEMATRIX_SCALE_MODE_ANISOTROPIC = 4
        D2D1_CONVOLVEMATRIX_SCALE_MODE_HIGH_QUALITY_CUBIC = 5
    End Enum

    Private Enum D2D1_BRIGHTNESS_PROP As Integer
        D2D1_BRIGHTNESS_PROP_WHITE_POINT = 0
        D2D1_BRIGHTNESS_PROP_BLACK_POINT = 1
    End Enum

    Private Enum D2D1_ARITHMETICCOMPOSITE_PROP As Integer
        D2D1_ARITHMETICCOMPOSITE_PROP_COEFFICIENTS = 0
        D2D1_ARITHMETICCOMPOSITE_PROP_CLAMP_OUTPUT = 1
    End Enum

    Private Enum D2D1_CROP_PROP As Integer
        D2D1_CROP_PROP_RECT = 0
        D2D1_CROP_PROP_BORDER_MODE = 1
    End Enum

    Private Enum D2D1_BORDER_PROP As Integer
        D2D1_BORDER_PROP_EDGE_MODE_X = 0
        D2D1_BORDER_PROP_EDGE_MODE_Y = 1
    End Enum

    Public Enum D2D1_BORDER_EDGE_MODE As Integer
        D2D1_BORDER_EDGE_MODE_CLAMP = 0
        D2D1_BORDER_EDGE_MODE_WRAP = 1
        D2D1_BORDER_EDGE_MODE_MIRROR = 2
    End Enum

    Private Enum D2D1_MORPHOLOGY_PROP As Integer
        D2D1_MORPHOLOGY_PROP_MODE = 0
        D2D1_MORPHOLOGY_PROP_WIDTH = 1
        D2D1_MORPHOLOGY_PROP_HEIGHT = 2
    End Enum

    Public Enum D2D1_MORPHOLOGY_MODE As Integer
        D2D1_MORPHOLOGY_MODE_ERODE = 0
        D2D1_MORPHOLOGY_MODE_DILATE = 1
    End Enum

    Private Enum D2D1_TILE_PROP As Integer
        D2D1_TILE_PROP_RECT = 0
    End Enum

    Private Enum D2D1_ATLAS_PROP As Integer
        D2D1_ATLAS_PROP_INPUT_RECT = 0
        D2D1_ATLAS_PROP_INPUT_PADDING_RECT = 1
    End Enum

    Private Enum D2D1_OPACITYMETADATA_PROP As Integer
        D2D1_OPACITYMETADATA_PROP_INPUT_OPAQUE_RECT = 0
    End Enum

    Private Enum D2D1_YCBCR_PROP As Integer
        D2D1_YCBCR_PROP_CHROMA_SUBSAMPLING = 0
        D2D1_YCBCR_PROP_TRANSFORM_MATRIX = 1
        D2D1_YCBCR_PROP_INTERPOLATION_MODE = 2
    End Enum

    Public Enum D2D1_YCBCR_CHROMA_SUBSAMPLING As Integer
        D2D1_YCBCR_CHROMA_SUBSAMPLING_AUTO = 0
        D2D1_YCBCR_CHROMA_SUBSAMPLING_420 = 1
        D2D1_YCBCR_CHROMA_SUBSAMPLING_422 = 2
        D2D1_YCBCR_CHROMA_SUBSAMPLING_444 = 3
        D2D1_YCBCR_CHROMA_SUBSAMPLING_440 = 4
    End Enum

    Public Enum D2D1_YCBCR_INTERPOLATION_MODE As Integer
        D2D1_YCBCR_INTERPOLATION_MODE_NEAREST_NEIGHBOR = 0
        D2D1_YCBCR_INTERPOLATION_MODE_LINEAR = 1
        D2D1_YCBCR_INTERPOLATION_MODE_CUBIC = 2
        D2D1_YCBCR_INTERPOLATION_MODE_MULTI_SAMPLE_LINEAR = 3
        D2D1_YCBCR_INTERPOLATION_MODE_ANISOTROPIC = 4
        D2D1_YCBCR_INTERPOLATION_MODE_HIGH_QUALITY_CUBIC = 5
    End Enum

    Private Enum D2D1_CONTRAST_PROP As Integer
        D2D1_CONTRAST_PROP_CONTRAST = 0
        D2D1_CONTRAST_PROP_CLAMP_INPUT = 1
    End Enum

    Private Enum D2D1_RGBTOHUE_PROP As Integer
        D2D1_RGBTOHUE_PROP_OUTPUT_COLOR_SPACE = 0
    End Enum

    Public Enum D2D1_RGBTOHUE_OUTPUT_COLOR_SPACE As Integer
        D2D1_RGBTOHUE_OUTPUT_COLOR_SPACE_HUE_SATURATION_VALUE = 0
        D2D1_RGBTOHUE_OUTPUT_COLOR_SPACE_HUE_SATURATION_LIGHTNESS = 1
    End Enum

    Private Enum D2D1_HUETORGB_PROP As Integer
        D2D1_HUETORGB_PROP_INPUT_COLOR_SPACE = 0
    End Enum

    Public Enum D2D1_HUETORGB_INPUT_COLOR_SPACE As Integer
        D2D1_HUETORGB_INPUT_COLOR_SPACE_HUE_SATURATION_VALUE = 0
        D2D1_HUETORGB_INPUT_COLOR_SPACE_HUE_SATURATION_LIGHTNESS = 1
    End Enum

    Private Enum D2D1_CHROMAKEY_PROP As Integer
        D2D1_CHROMAKEY_PROP_COLOR = 0
        D2D1_CHROMAKEY_PROP_TOLERANCE = 1
        D2D1_CHROMAKEY_PROP_INVERT_ALPHA = 2
        D2D1_CHROMAKEY_PROP_FEATHER = 3
    End Enum

    Private Enum D2D1_EMBOSS_PROP As Integer
        D2D1_EMBOSS_PROP_HEIGHT = 0
        D2D1_EMBOSS_PROP_DIRECTION = 1
    End Enum

    Private Enum D2D1_EXPOSURE_PROP As Integer
        D2D1_EXPOSURE_PROP_EXPOSURE_VALUE = 0
    End Enum

    Private Enum D2D1_POSTERIZE_PROP As Integer
        D2D1_POSTERIZE_PROP_RED_VALUE_COUNT = 0
        D2D1_POSTERIZE_PROP_GREEN_VALUE_COUNT = 1
        D2D1_POSTERIZE_PROP_BLUE_VALUE_COUNT = 2
    End Enum

    Private Enum D2D1_SEPIA_PROP As Integer
        D2D1_SEPIA_PROP_INTENSITY = 0
        D2D1_SEPIA_PROP_ALPHA_MODE = 1
    End Enum

    Private Enum D2D1_SHARPEN_PROP As Integer
        D2D1_SHARPEN_PROP_SHARPNESS = 0
        D2D1_SHARPEN_PROP_THRESHOLD = 1
    End Enum

    Private Enum D2D1_STRAIGHTEN_PROP As Integer
        D2D1_STRAIGHTEN_PROP_ANGLE = 0
        D2D1_STRAIGHTEN_PROP_MAINTAIN_SIZE = 1
        D2D1_STRAIGHTEN_PROP_SCALE_MODE = 2
    End Enum

    Public Enum D2D1_STRAIGHTEN_SCALE_MODE As Integer
        D2D1_STRAIGHTEN_SCALE_MODE_NEAREST_NEIGHBOR = 0
        D2D1_STRAIGHTEN_SCALE_MODE_LINEAR = 1
        D2D1_STRAIGHTEN_SCALE_MODE_CUBIC = 2
        D2D1_STRAIGHTEN_SCALE_MODE_MULTI_SAMPLE_LINEAR = 3
        D2D1_STRAIGHTEN_SCALE_MODE_ANISOTROPIC = 4
    End Enum

    Private Enum D2D1_TEMPERATUREANDTINT_PROP As Integer
        D2D1_TEMPERATUREANDTINT_PROP_TEMPERATURE = 0
        D2D1_TEMPERATUREANDTINT_PROP_TINT = 1
    End Enum

    Private Enum D2D1_VIGNETTE_PROP As Integer
        D2D1_VIGNETTE_PROP_COLOR = 0
        D2D1_VIGNETTE_PROP_TRANSITION_SIZE = 1
        D2D1_VIGNETTE_PROP_STRENGTH = 2
    End Enum

    Private Enum D2D1_EDGEDETECTION_PROP As Integer
        D2D1_EDGEDETECTION_PROP_STRENGTH = 0
        D2D1_EDGEDETECTION_PROP_BLUR_RADIUS = 1
        D2D1_EDGEDETECTION_PROP_MODE = 2
        D2D1_EDGEDETECTION_PROP_OVERLAY_EDGES = 3
        D2D1_EDGEDETECTION_PROP_ALPHA_MODE = 4
    End Enum

    Public Enum D2D1_EDGEDETECTION_MODE As Integer
        D2D1_EDGEDETECTION_MODE_SOBEL = 0
        D2D1_EDGEDETECTION_MODE_PREWITT = 1
    End Enum

    Private Enum D2D1_HIGHLIGHTSANDSHADOWS_PROP As Integer
        D2D1_HIGHLIGHTSANDSHADOWS_PROP_HIGHLIGHTS = 0
        D2D1_HIGHLIGHTSANDSHADOWS_PROP_SHADOWS = 1
        D2D1_HIGHLIGHTSANDSHADOWS_PROP_CLARITY = 2
        D2D1_HIGHLIGHTSANDSHADOWS_PROP_INPUT_GAMMA = 3
        D2D1_HIGHLIGHTSANDSHADOWS_PROP_MASK_BLUR_RADIUS = 4
    End Enum

    Public Enum D2D1_HIGHLIGHTSANDSHADOWS_INPUT_GAMMA As Integer
        D2D1_HIGHLIGHTSANDSHADOWS_INPUT_GAMMA_LINEAR = 0
        D2D1_HIGHLIGHTSANDSHADOWS_INPUT_GAMMA_SRGB = 1
    End Enum

    Private Enum D2D1_LOOKUPTABLE3D_PROP As Integer
        D2D1_LOOKUPTABLE3D_PROP_LUT = 0
        D2D1_LOOKUPTABLE3D_PROP_ALPHA_MODE = 1
    End Enum

    Private Enum D2D1_OPACITY_PROP As Integer
        D2D1_OPACITY_PROP_OPACITY = 0
    End Enum

    Private Enum D2D1_CROSSFADE_PROP As Integer
        D2D1_CROSSFADE_PROP_WEIGHT = 0
    End Enum

    Private Enum D2D1_TINT_PROP As Integer
        D2D1_TINT_PROP_COLOR = 0
        D2D1_TINT_PROP_CLAMP_OUTPUT = 1
    End Enum

#End Region

#Region "Structure"
    Private Structure WICRect
        Dim x As Integer
        Dim y As Integer
        Dim width As Integer
        Dim height As Integer
        Sub New(intX As Integer,
                intY As Integer,
                intWidth As Integer,
                intHeight As Integer)
            x = intX
            y = intY
            width = intWidth
            height = intHeight
        End Sub
    End Structure

    Private Structure D2D1_FACTORY_OPTIONS
        Dim eDebugLevel As D2D1_DEBUG_LEVEL
        Sub New(Optional DebugLevel As D2D1_DEBUG_LEVEL = D2D1_DEBUG_LEVEL.D2D1_DEBUG_LEVEL_NONE)
            eDebugLevel = DebugLevel
        End Sub
    End Structure

    Private Structure D2D1_PIXEL_FORMAT
        Dim eDxgiFormat As DXGI_FORMAT
        Dim eAlphaMode As D2D1_ALPHA_MODE
        Sub New(Optional AlphaMode As D2D1_ALPHA_MODE = D2D1_ALPHA_MODE.D2D1_ALPHA_MODE_PREMULTIPLIED,
                Optional DXGIFormat As DXGI_FORMAT = DXGI_FORMAT.DXGI_FORMAT_B8G8R8A8_UNORM)
            eAlphaMode = AlphaMode
            eDxgiFormat = DXGIFormat
        End Sub
    End Structure

    Private Structure D2D1_RENDER_TARGET_PROPERTIES
        Dim eTargetType As D2D1_RENDER_TARGET_TYPE
        Dim tPixelFormat As D2D1_PIXEL_FORMAT
        Dim sngDpiX As Single
        Dim sngDpiY As Single
        Dim eTargetUsage As D2D1_RENDER_TARGET_USAGE
        Dim eMinFeatureLevel As D2D1_FEATURE_LEVEL
        Sub New(PixelFormat As D2D1_PIXEL_FORMAT,
                Optional DpiX As Single = 0,
                Optional DpiY As Single = 0,
                Optional TargetType As D2D1_RENDER_TARGET_TYPE = D2D1_RENDER_TARGET_TYPE.D2D1_RENDER_TARGET_TYPE_DEFAULT,
                Optional TargetUsage As D2D1_RENDER_TARGET_USAGE = D2D1_RENDER_TARGET_USAGE.D2D1_RENDER_TARGET_USAGE_NONE,
                Optional MinFeatureLevel As D2D1_FEATURE_LEVEL = D2D1_FEATURE_LEVEL.D2D1_FEATURE_LEVEL_DEFAULT)
            eTargetType = TargetType
            tPixelFormat = PixelFormat
            sngDpiX = DpiX
            sngDpiY = DpiY
            eTargetUsage = TargetUsage
            eMinFeatureLevel = MinFeatureLevel
        End Sub
    End Structure

    Private Structure D2D1_TAG
        Dim value As ULong
    End Structure

    Public Structure D2D1_RECT_F
        Dim left As Single
        Dim top As Single
        Dim right As Single
        Dim bottom As Single
        Sub New(Optional sngLeft As Single = 0,
                Optional sngTop As Single = 0,
                Optional sngRight As Single = 0,
                Optional sngBottom As Single = 0)
            left = sngLeft
            top = sngTop
            right = sngRight
            bottom = sngBottom
        End Sub
    End Structure

    Public Structure D2D1_COLOR_F
        Dim r As Single
        Dim g As Single
        Dim b As Single
        Dim a As Single
        Sub New(Optional sngA As Single = 1,
                Optional sngR As Single = 0,
                Optional sngG As Single = 0,
                Optional sngB As Single = 0)
            r = sngR
            g = sngG
            b = sngB
            a = sngA
        End Sub
    End Structure

    Public Structure D2D1_POINT_2F
        Dim x As Single
        Dim y As Single
        Sub New(sngX As Single, sngY As Single)
            x = sngX
            y = sngY
        End Sub
    End Structure

    Public Structure D2D1_VECTOR_2F
        Dim x As Single
        Dim y As Single
        Sub New(sngX As Single, sngY As Single)
            x = sngX
            y = sngY
        End Sub
    End Structure

    Public Structure D2D1_VECTOR_3F
        Dim x As Single
        Dim y As Single
        Dim z As Single
        Sub New(sngX As Single, sngY As Single, sngZ As Single)
            x = sngX
            y = sngY
            z = sngZ
        End Sub
    End Structure

    Public Structure D2D1_VECTOR_4F
        Dim x As Single
        Dim y As Single
        Dim z As Single
        Dim w As Single
        Sub New(sngX As Single, sngY As Single, sngZ As Single, sngW As Single)
            x = sngX
            y = sngY
            z = sngZ
            w = sngW
        End Sub
    End Structure

    Public Structure D2D1_MATRIX_3X2_F
        Dim m_11 As Single
        Dim m_12 As Single
        Dim m_21 As Single
        Dim m_22 As Single
        Dim m_31 As Single
        Dim m_32 As Single
        Sub New(sng11 As Single, sng12 As Single,
                sng21 As Single, sng22 As Single,
                sng31 As Single, sng32 As Single)
            m_11 = sng11
            m_12 = sng12
            m_21 = sng21
            m_22 = sng22
            m_31 = sng31
            m_32 = sng32
        End Sub
    End Structure

    Public Structure D2D1_MATRIX_4X4_F
        Dim m_11 As Single
        Dim m_12 As Single
        Dim m_13 As Single
        Dim m_14 As Single
        Dim m_21 As Single
        Dim m_22 As Single
        Dim m_23 As Single
        Dim m_24 As Single
        Dim m_31 As Single
        Dim m_32 As Single
        Dim m_33 As Single
        Dim m_34 As Single
        Dim m_41 As Single
        Dim m_42 As Single
        Dim m_43 As Single
        Dim m_44 As Single
        Sub New(sng11 As Single, sng12 As Single, sng13 As Single, sng14 As Single,
                sng21 As Single, sng22 As Single, sng23 As Single, sng24 As Single,
                sng31 As Single, sng32 As Single, sng33 As Single, sng34 As Single,
                sng41 As Single, sng42 As Single, sng43 As Single, sng44 As Single)
            m_11 = sng11
            m_12 = sng12
            m_13 = sng13
            m_14 = sng14
            m_21 = sng21
            m_22 = sng22
            m_23 = sng23
            m_24 = sng24
            m_31 = sng31
            m_32 = sng32
            m_33 = sng33
            m_34 = sng34
            m_41 = sng41
            m_42 = sng42
            m_43 = sng43
            m_44 = sng44
        End Sub
    End Structure

    Public Structure D2D1_MATRIX_5X4_F
        Dim m_11 As Single
        Dim m_12 As Single
        Dim m_13 As Single
        Dim m_14 As Single
        Dim m_21 As Single
        Dim m_22 As Single
        Dim m_23 As Single
        Dim m_24 As Single
        Dim m_31 As Single
        Dim m_32 As Single
        Dim m_33 As Single
        Dim m_34 As Single
        Dim m_41 As Single
        Dim m_42 As Single
        Dim m_43 As Single
        Dim m_44 As Single
        Dim m_51 As Single
        Dim m_52 As Single
        Dim m_53 As Single
        Dim m_54 As Single
        Sub New(sng11 As Single, sng12 As Single, sng13 As Single, sng14 As Single,
                sng21 As Single, sng22 As Single, sng23 As Single, sng24 As Single,
                sng31 As Single, sng32 As Single, sng33 As Single, sng34 As Single,
                sng41 As Single, sng42 As Single, sng43 As Single, sng44 As Single,
                sng51 As Single, sng52 As Single, sng53 As Single, sng54 As Single)
            m_11 = sng11
            m_12 = sng12
            m_13 = sng13
            m_14 = sng14
            m_21 = sng21
            m_22 = sng22
            m_23 = sng23
            m_24 = sng24
            m_31 = sng31
            m_32 = sng32
            m_33 = sng33
            m_34 = sng34
            m_41 = sng41
            m_42 = sng42
            m_43 = sng43
            m_44 = sng44
            m_51 = sng51
            m_52 = sng52
            m_53 = sng53
            m_54 = sng54
        End Sub
    End Structure

#End Region

#Region "API"
    <DllImport("windowscodecs.dll", EntryPoint:="WICCreateImagingFactory_Proxy")>
    <PreserveSig> Private Shared Function CreateImagingFactory(<[In]> SDKVersion As Integer,
                                                               <Out> ByRef ppIImagingFactory As IntPtr) As Integer
    End Function

    <DllImport("windowscodecs.dll", EntryPoint:="IWICImagingFactory_CreateBitmap_Proxy")>
    <PreserveSig> Private Shared Function CreateWicBitmap(<[In]> pIImagingFactory As IntPtr,
                                                          <[In]> uiWidth As Integer,
                                                          <[In]> uiHeight As Integer,
                                                          <[In], MarshalAs(UnmanagedType.LPStruct)> pixelFormat As Guid,
                                                          <[In]> eOption As WICBitmapCreateCacheOption,
                                                          <Out> ByRef ppIWICBitmap As IntPtr) As Integer
    End Function

    <DllImport("windowscodecs.dll", EntryPoint:="WICConvertBitmapSource")>
    <PreserveSig> Private Shared Function ConvertWicBitmapSource(<[In]> ByRef dstFormat As Guid,
                                                                 <[In]> pISrc As IntPtr,
                                                                 <Out> ByRef ppIDst As IntPtr) As Integer
    End Function

    <DllImport("windowscodecs.dll", EntryPoint:="IWICBitmapSource_CopyPixels_Proxy")>
    <PreserveSig> Private Shared Function CopyWicPixels(<[In]> pIWICBitmapSource As IntPtr,
                                                        <[In]> prc As IntPtr,
                                                        <[In]> cbStride As Integer,
                                                        <[In]> cbBufferSize As Integer,
                                                        <[In]> pbBuffer As IntPtr) As Integer
    End Function

    <DllImport("windowscodecs.dll", EntryPoint:="IWICImagingFactory_CreateDecoderFromFilename_Proxy")>
    <PreserveSig> Private Shared Function CreateWICDecoderFromFilename(<[In]> pIImagingFactory As IntPtr,
                                                                       <[In], MarshalAs(UnmanagedType.LPWStr)> wzFilename As String,
                                                                       <[In], MarshalAs(UnmanagedType.LPStruct)> pguidVendor As Guid,
                                                                       <[In]> dwDesiredAccess As WICAccessOption,
                                                                       <[In]> metadataOptions As WICDecodeOptions,
                                                                       <Out> ByRef ppIWICBitmapDecoder As IntPtr) As Integer
    End Function

    <DllImport("windowscodecs.dll", EntryPoint:="IWICBitmapDecoder_GetFrameCount_Proxy")>
    <PreserveSig> Private Shared Function GetWICFrameCount(<[In]> pIWICBitmapDecoder As IntPtr,
                                                           <Out> ByRef pFrameCount As Integer) As Integer
    End Function

    <DllImport("windowscodecs.dll", EntryPoint:="IWICBitmapDecoder_GetFrame_Proxy")>
    <PreserveSig> Private Shared Function GetWICBitmapFrameDecode(<[In]> pIWICBitmapDecoder As IntPtr,
                                                                  <[In]> index As Integer,
                                                                  <Out> ByRef ppIWICBitmapFrameDecode As IntPtr) As Integer
    End Function

    <DllImport("windowscodecs.dll", EntryPoint:="IWICImagingFactory_CreateBitmapFromSource_Proxy")>
    <PreserveSig> Private Shared Function CreateWICBitmapFromSource(<[In]> pIImagingFactory As IntPtr,
                                                                    <[In]> pIBitmapSource As IntPtr,
                                                                    <[In]> options As WICBitmapCreateCacheOption,
                                                                    <Out> ByRef ppIWICBitmap As IntPtr) As Integer
    End Function

    <DllImport("windowscodecs.dll", EntryPoint:="IWICBitmapSource_GetSize_Proxy")>
    <PreserveSig> Private Shared Function GetWICBitmapSourceSize(<[In]> pIWICBitmapSource As IntPtr,
                                                                 <Out> ByRef puiWidth As Integer,
                                                                 <Out> ByRef puiHeight As Integer) As Integer
    End Function

    <DllImport("windowscodecs.dll", EntryPoint:="IWICBitmapSource_CopyPixels_Proxy")>
    <PreserveSig> Private Shared Function CopyWICBitmapPixels(<[In]> pIWICBitmapSource As IntPtr,
                                                              <[In]> prc As IntPtr,
                                                              <[In]> cbStride As Integer,
                                                              <[In]> cbBufferSize As Integer,
                                                              <[In]> pbBuffer As IntPtr) As Integer
    End Function

    <DllImport("D2d1.dll", EntryPoint:="D2D1CreateFactory")>
    <PreserveSig> Private Shared Function D2D1CreateFactory(<[In]> factoryType As D2D1_FACTORY_TYPE,
                                                            <[In]> ByRef riid As Guid,
                                                            <[In]> ByRef pFactoryOptions As D2D1_FACTORY_OPTIONS,
                                                            <Out> ByRef ppIFactory As IntPtr) As Integer
    End Function
#End Region

#Region "WIC & D2D1 Delegates"
    Private Delegate Function IWICImagingFactory_CreateBitmapFromHBITMAP(<[In]> this As IntPtr,
                                                                         <[In]> hBitmap As IntPtr,
                                                                         <[In]> hPalette As IntPtr,
                                                                         <[In]> options As WICBitmapAlphaChannelOption,
                                                                         <Out> ByRef ppIWICBitmap As IntPtr) As Integer

    Private Delegate Function ID2D1Factory_CreateWicBitmapRenderTarget(<[In]> this As IntPtr,
                                                                       <[In]> pIWICBitmap As IntPtr,
                                                                       <[In]> ByRef renderTargetProperties As D2D1_RENDER_TARGET_PROPERTIES,
                                                                       <Out> ByRef ppID2D1RenderTarget As IntPtr) As Integer

    Private Delegate Sub ID2D1DeviceContext_Clear(<[In]> this As IntPtr,
                                                  <[In]> ByRef color As D2D1_COLOR_F)

    Private Delegate Sub ID2D1DeviceContext_BeginDraw(<[In]> this As IntPtr)

    Private Delegate Function ID2D1DeviceContext_EndDraw(<[In]> this As IntPtr,
                                                         <Out> ByRef tag1 As D2D1_TAG,
                                                         <Out> ByRef tag2 As D2D1_TAG) As Integer

    Private Delegate Function ID2D1DeviceContext_CreateColorContext(<[In]> this As IntPtr,
                                                                    <[In]> space As D2D1_COLOR_SPACE,
                                                                    <[In]> pProfile As IntPtr,
                                                                    <[In]> profileSize As Integer,
                                                                    <Out> ByRef ppID2D1ColorContext As IntPtr) As Integer

    Private Delegate Function ID2D1DeviceContext_CreateEffect(<[In]> this As IntPtr,
                                                              <Out> ByRef effectId As Guid,
                                                              <Out> ByRef ppID2D1Effect As IntPtr) As Integer

    Private Delegate Function ID2D1DeviceContext_GetImageLocalBounds(<[In]> this As IntPtr,
                                                                     <[In]> pID2D1Image As IntPtr,
                                                                     <Out> ByRef localBounds As D2D1_RECT_F) As Integer

    Private Delegate Sub ID2D1DeviceContext_DrawImage(<[In]> this As IntPtr,
                                                      <[In]> pID2D1Image As IntPtr,
                                                      <[In]> ByRef targetOffset As D2D1_POINT_2F,
                                                      <[In]> ByRef imageRectangle As D2D1_RECT_F,
                                                      <[In]> interpolationMode As D2D1_INTERPOLATION_MODE,
                                                      <[In]> compositeMode As D2D1_COMPOSITE_MODE)


    'STDMETHOD(CreateLookupTable3D)(
    '     D2D1_BUFFER_PRECISION precision,
    '     _In_reads_(3) Const UINT32 *extents,
    '     _In_reads_(dataCount) CONST BYTE *data,
    '     UINT32 dataCount,
    '     _In_reads_(2) Const UINT32 *strides,
    '     _Outptr_ ID2D1LookupTable3D **lookupTable 
    '     ) PURE;
    Private Delegate Function ID2D1DeviceContext_CreateLookupTable3D(<[In]> this As IntPtr,
                                                                     <[In]> precision As D2D1_BUFFER_PRECISION,
                                                                     <[In], MarshalAs(UnmanagedType.LPArray)> extents As Integer(),
                                                                     <[In], MarshalAs(UnmanagedType.LPArray)> data As Byte(),
                                                                     <[In]> dataCount As Integer,
                                                                     <[In], MarshalAs(UnmanagedType.LPArray)> strides As Integer(),
                                                                     <Out> ByRef ppID2D1LookupTable3D As IntPtr) As Integer

    Private Delegate Function ID2D1Effect_SetValue(<[In]> this As IntPtr,
                                                   <[In]> index As Integer,
                                                   <[In]> type As D2D1_PROPERTY_TYPE,
                                                   <[In], MarshalAs(UnmanagedType.LPArray)> data As Byte(),
                                                   <[In]> dataSize As Integer) As Integer

    Private Delegate Function ID2D1Effect_GetValue(<[In]> this As IntPtr,
                                                   <[In]> index As Integer,
                                                   <[In]> type As D2D1_PROPERTY_TYPE,
                                                   <Out, MarshalAs(UnmanagedType.LPArray)> data As Byte(),
                                                   <[In]> dataSize As Integer) As Integer

    Private Delegate Sub ID2D1Effect_SetInput(<[In]> this As IntPtr,
                                              <[In]> index As Integer,
                                              <[In]> pID2D1Image As IntPtr,
                                              <[In], MarshalAs(UnmanagedType.Bool)> invalidate As Boolean)

    Private Delegate Sub ID2D1Effect_GetOutput(<[In]> this As IntPtr,
                                               <Out> ByRef ppID2D1Image As IntPtr)

#End Region

#Region "WIC & D2D1 Functions"
    Private Function WICImagingFactory_CreateBitmapFromHBITMAP(pIWICImagingFactory As IntPtr,
                                                               hBitmap As IntPtr,
                                                               hPalette As IntPtr,
                                                               options As WICBitmapAlphaChannelOption,
                                                               ByRef pIWICBitmap As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False

        If pIWICImagingFactory <> IntPtr.Zero AndAlso GetVTablePtr(pIWICImagingFactory, VTable_Interfaces.CreateBitmapFromHBITMAP, pVTablePtr) Then

            Dim Method As IWICImagingFactory_CreateBitmapFromHBITMAP = DirectCast(Marshal.GetDelegateForFunctionPointer(pVTablePtr,
                                                                                  GetType(IWICImagingFactory_CreateBitmapFromHBITMAP)),
                                                                                  IWICImagingFactory_CreateBitmapFromHBITMAP)

            If Method.Invoke(pIWICImagingFactory, hBitmap, hPalette, options, pIWICBitmap) = S_OK Then bolRet = True

        End If

        Return bolRet

    End Function

    Private Function D2D1Factory_CreateWicBitmapRenderTarget(pID2D1Factory As IntPtr,
                                                             pIWICBitmap As IntPtr,
                                                             renderTargetProperties As D2D1_RENDER_TARGET_PROPERTIES,
                                                             ByRef pID2D1RenderTarget As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False

        If pID2D1Factory <> IntPtr.Zero AndAlso GetVTablePtr(pID2D1Factory, VTable_Interfaces.CreateWicBitmapRenderTarget, pVTablePtr) Then

            Dim Method As ID2D1Factory_CreateWicBitmapRenderTarget = DirectCast(Marshal.GetDelegateForFunctionPointer(pVTablePtr,
                                                                                GetType(ID2D1Factory_CreateWicBitmapRenderTarget)),
                                                                                ID2D1Factory_CreateWicBitmapRenderTarget)

            If Method.Invoke(pID2D1Factory, pIWICBitmap, renderTargetProperties, pID2D1RenderTarget) = S_OK Then bolRet = True

        End If

        Return bolRet

    End Function

    Private Sub D2D1DeviceContext_Clear(pID2D1DeviceContext As IntPtr,
                                        Optional backColor As D2D1_COLOR_F? = Nothing)
        Dim pVTablePtr As IntPtr
        Dim color As D2D1_COLOR_F = If(backColor.HasValue, CType(backColor, D2D1_COLOR_F), New D2D1_COLOR_F(1, 0, 0, 0))

        If pID2D1DeviceContext <> IntPtr.Zero AndAlso GetVTablePtr(pID2D1DeviceContext, VTable_Interfaces.Clear, pVTablePtr) Then

            Dim Method As ID2D1DeviceContext_Clear = DirectCast(Marshal.GetDelegateForFunctionPointer(pVTablePtr,
                                                                GetType(ID2D1DeviceContext_Clear)),
                                                                ID2D1DeviceContext_Clear)

            Method.Invoke(pID2D1DeviceContext, color)

        End If

    End Sub

    Private Sub D2D1DeviceContext_BeginDraw(pID2D1DeviceContext As IntPtr)

        Dim pVTablePtr As IntPtr

        If pID2D1DeviceContext <> IntPtr.Zero AndAlso GetVTablePtr(pID2D1DeviceContext, VTable_Interfaces.BeginDraw, pVTablePtr) Then

            Dim Method As ID2D1DeviceContext_BeginDraw = DirectCast(Marshal.GetDelegateForFunctionPointer(pVTablePtr,
                                                                    GetType(ID2D1DeviceContext_BeginDraw)),
                                                                    ID2D1DeviceContext_BeginDraw)

            Method.Invoke(pID2D1DeviceContext)

        End If

    End Sub

    Private Function D2D1DeviceContext_EndDraw(pID2D1DeviceContext As IntPtr,
                                               Optional ByRef tag1 As D2D1_TAG = Nothing,
                                               Optional ByRef tag2 As D2D1_TAG = Nothing) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False

        If pID2D1DeviceContext <> IntPtr.Zero AndAlso GetVTablePtr(pID2D1DeviceContext, VTable_Interfaces.EndDraw, pVTablePtr) Then

            Dim Method As ID2D1DeviceContext_EndDraw = DirectCast(Marshal.GetDelegateForFunctionPointer(pVTablePtr,
                                                                  GetType(ID2D1DeviceContext_EndDraw)),
                                                                  ID2D1DeviceContext_EndDraw)

            If Method.Invoke(pID2D1DeviceContext, tag1, tag2) = S_OK Then bolRet = True

        End If

        Return bolRet

    End Function

    Private Function D2D1DeviceContext_CreateColorContext(pID2D1DeviceContext As IntPtr,
                                                          space As D2D1_COLOR_SPACE,
                                                          pProfile As IntPtr,
                                                          profileSize As Integer,
                                                          ByRef pID2D1ColorContext As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False

        If pID2D1DeviceContext <> IntPtr.Zero AndAlso GetVTablePtr(pID2D1DeviceContext, VTable_Interfaces.CreateColorContext, pVTablePtr) Then

            Dim Method As ID2D1DeviceContext_CreateColorContext = DirectCast(Marshal.GetDelegateForFunctionPointer(pVTablePtr,
                                                                             GetType(ID2D1DeviceContext_CreateColorContext)),
                                                                             ID2D1DeviceContext_CreateColorContext)

            If Method.Invoke(pID2D1DeviceContext, space, pProfile, profileSize, pID2D1ColorContext) = S_OK Then bolRet = True

        End If

        Return bolRet

    End Function

    Private Sub D2D1DeviceContext_DrawImage(pID2D1DeviceContext As IntPtr,
                                            pID2D1Image As IntPtr,
                                            targetOffset As D2D1_POINT_2F,
                                            imageRectangle As D2D1_RECT_F,
                                            Optional interpolationMode As D2D1_INTERPOLATION_MODE = D2D1_INTERPOLATION_MODE.D2D1_INTERPOLATION_MODE_LINEAR,
                                            Optional compositeMode As D2D1_COMPOSITE_MODE = D2D1_COMPOSITE_MODE.D2D1_COMPOSITE_MODE_SOURCE_OVER)
        Dim pVTablePtr As IntPtr

        If pID2D1DeviceContext <> IntPtr.Zero AndAlso GetVTablePtr(pID2D1DeviceContext, VTable_Interfaces.DrawImage, pVTablePtr) Then

            Dim Method As ID2D1DeviceContext_DrawImage = DirectCast(Marshal.GetDelegateForFunctionPointer(pVTablePtr,
                                                                    GetType(ID2D1DeviceContext_DrawImage)),
                                                                    ID2D1DeviceContext_DrawImage)

            Method.Invoke(pID2D1DeviceContext, pID2D1Image, targetOffset, imageRectangle, interpolationMode, compositeMode)

        End If

    End Sub

    Private Function D2D1DeviceContext_CreateLookupTable3D(pID2D1DeviceContext As IntPtr,
                                                           precision As D2D1_BUFFER_PRECISION,
                                                           extents As Integer(),
                                                           data As Byte(),
                                                           dataCount As Integer,
                                                           strides As Integer(),
                                                           ByRef pID2D1LookupTable3D As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False

        If pID2D1DeviceContext <> IntPtr.Zero AndAlso GetVTablePtr(pID2D1DeviceContext, VTable_Interfaces.CreateLookupTable3D, pVTablePtr) Then

            Dim Method As ID2D1DeviceContext_CreateLookupTable3D = DirectCast(Marshal.GetDelegateForFunctionPointer(pVTablePtr,
                                                                              GetType(ID2D1DeviceContext_CreateLookupTable3D)),
                                                                              ID2D1DeviceContext_CreateLookupTable3D)

            If Method.Invoke(pID2D1DeviceContext, precision, extents, data, dataCount, strides, pID2D1LookupTable3D) = S_OK Then bolRet = True

        End If

        Return bolRet

    End Function

    Private Function D2D1DeviceContext_CreateEffect(pID2D1DeviceContext As IntPtr,
                                                    effectId As Guid,
                                                    ByRef pID2D1Effect As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False

        If pID2D1DeviceContext <> IntPtr.Zero AndAlso GetVTablePtr(pID2D1DeviceContext, VTable_Interfaces.CreateEffect, pVTablePtr) Then

            Dim Method As ID2D1DeviceContext_CreateEffect = DirectCast(Marshal.GetDelegateForFunctionPointer(pVTablePtr,
                                                                       GetType(ID2D1DeviceContext_CreateEffect)),
                                                                       ID2D1DeviceContext_CreateEffect)

            If Method.Invoke(pID2D1DeviceContext, effectId, pID2D1Effect) = S_OK Then bolRet = True

        End If

        Return bolRet

    End Function

    Private Function D2D1DeviceContext_GetImageLocalBounds(pID2D1DeviceContext As IntPtr,
                                                           pID2D1Image As IntPtr,
                                                           ByRef localBounds As D2D1_RECT_F) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False

        If pID2D1DeviceContext <> IntPtr.Zero AndAlso GetVTablePtr(pID2D1DeviceContext, VTable_Interfaces.GetImageLocalBounds, pVTablePtr) Then

            Dim Method As ID2D1DeviceContext_GetImageLocalBounds = DirectCast(Marshal.GetDelegateForFunctionPointer(pVTablePtr,
                                                                              GetType(ID2D1DeviceContext_GetImageLocalBounds)),
                                                                              ID2D1DeviceContext_GetImageLocalBounds)

            If Method.Invoke(pID2D1DeviceContext, pID2D1Image, localBounds) = S_OK Then bolRet = True

        End If

        Return bolRet

    End Function

    Private Function D2D1Effect_SetValue(pID2D1Effect As IntPtr,
                                         index As Integer,
                                         type As D2D1_PROPERTY_TYPE,
                                         data As Byte(),
                                         dataSize As Integer) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False

        If pID2D1Effect <> IntPtr.Zero AndAlso GetVTablePtr(pID2D1Effect, VTable_Interfaces.SetValue, pVTablePtr) Then

            Dim Method As ID2D1Effect_SetValue = DirectCast(Marshal.GetDelegateForFunctionPointer(pVTablePtr,
                                                            GetType(ID2D1Effect_SetValue)),
                                                            ID2D1Effect_SetValue)

            If Method.Invoke(pID2D1Effect, index, type, data, dataSize) = S_OK Then bolRet = True

        End If

        Return bolRet

    End Function

    Private Function D2D1Effect_GetValue(pID2D1Effect As IntPtr,
                                         index As Integer,
                                         type As D2D1_PROPERTY_TYPE,
                                         ByRef data As Byte(),
                                         dataSize As Integer) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False

        If pID2D1Effect <> IntPtr.Zero AndAlso GetVTablePtr(pID2D1Effect, VTable_Interfaces.GetValue, pVTablePtr) Then

            Dim Method As ID2D1Effect_SetValue = DirectCast(Marshal.GetDelegateForFunctionPointer(pVTablePtr,
                                                            GetType(ID2D1Effect_SetValue)),
                                                            ID2D1Effect_SetValue)

            If Method.Invoke(pID2D1Effect, index, type, data, dataSize) = S_OK Then bolRet = True

        End If

        Return bolRet
    End Function


    Private Sub D2D1Effect_SetInput(pID2D1Effect As IntPtr,
                                    index As Integer,
                                    pID2D1Image As IntPtr,
                                    Optional invalidate As Boolean = True)
        Dim pVTablePtr As IntPtr

        If pID2D1Effect <> IntPtr.Zero AndAlso GetVTablePtr(pID2D1Effect, VTable_Interfaces.SetInput, pVTablePtr) Then

            Dim Method As ID2D1Effect_SetInput = DirectCast(Marshal.GetDelegateForFunctionPointer(pVTablePtr,
                                                            GetType(ID2D1Effect_SetInput)),
                                                            ID2D1Effect_SetInput)

            Method.Invoke(pID2D1Effect, index, pID2D1Image, invalidate)

        End If

    End Sub

    Private Function D2D1Effect_GetOutput(pID2D1Effect As IntPtr,
                                          ByRef pID2D1Image As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False

        If pID2D1Effect <> IntPtr.Zero AndAlso GetVTablePtr(pID2D1Effect, VTable_Interfaces.GetOutput, pVTablePtr) Then

            Dim Method As ID2D1Effect_GetOutput = DirectCast(Marshal.GetDelegateForFunctionPointer(pVTablePtr,
                                                             GetType(ID2D1Effect_GetOutput)),
                                                             ID2D1Effect_GetOutput)
            Method.Invoke(pID2D1Effect, pID2D1Image)

            If pID2D1Image <> IntPtr.Zero Then bolRet = True

        End If

        Return bolRet

    End Function

    Private Function GetVTablePtr(pInterface As IntPtr,
                                  intMethodNumber As Integer,
                                  ByRef pVTablePtr As IntPtr) As Boolean
        Dim bolRet As Boolean = False

        If pInterface <> IntPtr.Zero AndAlso intMethodNumber >= 0 Then

            Dim pVTablePtrs As IntPtr() = New IntPtr(0) {}

            Marshal.Copy(IntPtr.Add(Marshal.ReadIntPtr(pInterface),
                                    IntPtr.Size * intMethodNumber),
                                    pVTablePtrs, 0, 1)

            pVTablePtr = pVTablePtrs(0)

            bolRet = True

        End If

        Return bolRet

    End Function
#End Region

#Region "Variablen"
    Private m_bolDisposedValue As Boolean
    Private m_pIWICBitmap As IntPtr
    Private m_pID2D1Factory As IntPtr
    Private m_pIImagingFactory As IntPtr
    Private m_pID2D1RenderTarget As IntPtr
    Private m_pID2D1DeviceContext2 As IntPtr
#End Region

#Region "Class New"
    Public Sub New()

        ' ein IImagingFactory erstellen
        If CreateImagingFactory(WINCODEC_SDK_VERSION2,
                                m_pIImagingFactory) = S_OK Then

            ' ein ID2D1Factory erstellen
            If D2D1CreateFactory(D2D1_FACTORY_TYPE.D2D1_FACTORY_TYPE_SINGLE_THREADED,
                                 New Guid(IID_ID2D1Factory),
                                 New D2D1_FACTORY_OPTIONS(D2D1_DEBUG_LEVEL.D2D1_DEBUG_LEVEL_NONE),
                                 m_pID2D1Factory) = S_OK Then

                ' ein IWICBitmap erstellen
                If CreateWicBitmap(m_pIImagingFactory, 1, 1, New Guid(GUID_WICPixelFormat32bppPBGRA),
                                   WICBitmapCreateCacheOption.WICBitmapCacheOnDemand,
                                   m_pIWICBitmap) = S_OK Then

                    ' ein ID2D1RenderTarget vom IWICBitmap erstellen
                    If D2D1Factory_CreateWicBitmapRenderTarget(m_pID2D1Factory, m_pIWICBitmap,
                                                               New D2D1_RENDER_TARGET_PROPERTIES(New D2D1_PIXEL_FORMAT(D2D1_ALPHA_MODE.D2D1_ALPHA_MODE_PREMULTIPLIED,
                                                                                                                       DXGI_FORMAT.DXGI_FORMAT_B8G8R8A8_UNORM)),
                                                               m_pID2D1RenderTarget) Then

                        ' ein ID2D1DeviceContext2 vom ID2D1RenderTarget erstellen
                        If Marshal.QueryInterface(m_pID2D1RenderTarget,
                                                  New Guid(IID_ID2D1DeviceContext2),
                                                  m_pID2D1DeviceContext2) = S_OK Then

                        End If

                    End If

                End If

            End If

        End If

    End Sub
#End Region

#Region "Public Functions"
    Public Function ImageFileToIWICBitmap(strImageFile As String) As IntPtr

        Dim intFrameCount As Integer
        Dim pIWICBitmapConvert As IntPtr
        Dim pIWICBitmapDecoder As IntPtr
        Dim pIWICBitmapFrameDecode As IntPtr

        ' Existiert die Datei
        If File.Exists(strImageFile) Then

            ' einen IWICBitmapDecoder von der Datei erstellen
            If CreateWICDecoderFromFilename(m_pIImagingFactory, strImageFile, Guid.Empty,
                                            WICAccessOption.GENERIC_READ,
                                            WICDecodeOptions.WICDecodeMetadataCacheOnDemand,
                                            pIWICBitmapDecoder) = S_OK Then

                ' Anzahl der Frames ermitteln
                If GetWICFrameCount(pIWICBitmapDecoder, intFrameCount) = S_OK Then

                    ' ist mindestens ein Frame vorhanden
                    If intFrameCount > 0 Then

                        ' erstes Frame laden -> IWICBitmapFrameDecode
                        If GetWICBitmapFrameDecode(pIWICBitmapDecoder, 0, pIWICBitmapFrameDecode) = S_OK Then

                            ' IWICBitmapFrameDecode in das Pixelformat 32bppPBGRA konvertieren -> IWICBitmapConvert
                            If ConvertWicBitmapSource(New Guid(GUID_WICPixelFormat32bppPBGRA),
                                                      pIWICBitmapFrameDecode, pIWICBitmapConvert) = S_OK Then

                                ' ein IWICBitmap vom IWICBitmapConvert erstellen
                                CreateWICBitmapFromSource(m_pIImagingFactory, pIWICBitmapConvert,
                                                          WICBitmapCreateCacheOption.WICBitmapCacheOnDemand,
                                                          ImageFileToIWICBitmap)

                                ' IWICBitmapConvert löschen
                                Marshal.Release(pIWICBitmapConvert)

                            End If

                            ' IWICBitmapFrameDecode löschen
                            Marshal.Release(pIWICBitmapFrameDecode)

                        End If

                    End If

                End If

                ' IWICBitmapDecoder löschen
                Marshal.Release(pIWICBitmapDecoder)

            End If

        End If

    End Function

    Public Function ImageFileToID2D1Image(strImageFile As String,
                                          Optional scale As D2D1_VECTOR_2F? = Nothing,
                                          Optional orientation As D2D1_BITMAPSOURCE_ORIENTATION = D2D1_BITMAPSOURCE_ORIENTATION.D2D1_BITMAPSOURCE_ORIENTATION_DEFAULT,
                                          Optional interpolationMode As D2D1_BITMAPSOURCE_INTERPOLATION_MODE = D2D1_BITMAPSOURCE_INTERPOLATION_MODE.D2D1_BITMAPSOURCE_INTERPOLATION_MODE_LINEAR,
                                          Optional alphaMode As D2D1_BITMAPSOURCE_ALPHA_MODE = D2D1_BITMAPSOURCE_ALPHA_MODE.D2D1_BITMAPSOURCE_ALPHA_MODE_PREMULTIPLIED,
                                          Optional enableDpiCorrection As Boolean = False) As IntPtr

        Dim pIWICBitmap As IntPtr

        ' Existiert die Datei
        If File.Exists(strImageFile) Then

            Dim scaleBitmap As D2D1_VECTOR_2F = If(scale.HasValue, CType(scale, D2D1_VECTOR_2F), New D2D1_VECTOR_2F(1, 1))

            ' minimaler Wert
            If scaleBitmap.x < 0.01F Then scaleBitmap.x = 0.01F
            If scaleBitmap.y < 0.01F Then scaleBitmap.y = 0.01F

            ' Erstellt ein IWICBitmap von der Datei
            pIWICBitmap = ImageFileToIWICBitmap(strImageFile)

            ' ist ein IWICBitmap vorhanden
            If pIWICBitmap <> IntPtr.Zero Then

                ' konvertiert das IWICBitmap in ein ID2D1Image
                ImageFileToID2D1Image = CreateD2D1BitmapSourceEffect(pIWICBitmap, scaleBitmap, orientation,
                                                                     interpolationMode, alphaMode, enableDpiCorrection)

                ' IWICBitmap löschen
                Marshal.Release(pIWICBitmap)

            End If

        End If

    End Function

    Public Function BitmapToIWICBitmap(bitmap As Bitmap) As IntPtr

        If bitmap IsNot Nothing Then

            WICImagingFactory_CreateBitmapFromHBITMAP(m_pIImagingFactory, bitmap.GetHbitmap, IntPtr.Zero,
                                                      WICBitmapAlphaChannelOption.WICBitmapUsePremultipliedAlpha,
                                                      BitmapToIWICBitmap)

        End If

    End Function

    Public Function IWICBitmapToBitmap(pIWICBitmap As IntPtr) As Bitmap

        Dim intWidth As Integer
        Dim intHeight As Integer
        Dim bmpRet As Bitmap = Nothing
        Dim arrData As Byte() = New Byte() {}

        ' ist ein IWICBitmap vorhanden
        If pIWICBitmap <> IntPtr.Zero Then

            ' Dimensionen der IWICBitmap ermitteln
            If GetWICBitmapSourceSize(pIWICBitmap, intWidth, intHeight) = S_OK Then

                ' Stride berechnen
                Dim intStride As Integer = intWidth * 4

                ' ByteArray entsprechend dimensionieren
                Array.Resize(arrData, intHeight * intStride)

                ' ByteArray pinnen
                Dim hPinnedData As GCHandle = GCHandle.Alloc(arrData, GCHandleType.Pinned)

                ' Pointer vom gepinnten ByteArray
                Dim pPinnedData As IntPtr = hPinnedData.AddrOfPinnedObject

                ' Pixeldaten vom IWICBitmap in das ByteArray kopieren
                If CopyWICBitmapPixels(pIWICBitmap, IntPtr.Zero, intStride, arrData.Length, pPinnedData) = S_OK Then

                    ' neues Bitmap vom ByteArray in den gleichen Dimensionen und PixelFormat wie das IWICBitmap erstellen
                    bmpRet = New Bitmap(intWidth, intHeight, intStride, PixelFormat.Format32bppPArgb, pPinnedData)

                End If

                ' ByteArray unpinnen
                hPinnedData.Free()

            End If

        End If

        Return bmpRet

    End Function

    Public Function ID2D1ImageToBitmap(pID2D1Image As IntPtr,
                                       Optional backColor As D2D1_COLOR_F? = Nothing) As Bitmap

        Dim intWidth As Integer
        Dim intHeight As Integer
        Dim pIWICBitmap As IntPtr
        Dim pID2D1RenderTarget As IntPtr
        Dim pID2D1DeviceContext2 As IntPtr
        Dim tD2D1_RECT_F As D2D1_RECT_F
        Dim bmpRet As Bitmap = Nothing

        ' ist ein ID2D1Image vorhanden
        If pID2D1Image <> IntPtr.Zero Then

            Dim color As D2D1_COLOR_F = If(backColor.HasValue, CType(backColor, D2D1_COLOR_F), New D2D1_COLOR_F(0, 0, 0, 0))

            ' Dimensionen des ID2D1Image ermitteln
            If D2D1DeviceContext_GetImageLocalBounds(m_pID2D1DeviceContext2, pID2D1Image, tD2D1_RECT_F) Then

                ' maximale Begrenzung für bestimmte Effekte
                tD2D1_RECT_F.left = Clamp(tD2D1_RECT_F.left, -4096.0F, 4096.0F)
                tD2D1_RECT_F.top = Clamp(tD2D1_RECT_F.top, -4096.0F, 4096.0F)
                tD2D1_RECT_F.right = Clamp(tD2D1_RECT_F.right, -4096.0F, 4096.0F)
                tD2D1_RECT_F.bottom = Clamp(tD2D1_RECT_F.bottom, -4096.0F, 4096.0F)

                color.a = Clamp(color.a, 0F, 1.0F)
                color.r = Clamp(color.r, 0F, 1.0F)
                color.g = Clamp(color.g, 0F, 1.0F)
                color.b = Clamp(color.b, 0F, 1.0F)

                intWidth = CInt(tD2D1_RECT_F.right)
                intHeight = CInt(tD2D1_RECT_F.bottom)

                ' ein IWICBitmap erstellen
                If CreateWicBitmap(m_pIImagingFactory, intWidth, intHeight, New Guid(GUID_WICPixelFormat32bppPBGRA),
                                   WICBitmapCreateCacheOption.WICBitmapCacheOnDemand, pIWICBitmap) = S_OK Then

                    ' ein ID2D1RenderTarget erstellen
                    If D2D1Factory_CreateWicBitmapRenderTarget(m_pID2D1Factory, pIWICBitmap,
                                                               New D2D1_RENDER_TARGET_PROPERTIES(New D2D1_PIXEL_FORMAT(D2D1_ALPHA_MODE.D2D1_ALPHA_MODE_PREMULTIPLIED,
                                                                                                 DXGI_FORMAT.DXGI_FORMAT_B8G8R8A8_UNORM)),
                                                               pID2D1RenderTarget) Then

                        ' ein ID2D1DeviceContext2 vom ID2D1RenderTarget erstellen
                        If Marshal.QueryInterface(pID2D1RenderTarget,
                                                  New Guid(IID_ID2D1DeviceContext2),
                                                  pID2D1DeviceContext2) = S_OK Then

                            ' mit dem zeichnen in den ID2D1DeviceContext2 beginnen
                            D2D1DeviceContext_BeginDraw(pID2D1DeviceContext2)

                            ' Hintergrundfarbe festlegen
                            D2D1DeviceContext_Clear(pID2D1DeviceContext2, color)

                            ' zeichnet das ID2D1Image in den ID2D1DeviceContext2
                            D2D1DeviceContext_DrawImage(pID2D1DeviceContext2, pID2D1Image, New D2D1_POINT_2F(tD2D1_RECT_F.left, tD2D1_RECT_F.top), tD2D1_RECT_F)

                            ' zeichnen beenden
                            D2D1DeviceContext_EndDraw(pID2D1DeviceContext2)

                            ' ID2D1DeviceContext2 löschen
                            Marshal.Release(pID2D1DeviceContext2)

                        End If

                        ' ID2D1RenderTarget löschen
                        Marshal.Release(pID2D1RenderTarget)

                    End If

                    ' IWICBitmap in ein Bitmap konvertieren
                    bmpRet = IWICBitmapToBitmap(pIWICBitmap)

                    ' IWICBitmap löschen
                    Marshal.Release(pIWICBitmap)

                End If

            End If

        End If

        Return bmpRet

    End Function
#End Region

#Region "Public Functions D2D1-Effekts"
    ' Das Prinzip ist für jeden D2D1Effect gleich. Daher gibt es nur für den ersten Effekt
    ' eine Kommentierung.

    Public Function CreateD2D1TintEffect(pID2D1Image As IntPtr,
                                        Optional tintColor As D2D1_VECTOR_4F? = Nothing,
                                        Optional clampOutput As Boolean = False) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        ' ist ein ID2D1Image vorhanden
        If pID2D1Image <> IntPtr.Zero Then

            ' Standardwerte, wenn nichts vorgegeben wurde
            Dim color As D2D1_VECTOR_4F = If(tintColor.HasValue, CType(tintColor, D2D1_VECTOR_4F), New D2D1_VECTOR_4F(1, 1, 1, 1))

            ' min/max Wert Überprüfung
            color.x = Clamp(color.x, 0F, 1.0F)
            color.y = Clamp(color.y, 0F, 1.0F)
            color.z = Clamp(color.z, 0F, 1.0F)
            color.w = Clamp(color.w, 0F, 1.0F)

            ' D2D1Effect erstellen
            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Tint),
                                              pID2D1Effect) Then

                ' D2D1Image zuweisen (Input)
                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                ' Konvertierung eines Parameters zu einem ByteArray
                bytData = ToByteArray(color)

                ' Parameter zuweisen. Wichtig: Der D2D1_PROPERTY_TYPE muss zum
                ' übergebenen Parameter passen!
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_TINT_PROP.D2D1_TINT_PROP_COLOR,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR4,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(clampOutput)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_TINT_PROP.D2D1_TINT_PROP_CLAMP_OUTPUT,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                           bytData, bytData.Length) Then

                        ' durch den Effekt verändertes D2D1Image holen (Output)
                        D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1TintEffect)

                    End If

                End If

                ' D2D1Effect löschen
                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1CropEffect(pID2D1Image As IntPtr,
                                         Optional cropRect As D2D1_VECTOR_4F? = Nothing,
                                         Optional borderMode As D2D1_BORDER_MODE = D2D1_BORDER_MODE.D2D1_BORDER_MODE_SOFT) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            Dim rect As D2D1_VECTOR_4F = If(cropRect.HasValue, CType(cropRect, D2D1_VECTOR_4F), New D2D1_VECTOR_4F(0, 0, 0, 0))

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Crop),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(rect)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_CROP_PROP.D2D1_CROP_PROP_RECT,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR4,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(CInt(borderMode))
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_CROP_PROP.D2D1_CROP_PROP_BORDER_MODE,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                           bytData, bytData.Length) Then

                        D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1CropEffect)

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1TileEffect(pID2D1Image As IntPtr,
                                         Optional tileRect As D2D1_VECTOR_4F? = Nothing) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            Dim rect As D2D1_VECTOR_4F = If(tileRect.HasValue, CType(tileRect, D2D1_VECTOR_4F), New D2D1_VECTOR_4F(0, 0, 0, 0))

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Tile),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(rect)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_TILE_PROP.D2D1_TILE_PROP_RECT,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR4,
                                       bytData, bytData.Length) Then

                    D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1TileEffect)

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1YCbCrEffect(pID2D1Image0 As IntPtr,
                                          pID2D1Image1 As IntPtr,
                                          Optional transformMatrix As D2D1_MATRIX_3X2_F? = Nothing,
                                          Optional chromaSubsampling As D2D1_YCBCR_CHROMA_SUBSAMPLING = D2D1_YCBCR_CHROMA_SUBSAMPLING.D2D1_YCBCR_CHROMA_SUBSAMPLING_AUTO,
                                          Optional interpolationMode As D2D1_YCBCR_INTERPOLATION_MODE = D2D1_YCBCR_INTERPOLATION_MODE.D2D1_YCBCR_INTERPOLATION_MODE_LINEAR) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image0 <> IntPtr.Zero AndAlso pID2D1Image1 <> IntPtr.Zero Then

            Dim matrix As D2D1_MATRIX_3X2_F = If(transformMatrix.HasValue, CType(transformMatrix, D2D1_MATRIX_3X2_F), New D2D1_MATRIX_3X2_F(1, 0, 0, 1, 0, 0))

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1YCbCr),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image0)
                D2D1Effect_SetInput(pID2D1Effect, 1, pID2D1Image1)

                bytData = ToByteArray(CInt(chromaSubsampling))
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_YCBCR_PROP.D2D1_YCBCR_PROP_CHROMA_SUBSAMPLING,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(matrix)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_YCBCR_PROP.D2D1_YCBCR_PROP_TRANSFORM_MATRIX,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_MATRIX_3X2,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(CInt(interpolationMode))
                        If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_YCBCR_PROP.D2D1_YCBCR_PROP_INTERPOLATION_MODE,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                           bytData, bytData.Length) Then

                            D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1YCbCrEffect)

                        End If

                    End If

                End If
                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1AtlasEffect(pID2D1Image As IntPtr,
                                          Optional inputRect As D2D1_VECTOR_4F? = Nothing,
                                          Optional inputPaddingRect As D2D1_VECTOR_4F? = Nothing) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            Dim input As D2D1_VECTOR_4F = If(inputRect.HasValue, CType(inputRect, D2D1_VECTOR_4F), New D2D1_VECTOR_4F(0, 0, 0, 0))
            Dim padding As D2D1_VECTOR_4F = If(inputPaddingRect.HasValue, CType(inputPaddingRect, D2D1_VECTOR_4F), New D2D1_VECTOR_4F(0, 0, 0, 0))

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Atlas),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(input)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_ATLAS_PROP.D2D1_ATLAS_PROP_INPUT_RECT,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR4,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(padding)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_ATLAS_PROP.D2D1_ATLAS_PROP_INPUT_PADDING_RECT,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR4,
                                           bytData, bytData.Length) Then

                        D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1AtlasEffect)

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1BlendEffect(pID2D1Image0 As IntPtr,
                                          pID2D1Image1 As IntPtr,
                                          Optional borderMode As D2D1_BORDER_MODE = D2D1_BORDER_MODE.D2D1_BORDER_MODE_SOFT) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image0 <> IntPtr.Zero AndAlso pID2D1Image1 <> IntPtr.Zero Then

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Blend),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image0)
                D2D1Effect_SetInput(pID2D1Effect, 1, pID2D1Image1)

                bytData = ToByteArray(CInt(borderMode))
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_BLEND_PROP.D2D1_BLEND_PROP_MODE,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                       bytData, bytData.Length) Then

                    D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1BlendEffect)

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1FloodEffect(Optional floodColor As D2D1_VECTOR_4F? = Nothing) As IntPtr

        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        Dim color As D2D1_VECTOR_4F = If(floodColor.HasValue, CType(floodColor, D2D1_VECTOR_4F), New D2D1_VECTOR_4F(0, 0, 0, 1))

        color.x = Clamp(color.x, 0F, 1.0F)
        color.y = Clamp(color.y, 0F, 1.0F)
        color.z = Clamp(color.z, 0F, 1.0F)
        color.w = Clamp(color.w, 0F, 1.0F)

        If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Flood),
                                              pID2D1Effect) Then

            bytData = ToByteArray(color)
            If D2D1Effect_SetValue(pID2D1Effect,
                                   D2D1_FLOOD_PROP.D2D1_FLOOD_PROP_COLOR,
                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR4,
                                   bytData, bytData.Length) Then

                D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1FloodEffect)

            End If

            Marshal.Release(pID2D1Effect)

        End If

    End Function

    Public Function CreateD2D1SepiaEffect(pID2D1Image As IntPtr,
                                          Optional intensity As Single = 0.5F,
                                          Optional alphaMode As D2D1_ALPHA_MODE = D2D1_ALPHA_MODE.D2D1_ALPHA_MODE_PREMULTIPLIED) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            intensity = Clamp(intensity, 0F, 1.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Sepia),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(intensity)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_SEPIA_PROP.D2D1_SEPIA_PROP_INTENSITY,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(CInt(alphaMode))
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_SEPIA_PROP.D2D1_SEPIA_PROP_ALPHA_MODE,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                           bytData, bytData.Length) Then

                        D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1SepiaEffect)

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1ScaleEffect(pID2D1Image As IntPtr,
                                          Optional scaleImage As D2D1_VECTOR_2F? = Nothing,
                                          Optional centerPoint As D2D1_VECTOR_2F? = Nothing,
                                          Optional sharpness As Single = 0F,
                                          Optional interpolationMode As D2D1_SCALE_INTERPOLATION_MODE = D2D1_SCALE_INTERPOLATION_MODE.D2D1_SCALE_INTERPOLATION_MODE_LINEAR,
                                          Optional borderMode As D2D1_BORDER_MODE = D2D1_BORDER_MODE.D2D1_BORDER_MODE_SOFT) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            sharpness = Clamp(sharpness, 0F, 1.0F)
            Dim scale As D2D1_VECTOR_2F = If(scaleImage.HasValue, CType(scaleImage, D2D1_VECTOR_2F), New D2D1_VECTOR_2F(1, 1))
            Dim center As D2D1_VECTOR_2F = If(centerPoint.HasValue, CType(centerPoint, D2D1_VECTOR_2F), New D2D1_VECTOR_2F(0, 0))

            If scale.x < 0.01F Then scale.x = 0.01F
            If scale.y < 0.01F Then scale.y = 0.01F

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Scale),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(scale)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_SCALE_PROP.D2D1_SCALE_PROP_SCALE,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR2,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(center)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_SCALE_PROP.D2D1_SCALE_PROP_CENTER_POINT,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR2,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(CInt(interpolationMode))
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_SCALE_PROP.D2D1_SCALE_PROP_INTERPOLATION_MODE,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                               bytData, bytData.Length) Then

                            bytData = ToByteArray(CInt(borderMode))
                            If D2D1Effect_SetValue(pID2D1Effect,
                                                   D2D1_SCALE_PROP.D2D1_SCALE_PROP_BORDER_MODE,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                                   bytData, bytData.Length) Then

                                bytData = ToByteArray(sharpness)
                                If D2D1Effect_SetValue(pID2D1Effect,
                                                       D2D1_SCALE_PROP.D2D1_SCALE_PROP_SHARPNESS,
                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                       bytData, bytData.Length) Then

                                    D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1ScaleEffect)

                                End If

                            End If

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1ShadowEffect(pID2D1Image As IntPtr,
                                           Optional shadowColor As D2D1_VECTOR_4F? = Nothing,
                                           Optional standardDeviation As Single = 3.0F,
                                           Optional optimation As D2D1_SHADOW_OPTIMIZATION = D2D1_SHADOW_OPTIMIZATION.D2D1_SHADOW_OPTIMIZATION_BALANCED) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            If standardDeviation < 0F Then standardDeviation = 0F
            Dim color As D2D1_VECTOR_4F = If(shadowColor.HasValue, CType(shadowColor, D2D1_VECTOR_4F), New D2D1_VECTOR_4F(0, 0, 0, 1))

            color.x = Clamp(color.x, 0F, 1.0F)
            color.y = Clamp(color.y, 0F, 1.0F)
            color.z = Clamp(color.z, 0F, 1.0F)
            color.w = Clamp(color.w, 0F, 1.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Shadow),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(standardDeviation)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_SHADOW_PROP.D2D1_SHADOW_PROP_BLUR_STANDARD_DEVIATION,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(color)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_SHADOW_PROP.D2D1_SHADOW_PROP_COLOR,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR4,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(CInt(optimation))
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_SHADOW_PROP.D2D1_SHADOW_PROP_OPTIMIZATION,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                               bytData, bytData.Length) Then

                            D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1ShadowEffect)

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1BorderEffect(pID2D1Image As IntPtr,
                                           Optional edgeModeX As D2D1_BORDER_EDGE_MODE = D2D1_BORDER_EDGE_MODE.D2D1_BORDER_EDGE_MODE_CLAMP,
                                           Optional edgeModeY As D2D1_BORDER_EDGE_MODE = D2D1_BORDER_EDGE_MODE.D2D1_BORDER_EDGE_MODE_CLAMP) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Border),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(CInt(edgeModeX))
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_BORDER_PROP.D2D1_BORDER_PROP_EDGE_MODE_X,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(CInt(edgeModeY))
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_BORDER_PROP.D2D1_BORDER_PROP_EDGE_MODE_Y,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                           bytData, bytData.Length) Then

                        D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1BorderEffect)

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1EmbossEffect(pID2D1Image As IntPtr,
                                           Optional height As Single = 1.0F,
                                           Optional direction As Single = 0F) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            height = Clamp(height, 0F, 10.0F)
            direction = Clamp(direction, 0F, 360.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Emboss),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(height)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_EMBOSS_PROP.D2D1_EMBOSS_PROP_HEIGHT,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(direction)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_EMBOSS_PROP.D2D1_EMBOSS_PROP_DIRECTION,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                           bytData, bytData.Length) Then

                        D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1EmbossEffect)

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1InvertEffect(pID2D1Image As IntPtr) As IntPtr

        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Invert),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)


                D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1InvertEffect)


                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1SharpenEffect(pID2D1Image As IntPtr,
                                            Optional sharpness As Single = 1.0F,
                                            Optional threshold As Single = 0F) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            sharpness = Clamp(sharpness, 0F, 10.0F)
            threshold = Clamp(threshold, 0F, 360.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Sharpen),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(sharpness)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_SHARPEN_PROP.D2D1_SHARPEN_PROP_SHARPNESS,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(threshold)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_SHARPEN_PROP.D2D1_SHARPEN_PROP_THRESHOLD,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                           bytData, bytData.Length) Then

                        D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1SharpenEffect)

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1OpacityEffect(pID2D1Image As IntPtr,
                                            Optional opacity As Single = 1.0F) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            opacity = Clamp(opacity, 0F, 1.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Opacity),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(opacity)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_OPACITY_PROP.D2D1_OPACITY_PROP_OPACITY,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1OpacityEffect)

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1ExposureEffect(pID2D1Image As IntPtr,
                                             Optional exposure As Single = 0F) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            exposure = Clamp(exposure, -2.0F, 2.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Exposure),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(exposure)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_EXPOSURE_PROP.D2D1_EXPOSURE_PROP_EXPOSURE_VALUE,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1ExposureEffect)

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1VignetteEffect(pID2D1Image As IntPtr,
                                             Optional vignetteColor As D2D1_VECTOR_4F? = Nothing,
                                             Optional transitionSize As Single = 0.1F,
                                             Optional strength As Single = 0.5F) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then


            Dim color As D2D1_VECTOR_4F = If(vignetteColor.HasValue, CType(vignetteColor, D2D1_VECTOR_4F), New D2D1_VECTOR_4F(0, 0, 0, 1))

            color.x = Clamp(color.x, 0F, 1.0F)
            color.y = Clamp(color.y, 0F, 1.0F)
            color.z = Clamp(color.z, 0F, 1.0F)
            color.w = Clamp(color.w, 0F, 1.0F)

            strength = Clamp(strength, 0F, 1.0F)
            transitionSize = Clamp(transitionSize, 0F, 1.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Vignette),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(color)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_VIGNETTE_PROP.D2D1_VIGNETTE_PROP_COLOR,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR4,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(transitionSize)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_VIGNETTE_PROP.D2D1_VIGNETTE_PROP_TRANSITION_SIZE,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(strength)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_VIGNETTE_PROP.D2D1_VIGNETTE_PROP_STRENGTH,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                               bytData, bytData.Length) Then

                            D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1VignetteEffect)

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1ContrastEffect(pID2D1Image As IntPtr,
                                             Optional contrast As Single = 0F,
                                             Optional clampInput As Boolean = False) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            contrast = Clamp(contrast, -1.0F, 1.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Contrast),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(contrast)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_CONTRAST_PROP.D2D1_CONTRAST_PROP_CONTRAST,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(clampInput)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_CONTRAST_PROP.D2D1_CONTRAST_PROP_CLAMP_INPUT,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                           bytData, bytData.Length) Then

                        D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1ContrastEffect)

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1RgbToHueEffect(pID2D1Image As IntPtr,
                                             Optional colorSpace As D2D1_RGBTOHUE_OUTPUT_COLOR_SPACE = D2D1_RGBTOHUE_OUTPUT_COLOR_SPACE.D2D1_RGBTOHUE_OUTPUT_COLOR_SPACE_HUE_SATURATION_VALUE) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1RgbToHue),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(CInt(colorSpace))
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_RGBTOHUE_PROP.D2D1_RGBTOHUE_PROP_OUTPUT_COLOR_SPACE,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                       bytData, bytData.Length) Then

                    D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1RgbToHueEffect)

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1HueToRgbEffect(pID2D1Image As IntPtr,
                                             Optional colorSpace As D2D1_HUETORGB_INPUT_COLOR_SPACE = D2D1_HUETORGB_INPUT_COLOR_SPACE.D2D1_HUETORGB_INPUT_COLOR_SPACE_HUE_SATURATION_VALUE) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1HueToRgb),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(CInt(colorSpace))
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_HUETORGB_PROP.D2D1_HUETORGB_PROP_INPUT_COLOR_SPACE,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                       bytData, bytData.Length) Then

                    D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1HueToRgbEffect)

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1CrossFadeEffect(pID2D1Image0 As IntPtr,
                                              pID2D1Image1 As IntPtr,
                                              Optional weight As Single = 0.5F) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image0 <> IntPtr.Zero AndAlso pID2D1Image1 <> IntPtr.Zero Then

            weight = Clamp(weight, 0F, 1.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1CrossFade),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image0)
                D2D1Effect_SetInput(pID2D1Effect, 1, pID2D1Image1)

                bytData = ToByteArray(weight)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_CROSSFADE_PROP.D2D1_CROSSFADE_PROP_WEIGHT,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1CrossFadeEffect)

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1AlphaMaskEffect(pID2D1Image0 As IntPtr,
                                              pID2D1Image1 As IntPtr) As IntPtr
        Dim pID2D1Effect As IntPtr

        If pID2D1Image0 <> IntPtr.Zero AndAlso pID2D1Image1 <> IntPtr.Zero Then

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1AlphaMask),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image0)
                D2D1Effect_SetInput(pID2D1Effect, 1, pID2D1Image1)

                D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1AlphaMaskEffect)

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1PosterizeEffect(pID2D1Image As IntPtr,
                                              Optional redCount As UInteger = 4UI,
                                              Optional greenCount As UInteger = 4UI,
                                              Optional blueCount As UInteger = 4UI) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            redCount = Clamp(redCount, 2UI, 16UI)
            greenCount = Clamp(greenCount, 2UI, 16UI)
            blueCount = Clamp(blueCount, 2UI, 16UI)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Posterize),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(redCount)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_POSTERIZE_PROP.D2D1_POSTERIZE_PROP_RED_VALUE_COUNT,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_UINT32,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(greenCount)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_POSTERIZE_PROP.D2D1_POSTERIZE_PROP_GREEN_VALUE_COUNT,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_UINT32,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(blueCount)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_POSTERIZE_PROP.D2D1_POSTERIZE_PROP_BLUE_VALUE_COUNT,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_UINT32,
                                               bytData, bytData.Length) Then

                            D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1PosterizeEffect)

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1GrayscaleEffect(pID2D1Image As IntPtr) As IntPtr

        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Grayscale),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1GrayscaleEffect)

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1ChromaKeyEffect(pID2D1Image As IntPtr,
                                              Optional chromaKeyColor As D2D1_VECTOR_3F? = Nothing,
                                              Optional tolerance As Single = 0.1F,
                                              Optional invertAlpha As Boolean = False,
                                              Optional feather As Boolean = False) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            Dim color As D2D1_VECTOR_3F = If(chromaKeyColor.HasValue, CType(chromaKeyColor, D2D1_VECTOR_3F), New D2D1_VECTOR_3F(0, 0, 0))

            color.x = Clamp(color.x, 0F, 1.0F)
            color.y = Clamp(color.y, 0F, 1.0F)
            color.z = Clamp(color.z, 0F, 1.0F)

            tolerance = Clamp(tolerance, 0F, 1.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1ChromaKey),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(color)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_CHROMAKEY_PROP.D2D1_CHROMAKEY_PROP_COLOR,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR3,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(tolerance)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_CHROMAKEY_PROP.D2D1_CHROMAKEY_PROP_TOLERANCE,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(invertAlpha)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_CHROMAKEY_PROP.D2D1_CHROMAKEY_PROP_INVERT_ALPHA,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                               bytData, bytData.Length) Then

                            bytData = ToByteArray(feather)
                            If D2D1Effect_SetValue(pID2D1Effect,
                                                   D2D1_CHROMAKEY_PROP.D2D1_CHROMAKEY_PROP_FEATHER,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                   bytData, bytData.Length) Then

                                D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1ChromaKeyEffect)

                            End If

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1CompositeEffect(pID2D1Image0 As IntPtr,
                                              pID2D1Image1 As IntPtr,
                                              Optional compositeMode As D2D1_COMPOSITE_MODE = D2D1_COMPOSITE_MODE.D2D1_COMPOSITE_MODE_SOURCE_OVER) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image0 <> IntPtr.Zero AndAlso pID2D1Image1 <> IntPtr.Zero Then

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Composite),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image0)
                D2D1Effect_SetInput(pID2D1Effect, 1, pID2D1Image1)

                bytData = ToByteArray(CInt(compositeMode))
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_COMPOSITE_PROP.D2D1_COMPOSITE_PROP_MODE,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                       bytData, bytData.Length) Then

                    D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1CompositeEffect)

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function


    Public Function CreateD2D1HistogramEffect(pID2D1Image As IntPtr,
                                              ByRef intHistogramDataSize As Integer,
                                              ByRef sngHistogramData As Single(),
                                              Optional channelSelector As D2D1_CHANNEL_SELECTOR = D2D1_CHANNEL_SELECTOR.D2D1_CHANNEL_SELECTOR_A) As Boolean


        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr
        Dim pHistogramEffect As IntPtr
        Dim bolRet As Boolean = False

        If pID2D1Image <> IntPtr.Zero Then

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Histogram),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(CInt(channelSelector))
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_HISTOGRAM_PROP.D2D1_HISTOGRAM_PROP_CHANNEL_SELECT,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                       bytData, bytData.Length) Then

                    D2D1Effect_GetOutput(pID2D1Effect, pHistogramEffect)

                    If pHistogramEffect <> IntPtr.Zero Then

                        ' Wichtig: Das zurückgegebene D2D1Image vom Effekt muss einmal in einen
                        ' D2D1DeviceContext gezeichnet werden damit die nachfolgenden D2D1Effect_GetValue
                        ' Werte zurück liefern!

                        D2D1DeviceContext_BeginDraw(m_pID2D1DeviceContext2)

                        D2D1DeviceContext_DrawImage(m_pID2D1DeviceContext2, pHistogramEffect, New D2D1_POINT_2F(0, 0), New D2D1_RECT_F(0, 0, 0, 0))

                        D2D1DeviceContext_EndDraw(m_pID2D1DeviceContext2)

                        bytData = ToByteArray(intHistogramDataSize)
                        If D2D1Effect_GetValue(pID2D1Effect,
                                                   D2D1_HISTOGRAM_PROP.D2D1_HISTOGRAM_PROP_NUM_BINS,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_UINT32,
                                                   bytData, bytData.Length) Then

                            intHistogramDataSize = CInt(BitConverter.ToUInt32(bytData, 0))

                            Array.Resize(sngHistogramData, intHistogramDataSize)
                            bytData = ToByteArray(sngHistogramData)
                            If D2D1Effect_GetValue(pID2D1Effect,
                                                       D2D1_HISTOGRAM_PROP.D2D1_HISTOGRAM_PROP_HISTOGRAM_OUTPUT,
                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BLOB,
                                                       bytData, bytData.Length) Then

                                For intItem = 0 To intHistogramDataSize - 1

                                    sngHistogramData(intItem) = BitConverter.ToSingle(bytData, intItem * Marshal.SizeOf(Of Single))

                                Next

                                bolRet = True

                            End If

                        End If

                        Marshal.Release(pHistogramEffect)

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

        Return bolRet

    End Function

    Public Function CreateD2D1HdrToneMapEffect(pID2D1Image As IntPtr,
                                               Optional inputMaxLuminance As Single = 100.0F,
                                               Optional outputMaxLuminance As Single = 100.0F,
                                               Optional displayMode As D2D1_HDRTONEMAP_DISPLAY_MODE = D2D1_HDRTONEMAP_DISPLAY_MODE.D2D1_HDRTONEMAP_DISPLAY_MODE_HDR) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            If inputMaxLuminance < 2 Then inputMaxLuminance = 2
            If outputMaxLuminance < 2 Then outputMaxLuminance = 2

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1HdrToneMap),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(inputMaxLuminance)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_HDRTONEMAP_PROP.D2D1_HDRTONEMAP_PROP_INPUT_MAX_LUMINANCE,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(outputMaxLuminance)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_HDRTONEMAP_PROP.D2D1_HDRTONEMAP_PROP_OUTPUT_MAX_LUMINANCE,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(CInt(displayMode))
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_HDRTONEMAP_PROP.D2D1_HDRTONEMAP_PROP_DISPLAY_MODE,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                               bytData, bytData.Length) Then

                            D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1HdrToneMapEffect)

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1SaturationEffect(pID2D1Image As IntPtr,
                                               Optional saturation As Single = 1.0F) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            saturation = Clamp(saturation, 0F, 1.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Saturation),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(saturation)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_SATURATION_PROP.D2D1_SATURATION_PROP_SATURATION,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1SaturationEffect)

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1StraightenEffect(pID2D1Image As IntPtr,
                                               Optional angle As Single = 0F,
                                               Optional maintainSize As Boolean = False,
                                               Optional scaleMode As D2D1_STRAIGHTEN_SCALE_MODE = D2D1_STRAIGHTEN_SCALE_MODE.D2D1_STRAIGHTEN_SCALE_MODE_LINEAR) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            angle = Clamp(angle, -45.0F, 45.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Straighten),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(angle)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_STRAIGHTEN_PROP.D2D1_STRAIGHTEN_PROP_ANGLE,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(maintainSize)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_STRAIGHTEN_PROP.D2D1_STRAIGHTEN_PROP_MAINTAIN_SIZE,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(CInt(scaleMode))
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_STRAIGHTEN_PROP.D2D1_STRAIGHTEN_PROP_SCALE_MODE,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                               bytData, bytData.Length) Then

                            D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1StraightenEffect)

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1BrightnessEffect(pID2D1Image As IntPtr,
                                               Optional blackPoint As D2D1_VECTOR_2F? = Nothing,
                                               Optional whitePoint As D2D1_VECTOR_2F? = Nothing) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            Dim black As D2D1_VECTOR_2F = If(blackPoint.HasValue, CType(blackPoint, D2D1_VECTOR_2F), New D2D1_VECTOR_2F(0, 0))
            Dim white As D2D1_VECTOR_2F = If(whitePoint.HasValue, CType(whitePoint, D2D1_VECTOR_2F), New D2D1_VECTOR_2F(1, 1))

            black.x = Clamp(black.x, 0F, 1.0F)
            black.y = Clamp(black.y, 0F, 1.0F)
            white.x = Clamp(white.x, 0F, 1.0F)
            white.y = Clamp(white.y, 0F, 1.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Brightness),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(white)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_BRIGHTNESS_PROP.D2D1_BRIGHTNESS_PROP_WHITE_POINT,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR2,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(black)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_BRIGHTNESS_PROP.D2D1_BRIGHTNESS_PROP_BLACK_POINT,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR2,
                                           bytData, bytData.Length) Then

                        D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1BrightnessEffect)

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1MorphologyEffect(pID2D1Image As IntPtr,
                                               Optional width As UInteger = 1UI,
                                               Optional height As UInteger = 1UI,
                                               Optional morphologyMode As D2D1_MORPHOLOGY_MODE = D2D1_MORPHOLOGY_MODE.D2D1_MORPHOLOGY_MODE_ERODE) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            width = Clamp(width, 1UI, 100UI)
            height = Clamp(height, 1UI, 100UI)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Morphology),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(CInt(morphologyMode))
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_MORPHOLOGY_PROP.D2D1_MORPHOLOGY_PROP_MODE,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(width)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_MORPHOLOGY_PROP.D2D1_MORPHOLOGY_PROP_WIDTH,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_UINT32,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(height)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_MORPHOLOGY_PROP.D2D1_MORPHOLOGY_PROP_HEIGHT,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_UINT32,
                                               bytData, bytData.Length) Then

                            D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1MorphologyEffect)

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1TurbulenceEffect(size As D2D1_VECTOR_2F,
                                               Optional baseFrequency As D2D1_VECTOR_2F? = Nothing,
                                               Optional offset As D2D1_VECTOR_2F? = Nothing,
                                               Optional noise As D2D1_TURBULENCE_NOISE = D2D1_TURBULENCE_NOISE.D2D1_TURBULENCE_NOISE_FRACTAL_SUM,
                                               Optional numOctaves As UInteger = 1UI,
                                               Optional seed As Integer = 0,
                                               Optional stitchable As Boolean = False) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        Dim frequency As D2D1_VECTOR_2F = If(baseFrequency.HasValue, CType(baseFrequency, D2D1_VECTOR_2F), New D2D1_VECTOR_2F(0.01, 0.01))
        Dim offs As D2D1_VECTOR_2F = If(offset.HasValue, CType(offset, D2D1_VECTOR_2F), New D2D1_VECTOR_2F(0, 0))

        If numOctaves < 1UI Then numOctaves = 1UI

        If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                          New Guid(CLSID_D2D1Turbulence),
                                          pID2D1Effect) Then

            bytData = ToByteArray(offs)
            If D2D1Effect_SetValue(pID2D1Effect,
                                   D2D1_TURBULENCE_PROP.D2D1_TURBULENCE_PROP_OFFSET,
                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR2,
                                   bytData, bytData.Length) Then

                bytData = ToByteArray(size)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_TURBULENCE_PROP.D2D1_TURBULENCE_PROP_SIZE,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR2,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(frequency)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_TURBULENCE_PROP.D2D1_TURBULENCE_PROP_BASE_FREQUENCY,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR2,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(numOctaves)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_TURBULENCE_PROP.D2D1_TURBULENCE_PROP_NUM_OCTAVES,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_UINT32,
                                               bytData, bytData.Length) Then

                            bytData = ToByteArray(seed)
                            If D2D1Effect_SetValue(pID2D1Effect,
                                                   D2D1_TURBULENCE_PROP.D2D1_TURBULENCE_PROP_SEED,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_INT32,
                                                   bytData, bytData.Length) Then

                                bytData = ToByteArray(CInt(noise))
                                If D2D1Effect_SetValue(pID2D1Effect,
                                                       D2D1_TURBULENCE_PROP.D2D1_TURBULENCE_PROP_NOISE,
                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                                       bytData, bytData.Length) Then

                                    bytData = ToByteArray(stitchable)
                                    If D2D1Effect_SetValue(pID2D1Effect,
                                                           D2D1_TURBULENCE_PROP.D2D1_TURBULENCE_PROP_STITCHABLE,
                                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                           bytData, bytData.Length) Then

                                        D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1TurbulenceEffect)

                                    End If

                                End If

                            End If

                        End If

                    End If

                End If

            End If

            Marshal.Release(pID2D1Effect)

        End If

    End Function

    Public Function CreateD2D1PremultiplyEffect(pID2D1Image As IntPtr) As IntPtr

        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1Premultiply),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1PremultiplyEffect)

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1HueRotationEffect(pID2D1Image As IntPtr,
                                                Optional angle As Single = 0F) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            angle = Clamp(angle, 0F, 360.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1HueRotation),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(angle)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_HUEROTATION_PROP.D2D1_HUEROTATION_PROP_ANGLE,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1HueRotationEffect)

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1SpotDiffuseEffect(pID2D1Image As IntPtr,
                                                Optional spotColor As D2D1_VECTOR_3F? = Nothing,
                                                Optional lightPosition As D2D1_VECTOR_3F? = Nothing,
                                                Optional kernelUnitLength As D2D1_VECTOR_2F? = Nothing,
                                                Optional pointsAt As D2D1_VECTOR_3F? = Nothing,
                                                Optional focus As Single = 1.0F,
                                                Optional limitingConeAngle As Single = 90.0F,
                                                Optional diffuseConstant As Single = 1.0F,
                                                Optional surfaceScale As Single = 1.0F,
                                                Optional scaleMode As D2D1_SPOTDIFFUSE_SCALE_MODE = D2D1_SPOTDIFFUSE_SCALE_MODE.D2D1_SPOTDIFFUSE_SCALE_MODE_LINEAR) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            Dim color As D2D1_VECTOR_3F = If(spotColor.HasValue, CType(spotColor, D2D1_VECTOR_3F), New D2D1_VECTOR_3F(1, 1, 1))
            Dim position As D2D1_VECTOR_3F = If(lightPosition.HasValue, CType(lightPosition, D2D1_VECTOR_3F), New D2D1_VECTOR_3F(0, 0, 0))
            Dim unitLength As D2D1_VECTOR_2F = If(kernelUnitLength.HasValue, CType(kernelUnitLength, D2D1_VECTOR_2F), New D2D1_VECTOR_2F(1, 1))
            Dim points As D2D1_VECTOR_3F = If(pointsAt.HasValue, CType(pointsAt, D2D1_VECTOR_3F), New D2D1_VECTOR_3F(0, 0, 0))

            color.x = Clamp(color.x, 0F, 1.0F)
            color.y = Clamp(color.y, 0F, 1.0F)
            color.z = Clamp(color.z, 0F, 1.0F)

            focus = Clamp(focus, 0F, 200.0F)
            limitingConeAngle = Clamp(limitingConeAngle, 0F, 90.0F)
            diffuseConstant = Clamp(diffuseConstant, 0F, 10000.0F)
            surfaceScale = Clamp(surfaceScale, 0F, 10000.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1SpotDiffuse),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(position)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_SPOTDIFFUSE_PROP.D2D1_SPOTDIFFUSE_PROP_LIGHT_POSITION,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR3,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(points)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_SPOTDIFFUSE_PROP.D2D1_SPOTDIFFUSE_PROP_POINTS_AT,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR3,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(focus)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_SPOTDIFFUSE_PROP.D2D1_SPOTDIFFUSE_PROP_FOCUS,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                               bytData, bytData.Length) Then

                            bytData = ToByteArray(limitingConeAngle)
                            If D2D1Effect_SetValue(pID2D1Effect,
                                                   D2D1_SPOTDIFFUSE_PROP.D2D1_SPOTDIFFUSE_PROP_LIMITING_CONE_ANGLE,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                   bytData, bytData.Length) Then

                                bytData = ToByteArray(diffuseConstant)
                                If D2D1Effect_SetValue(pID2D1Effect,
                                                       D2D1_SPOTDIFFUSE_PROP.D2D1_SPOTDIFFUSE_PROP_DIFFUSE_CONSTANT,
                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                       bytData, bytData.Length) Then

                                    bytData = ToByteArray(surfaceScale)
                                    If D2D1Effect_SetValue(pID2D1Effect,
                                                           D2D1_SPOTDIFFUSE_PROP.D2D1_SPOTDIFFUSE_PROP_SURFACE_SCALE,
                                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                           bytData, bytData.Length) Then

                                        bytData = ToByteArray(color)
                                        If D2D1Effect_SetValue(pID2D1Effect,
                                                               D2D1_SPOTDIFFUSE_PROP.D2D1_SPOTDIFFUSE_PROP_COLOR,
                                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR3,
                                                               bytData, bytData.Length) Then

                                            bytData = ToByteArray(unitLength)
                                            If D2D1Effect_SetValue(pID2D1Effect,
                                                                   D2D1_SPOTDIFFUSE_PROP.D2D1_SPOTDIFFUSE_PROP_KERNEL_UNIT_LENGTH,
                                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR2,
                                                                   bytData, bytData.Length) Then

                                                bytData = ToByteArray(CInt(scaleMode))
                                                If D2D1Effect_SetValue(pID2D1Effect,
                                                                       D2D1_SPOTDIFFUSE_PROP.D2D1_SPOTDIFFUSE_PROP_SCALE_MODE,
                                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                                                       bytData, bytData.Length) Then

                                                    D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1SpotDiffuseEffect)

                                                End If

                                            End If

                                        End If

                                    End If

                                End If

                            End If

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1ColorMatrixEffect(pID2D1Image As IntPtr,
                                                Optional colorMatrix As D2D1_MATRIX_5X4_F? = Nothing,
                                                Optional alphaMode As D2D1_COLORMATRIX_ALPHA_MODE = D2D1_COLORMATRIX_ALPHA_MODE.D2D1_COLORMATRIX_ALPHA_MODE_PREMULTIPLIED,
                                                Optional clampOutput As Boolean = False) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            Dim matrix As D2D1_MATRIX_5X4_F = If(colorMatrix.HasValue, CType(colorMatrix, D2D1_MATRIX_5X4_F), New D2D1_MATRIX_5X4_F(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0))

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1ColorMatrix),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(matrix)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_COLORMATRIX_PROP.D2D1_COLORMATRIX_PROP_COLOR_MATRIX,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_MATRIX_5X4,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(CInt(alphaMode))
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_COLORMATRIX_PROP.D2D1_COLORMATRIX_PROP_ALPHA_MODE,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(clampOutput)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_COLORMATRIX_PROP.D2D1_COLORMATRIX_PROP_CLAMP_OUTPUT,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                               bytData, bytData.Length) Then

                            D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1ColorMatrixEffect)

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D13DTransformEffect(pID2D1Image As IntPtr,
                                                Optional transformMatrix As D2D1_MATRIX_4X4_F? = Nothing,
                                                Optional interpolationMode As D2D1_3DTRANSFORM_INTERPOLATION_MODE = D2D1_3DTRANSFORM_INTERPOLATION_MODE.D2D1_3DTRANSFORM_INTERPOLATION_MODE_LINEAR,
                                                Optional borderMode As D2D1_BORDER_MODE = D2D1_BORDER_MODE.D2D1_BORDER_MODE_SOFT) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            Dim matrix As D2D1_MATRIX_4X4_F = If(transformMatrix.HasValue, CType(transformMatrix, D2D1_MATRIX_4X4_F), New D2D1_MATRIX_4X4_F(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1))

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D13DTransform),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(CInt(interpolationMode))
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_3DTRANSFORM_PROP.D2D1_3DTRANSFORM_PROP_INTERPOLATION_MODE,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(CInt(borderMode))
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_3DTRANSFORM_PROP.D2D1_3DTRANSFORM_PROP_BORDER_MODE,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(matrix)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_3DTRANSFORM_PROP.D2D1_3DTRANSFORM_PROP_TRANSFORM_MATRIX,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_MATRIX_4X4,
                                               bytData, bytData.Length) Then

                            D2D1Effect_GetOutput(pID2D1Effect, CreateD2D13DTransformEffect)

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1GaussianBlurEffect(pID2D1Image As IntPtr,
                                                 Optional standardDeviation As Single = 3.0F,
                                                 Optional optimation As D2D1_GAUSSIANBLUR_OPTIMIZATION = D2D1_GAUSSIANBLUR_OPTIMIZATION.D2D1_GAUSSIANBLUR_OPTIMIZATION_BALANCED,
                                                 Optional borderMode As D2D1_BORDER_MODE = D2D1_BORDER_MODE.D2D1_BORDER_MODE_SOFT) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            If standardDeviation < 0 Then standardDeviation = 0

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1GaussianBlur),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(standardDeviation)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_GAUSSIANBLUR_PROP.D2D1_GAUSSIANBLUR_PROP_STANDARD_DEVIATION,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(CInt(optimation))
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_GAUSSIANBLUR_PROP.D2D1_GAUSSIANBLUR_PROP_OPTIMIZATION,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(CInt(borderMode))
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_GAUSSIANBLUR_PROP.D2D1_GAUSSIANBLUR_PROP_BORDER_MODE,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                               bytData, bytData.Length) Then

                            D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1GaussianBlurEffect)

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1BitmapSourceEffect(pIWICBitmap As IntPtr,
                                                 Optional scaleBitmap As D2D1_VECTOR_2F? = Nothing,
                                                 Optional orientation As D2D1_BITMAPSOURCE_ORIENTATION = D2D1_BITMAPSOURCE_ORIENTATION.D2D1_BITMAPSOURCE_ORIENTATION_DEFAULT,
                                                 Optional interpolationMode As D2D1_BITMAPSOURCE_INTERPOLATION_MODE = D2D1_BITMAPSOURCE_INTERPOLATION_MODE.D2D1_BITMAPSOURCE_INTERPOLATION_MODE_LINEAR,
                                                 Optional alphaMode As D2D1_BITMAPSOURCE_ALPHA_MODE = D2D1_BITMAPSOURCE_ALPHA_MODE.D2D1_BITMAPSOURCE_ALPHA_MODE_PREMULTIPLIED,
                                                 Optional enableDpiCorrection As Boolean = False) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pIWICBitmap <> IntPtr.Zero Then

            Dim scale As D2D1_VECTOR_2F = If(scaleBitmap.HasValue, CType(scaleBitmap, D2D1_VECTOR_2F), New D2D1_VECTOR_2F(1, 1))

            If scale.x < 0.01F Then scale.x = 0.01F
            If scale.y < 0.01F Then scale.y = 0.01F

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1BitmapSource),
                                              pID2D1Effect) Then

                bytData = ToByteArray(pIWICBitmap)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_BITMAPSOURCE_PROP.D2D1_BITMAPSOURCE_PROP_WIC_BITMAP_SOURCE,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_IUNKNOWN,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(scale)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_BITMAPSOURCE_PROP.D2D1_BITMAPSOURCE_PROP_SCALE,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR2,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(CInt(interpolationMode))
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_BITMAPSOURCE_PROP.D2D1_BITMAPSOURCE_PROP_INTERPOLATION_MODE,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                               bytData, bytData.Length) Then

                            bytData = ToByteArray(enableDpiCorrection)
                            If D2D1Effect_SetValue(pID2D1Effect,
                                                   D2D1_BITMAPSOURCE_PROP.D2D1_BITMAPSOURCE_PROP_ENABLE_DPI_CORRECTION,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                   bytData, bytData.Length) Then

                                bytData = ToByteArray(CInt(alphaMode))
                                If D2D1Effect_SetValue(pID2D1Effect,
                                                       D2D1_BITMAPSOURCE_PROP.D2D1_BITMAPSOURCE_PROP_ALPHA_MODE,
                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                                       bytData, bytData.Length) Then

                                    bytData = ToByteArray(CInt(orientation))
                                    If D2D1Effect_SetValue(pID2D1Effect,
                                                           D2D1_BITMAPSOURCE_PROP.D2D1_BITMAPSOURCE_PROP_ORIENTATION,
                                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                                           bytData, bytData.Length) Then

                                        D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1BitmapSourceEffect)

                                    End If

                                End If

                            End If

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1PointDiffuseEffect(pID2D1Image As IntPtr,
                                                 Optional pointColor As D2D1_VECTOR_3F? = Nothing,
                                                 Optional lightPosition As D2D1_VECTOR_3F? = Nothing,
                                                 Optional kernelUnitLength As D2D1_VECTOR_2F? = Nothing,
                                                 Optional diffuseConstant As Single = 1.0F,
                                                 Optional surfaceScale As Single = 1.0F,
                                                 Optional scaleMode As D2D1_POINTDIFFUSE_SCALE_MODE = D2D1_POINTDIFFUSE_SCALE_MODE.D2D1_POINTDIFFUSE_SCALE_MODE_LINEAR) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            Dim color As D2D1_VECTOR_3F = If(pointColor.HasValue, CType(pointColor, D2D1_VECTOR_3F), New D2D1_VECTOR_3F(1, 1, 1))
            Dim position As D2D1_VECTOR_3F = If(lightPosition.HasValue, CType(lightPosition, D2D1_VECTOR_3F), New D2D1_VECTOR_3F(0, 0, 0))
            Dim unitLength As D2D1_VECTOR_2F = If(kernelUnitLength.HasValue, CType(kernelUnitLength, D2D1_VECTOR_2F), New D2D1_VECTOR_2F(1, 1))

            color.x = Clamp(color.x, 0F, 1.0F)
            color.y = Clamp(color.y, 0F, 1.0F)
            color.z = Clamp(color.z, 0F, 1.0F)

            diffuseConstant = Clamp(diffuseConstant, 0F, 10000.0F)
            surfaceScale = Clamp(surfaceScale, 0F, 10000.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1PointDiffuse),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(position)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_POINTDIFFUSE_PROP.D2D1_POINTDIFFUSE_PROP_LIGHT_POSITION,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR3,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(diffuseConstant)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_POINTDIFFUSE_PROP.D2D1_POINTDIFFUSE_PROP_DIFFUSE_CONSTANT,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(surfaceScale)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_POINTDIFFUSE_PROP.D2D1_POINTDIFFUSE_PROP_SURFACE_SCALE,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                               bytData, bytData.Length) Then

                            bytData = ToByteArray(color)
                            If D2D1Effect_SetValue(pID2D1Effect,
                                                   D2D1_POINTDIFFUSE_PROP.D2D1_POINTDIFFUSE_PROP_COLOR,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR3,
                                                   bytData, bytData.Length) Then

                                bytData = ToByteArray(unitLength)
                                If D2D1Effect_SetValue(pID2D1Effect,
                                                       D2D1_POINTDIFFUSE_PROP.D2D1_POINTDIFFUSE_PROP_KERNEL_UNIT_LENGTH,
                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR2,
                                                       bytData, bytData.Length) Then

                                    bytData = ToByteArray(CInt(scaleMode))
                                    If D2D1Effect_SetValue(pID2D1Effect,
                                                           D2D1_POINTDIFFUSE_PROP.D2D1_POINTDIFFUSE_PROP_SCALE_MODE,
                                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                                           bytData, bytData.Length) Then

                                        D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1PointDiffuseEffect)

                                    End If

                                End If

                            End If

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1SpotSpecularEffect(pID2D1Image As IntPtr,
                                                 Optional spotColor As D2D1_VECTOR_3F? = Nothing,
                                                 Optional lightPosition As D2D1_VECTOR_3F? = Nothing,
                                                 Optional kernelUnitLength As D2D1_VECTOR_2F? = Nothing,
                                                 Optional pointsAt As D2D1_VECTOR_3F? = Nothing,
                                                 Optional focus As Single = 1.0F,
                                                 Optional limitingConeAngle As Single = 90.0F,
                                                 Optional specularExponent As Single = 1.0F,
                                                 Optional specularConstant As Single = 1.0F,
                                                 Optional surfaceScale As Single = 1.0F,
                                                 Optional scaleMode As D2D1_SPOTSPECULAR_SCALE_MODE = D2D1_SPOTSPECULAR_SCALE_MODE.D2D1_SPOTSPECULAR_SCALE_MODE_LINEAR) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            Dim color As D2D1_VECTOR_3F = If(spotColor.HasValue, CType(spotColor, D2D1_VECTOR_3F), New D2D1_VECTOR_3F(1, 1, 1))
            Dim position As D2D1_VECTOR_3F = If(lightPosition.HasValue, CType(lightPosition, D2D1_VECTOR_3F), New D2D1_VECTOR_3F(0, 0, 0))
            Dim unitLength As D2D1_VECTOR_2F = If(kernelUnitLength.HasValue, CType(kernelUnitLength, D2D1_VECTOR_2F), New D2D1_VECTOR_2F(1, 1))
            Dim points As D2D1_VECTOR_3F = If(pointsAt.HasValue, CType(pointsAt, D2D1_VECTOR_3F), New D2D1_VECTOR_3F(0, 0, 0))

            color.x = Clamp(color.x, 0F, 1.0F)
            color.y = Clamp(color.y, 0F, 1.0F)
            color.z = Clamp(color.z, 0F, 1.0F)

            focus = Clamp(focus, 0F, 200.0F)
            limitingConeAngle = Clamp(limitingConeAngle, 0F, 90.0F)
            specularExponent = Clamp(specularExponent, 1.0F, 128.0F)
            specularConstant = Clamp(specularConstant, 0F, 10000.0F)
            surfaceScale = Clamp(surfaceScale, 0F, 10000.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1SpotSpecular),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(position)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_SPOTSPECULAR_PROP.D2D1_SPOTSPECULAR_PROP_LIGHT_POSITION,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR3,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(points)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_SPOTSPECULAR_PROP.D2D1_SPOTSPECULAR_PROP_POINTS_AT,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR3,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(focus)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_SPOTSPECULAR_PROP.D2D1_SPOTSPECULAR_PROP_FOCUS,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                               bytData, bytData.Length) Then

                            bytData = ToByteArray(limitingConeAngle)
                            If D2D1Effect_SetValue(pID2D1Effect,
                                                   D2D1_SPOTSPECULAR_PROP.D2D1_SPOTSPECULAR_PROP_LIMITING_CONE_ANGLE,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                   bytData, bytData.Length) Then

                                bytData = ToByteArray(specularExponent)
                                If D2D1Effect_SetValue(pID2D1Effect,
                                                       D2D1_SPOTSPECULAR_PROP.D2D1_SPOTSPECULAR_PROP_SPECULAR_EXPONENT,
                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                       bytData, bytData.Length) Then

                                    bytData = ToByteArray(specularConstant)
                                    If D2D1Effect_SetValue(pID2D1Effect,
                                                           D2D1_SPOTSPECULAR_PROP.D2D1_SPOTSPECULAR_PROP_SPECULAR_CONSTANT,
                                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                            bytData, bytData.Length) Then

                                        bytData = ToByteArray(surfaceScale)
                                        If D2D1Effect_SetValue(pID2D1Effect,
                                                               D2D1_SPOTSPECULAR_PROP.D2D1_SPOTSPECULAR_PROP_SURFACE_SCALE,
                                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                               bytData, bytData.Length) Then

                                            bytData = ToByteArray(color)
                                            If D2D1Effect_SetValue(pID2D1Effect,
                                                                   D2D1_SPOTSPECULAR_PROP.D2D1_SPOTSPECULAR_PROP_COLOR,
                                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR3,
                                                                   bytData, bytData.Length) Then

                                                bytData = ToByteArray(unitLength)
                                                If D2D1Effect_SetValue(pID2D1Effect,
                                                                       D2D1_SPOTSPECULAR_PROP.D2D1_SPOTSPECULAR_PROP_KERNEL_UNIT_LENGTH,
                                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR2,
                                                                       bytData, bytData.Length) Then

                                                    bytData = ToByteArray(CInt(scaleMode))
                                                    If D2D1Effect_SetValue(pID2D1Effect,
                                                                           D2D1_SPOTSPECULAR_PROP.D2D1_SPOTSPECULAR_PROP_SCALE_MODE,
                                                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                                                           bytData, bytData.Length) Then

                                                        D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1SpotSpecularEffect)


                                                    End If

                                                End If

                                            End If

                                        End If

                                    End If

                                End If

                            End If

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1EdgeDetectionEffect(pID2D1Image As IntPtr,
                                                  Optional stength As Single = 0.5F,
                                                  Optional blurRadius As Single = 0F,
                                                  Optional edgedetectionMode As D2D1_EDGEDETECTION_MODE = D2D1_EDGEDETECTION_MODE.D2D1_EDGEDETECTION_MODE_SOBEL,
                                                  Optional overlayEdges As Boolean = False,
                                                  Optional alphaMode As D2D1_ALPHA_MODE = D2D1_ALPHA_MODE.D2D1_ALPHA_MODE_PREMULTIPLIED) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            stength = Clamp(stength, 0F, 1.0F)
            blurRadius = Clamp(blurRadius, 0F, 10.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1EdgeDetection),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(stength)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_EDGEDETECTION_PROP.D2D1_EDGEDETECTION_PROP_STRENGTH,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(blurRadius)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_EDGEDETECTION_PROP.D2D1_EDGEDETECTION_PROP_BLUR_RADIUS,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(CInt(edgedetectionMode))
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_EDGEDETECTION_PROP.D2D1_EDGEDETECTION_PROP_MODE,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                               bytData, bytData.Length) Then

                            bytData = ToByteArray(overlayEdges)
                            If D2D1Effect_SetValue(pID2D1Effect,
                                                   D2D1_EDGEDETECTION_PROP.D2D1_EDGEDETECTION_PROP_OVERLAY_EDGES,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                   bytData, bytData.Length) Then

                                bytData = ToByteArray(CInt(alphaMode))
                                If D2D1Effect_SetValue(pID2D1Effect,
                                                       D2D1_EDGEDETECTION_PROP.D2D1_EDGEDETECTION_PROP_ALPHA_MODE,
                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                                       bytData, bytData.Length) Then

                                    D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1EdgeDetectionEffect)

                                End If

                            End If

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1TableTransferEffect(pID2D1Image As IntPtr,
                                                  Optional redTable As Single() = Nothing,
                                                  Optional greenTable As Single() = Nothing,
                                                  Optional blueTable As Single() = Nothing,
                                                  Optional alphaTable As Single() = Nothing,
                                                  Optional redDisable As Boolean = False,
                                                  Optional greenDisable As Boolean = False,
                                                  Optional blueDisable As Boolean = False,
                                                  Optional alphaDisable As Boolean = False,
                                                  Optional clampOutput As Boolean = False) As IntPtr
        Dim bytData As Byte()
        Dim intCount As Integer
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            If redTable Is Nothing Then redTable = New Single() {0F, 1.0F}
            If greenTable Is Nothing Then greenTable = New Single() {0F, 1.0F}
            If blueTable Is Nothing Then blueTable = New Single() {0F, 1.0F}
            If alphaTable Is Nothing Then alphaTable = New Single() {0F, 1.0F}

            intCount = redTable.Length
            For intItem = 0 To intCount - 1
                redTable(intItem) = Clamp(redTable(intItem), 0F, 1.0F)
            Next

            intCount = greenTable.Length
            For intItem = 0 To intCount - 1
                greenTable(intItem) = Clamp(greenTable(intItem), 0F, 1.0F)
            Next

            intCount = blueTable.Length
            For intItem = 0 To intCount - 1
                blueTable(intItem) = Clamp(blueTable(intItem), 0F, 1.0F)
            Next

            intCount = alphaTable.Length
            For intItem = 0 To intCount - 1
                alphaTable(intItem) = Clamp(alphaTable(intItem), 0F, 1.0F)
            Next

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1TableTransfer),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(redTable)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_TABLETRANSFER_PROP.D2D1_TABLETRANSFER_PROP_RED_TABLE,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BLOB,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(redDisable)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_TABLETRANSFER_PROP.D2D1_TABLETRANSFER_PROP_RED_DISABLE,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(greenTable)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_TABLETRANSFER_PROP.D2D1_TABLETRANSFER_PROP_GREEN_TABLE,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BLOB,
                                               bytData, bytData.Length) Then

                            bytData = ToByteArray(greenDisable)
                            If D2D1Effect_SetValue(pID2D1Effect,
                                                   D2D1_TABLETRANSFER_PROP.D2D1_TABLETRANSFER_PROP_GREEN_DISABLE,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                   bytData, bytData.Length) Then

                                bytData = ToByteArray(blueTable)
                                If D2D1Effect_SetValue(pID2D1Effect,
                                                       D2D1_TABLETRANSFER_PROP.D2D1_TABLETRANSFER_PROP_BLUE_TABLE,
                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BLOB,
                                                       bytData, bytData.Length) Then

                                    bytData = ToByteArray(blueDisable)
                                    If D2D1Effect_SetValue(pID2D1Effect,
                                                           D2D1_TABLETRANSFER_PROP.D2D1_TABLETRANSFER_PROP_BLUE_DISABLE,
                                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                            bytData, bytData.Length) Then

                                        bytData = ToByteArray(alphaTable)
                                        If D2D1Effect_SetValue(pID2D1Effect,
                                                               D2D1_TABLETRANSFER_PROP.D2D1_TABLETRANSFER_PROP_ALPHA_TABLE,
                                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BLOB,
                                                               bytData, bytData.Length) Then

                                            bytData = ToByteArray(alphaDisable)
                                            If D2D1Effect_SetValue(pID2D1Effect,
                                                                   D2D1_TABLETRANSFER_PROP.D2D1_TABLETRANSFER_PROP_ALPHA_DISABLE,
                                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                                   bytData, bytData.Length) Then

                                                bytData = ToByteArray(clampOutput)
                                                If D2D1Effect_SetValue(pID2D1Effect,
                                                                       D2D1_TABLETRANSFER_PROP.D2D1_TABLETRANSFER_PROP_CLAMP_OUTPUT,
                                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                                       bytData, bytData.Length) Then

                                                    D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1TableTransferEffect)


                                                End If

                                            End If

                                        End If

                                    End If

                                End If

                            End If

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1GammaTransferEffect(pID2D1Image As IntPtr,
                                                  Optional redAmplitude As Single = 1.0F,
                                                  Optional redExponent As Single = 1.0F,
                                                  Optional redOffset As Single = 0F,
                                                  Optional greenAmplitude As Single = 1.0F,
                                                  Optional greenExponent As Single = 1.0F,
                                                  Optional greenOffset As Single = 0F,
                                                  Optional blueAmplitude As Single = 1.0F,
                                                  Optional blueExponent As Single = 1.0F,
                                                  Optional blueOffset As Single = 0F,
                                                  Optional alphaAmplitude As Single = 1.0F,
                                                  Optional alphaExponent As Single = 1.0F,
                                                  Optional alphaOffset As Single = 0F,
                                                  Optional redDisable As Boolean = False,
                                                  Optional greenDisable As Boolean = False,
                                                  Optional blueDisable As Boolean = False,
                                                  Optional alphaDisable As Boolean = False,
                                                  Optional clampOutput As Boolean = False) As IntPtr

        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1GammaTransfer),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(redAmplitude)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_GAMMATRANSFER_PROP.D2D1_GAMMATRANSFER_PROP_RED_AMPLITUDE,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(redExponent)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_GAMMATRANSFER_PROP.D2D1_GAMMATRANSFER_PROP_RED_EXPONENT,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(redOffset)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_GAMMATRANSFER_PROP.D2D1_GAMMATRANSFER_PROP_RED_OFFSET,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                               bytData, bytData.Length) Then

                            bytData = ToByteArray(redDisable)
                            If D2D1Effect_SetValue(pID2D1Effect,
                                                   D2D1_GAMMATRANSFER_PROP.D2D1_GAMMATRANSFER_PROP_RED_DISABLE,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                   bytData, bytData.Length) Then

                                bytData = ToByteArray(greenAmplitude)
                                If D2D1Effect_SetValue(pID2D1Effect,
                                                       D2D1_GAMMATRANSFER_PROP.D2D1_GAMMATRANSFER_PROP_GREEN_AMPLITUDE,
                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                       bytData, bytData.Length) Then

                                    bytData = ToByteArray(greenExponent)
                                    If D2D1Effect_SetValue(pID2D1Effect,
                                                           D2D1_GAMMATRANSFER_PROP.D2D1_GAMMATRANSFER_PROP_GREEN_EXPONENT,
                                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                           bytData, bytData.Length) Then

                                        bytData = ToByteArray(greenOffset)
                                        If D2D1Effect_SetValue(pID2D1Effect,
                                                               D2D1_GAMMATRANSFER_PROP.D2D1_GAMMATRANSFER_PROP_GREEN_OFFSET,
                                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                               bytData, bytData.Length) Then

                                            bytData = ToByteArray(greenDisable)
                                            If D2D1Effect_SetValue(pID2D1Effect,
                                                                   D2D1_GAMMATRANSFER_PROP.D2D1_GAMMATRANSFER_PROP_GREEN_DISABLE,
                                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                                   bytData, bytData.Length) Then

                                                bytData = ToByteArray(blueAmplitude)
                                                If D2D1Effect_SetValue(pID2D1Effect,
                                                                       D2D1_GAMMATRANSFER_PROP.D2D1_GAMMATRANSFER_PROP_BLUE_AMPLITUDE,
                                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                                       bytData, bytData.Length) Then

                                                    bytData = ToByteArray(blueExponent)
                                                    If D2D1Effect_SetValue(pID2D1Effect,
                                                                           D2D1_GAMMATRANSFER_PROP.D2D1_GAMMATRANSFER_PROP_BLUE_EXPONENT,
                                                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                                           bytData, bytData.Length) Then

                                                        bytData = ToByteArray(blueOffset)
                                                        If D2D1Effect_SetValue(pID2D1Effect,
                                                                               D2D1_GAMMATRANSFER_PROP.D2D1_GAMMATRANSFER_PROP_BLUE_OFFSET,
                                                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                                               bytData, bytData.Length) Then

                                                            bytData = ToByteArray(blueDisable)
                                                            If D2D1Effect_SetValue(pID2D1Effect,
                                                                                   D2D1_GAMMATRANSFER_PROP.D2D1_GAMMATRANSFER_PROP_BLUE_DISABLE,
                                                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                                                   bytData, bytData.Length) Then

                                                                bytData = ToByteArray(alphaAmplitude)
                                                                If D2D1Effect_SetValue(pID2D1Effect,
                                                                                       D2D1_GAMMATRANSFER_PROP.D2D1_GAMMATRANSFER_PROP_ALPHA_AMPLITUDE,
                                                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                                                       bytData, bytData.Length) Then

                                                                    bytData = ToByteArray(alphaExponent)
                                                                    If D2D1Effect_SetValue(pID2D1Effect,
                                                                                           D2D1_GAMMATRANSFER_PROP.D2D1_GAMMATRANSFER_PROP_ALPHA_EXPONENT,
                                                                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                                                           bytData, bytData.Length) Then

                                                                        bytData = ToByteArray(alphaOffset)
                                                                        If D2D1Effect_SetValue(pID2D1Effect,
                                                                                               D2D1_GAMMATRANSFER_PROP.D2D1_GAMMATRANSFER_PROP_ALPHA_OFFSET,
                                                                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                                                               bytData, bytData.Length) Then

                                                                            bytData = ToByteArray(alphaDisable)
                                                                            If D2D1Effect_SetValue(pID2D1Effect,
                                                                                                   D2D1_GAMMATRANSFER_PROP.D2D1_GAMMATRANSFER_PROP_ALPHA_DISABLE,
                                                                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                                                                   bytData, bytData.Length) Then

                                                                                bytData = ToByteArray(clampOutput)
                                                                                If D2D1Effect_SetValue(pID2D1Effect,
                                                                                                       D2D1_GAMMATRANSFER_PROP.D2D1_GAMMATRANSFER_PROP_CLAMP_OUTPUT,
                                                                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                                                                       bytData, bytData.Length) Then

                                                                                    D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1GammaTransferEffect)

                                                                                End If

                                                                            End If

                                                                        End If

                                                                    End If

                                                                End If

                                                            End If

                                                        End If

                                                    End If

                                                End If

                                            End If

                                        End If

                                    End If

                                End If

                            End If

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1UnPremultiplyEffect(pID2D1Image As IntPtr) As IntPtr

        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1UnPremultiply),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1UnPremultiplyEffect)

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1PointSpecularEffect(pID2D1Image As IntPtr,
                                                  Optional pointColor As D2D1_VECTOR_3F? = Nothing,
                                                  Optional lightPosition As D2D1_VECTOR_3F? = Nothing,
                                                  Optional kernelUnitLength As D2D1_VECTOR_2F? = Nothing,
                                                  Optional specularExponent As Single = 1.0F,
                                                  Optional specularConstant As Single = 1.0F,
                                                  Optional surfaceScale As Single = 1.0F,
                                                  Optional scaleMode As D2D1_POINTSPECULAR_SCALE_MODE = D2D1_POINTSPECULAR_SCALE_MODE.D2D1_POINTSPECULAR_SCALE_MODE_LINEAR) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            Dim color As D2D1_VECTOR_3F = If(pointColor.HasValue, CType(pointColor, D2D1_VECTOR_3F), New D2D1_VECTOR_3F(1, 1, 1))
            Dim position As D2D1_VECTOR_3F = If(lightPosition.HasValue, CType(lightPosition, D2D1_VECTOR_3F), New D2D1_VECTOR_3F(0, 0, 0))
            Dim unitLength As D2D1_VECTOR_2F = If(kernelUnitLength.HasValue, CType(kernelUnitLength, D2D1_VECTOR_2F), New D2D1_VECTOR_2F(1, 1))

            color.x = Clamp(color.x, 0F, 1.0F)
            color.y = Clamp(color.y, 0F, 1.0F)
            color.z = Clamp(color.z, 0F, 1.0F)

            specularExponent = Clamp(specularExponent, 1.0F, 128.0F)
            specularConstant = Clamp(specularConstant, 0F, 10000.0F)
            surfaceScale = Clamp(surfaceScale, 0F, 10000.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1PointSpecular),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(position)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_POINTSPECULAR_PROP.D2D1_POINTSPECULAR_PROP_LIGHT_POSITION,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR3,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(specularExponent)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_POINTSPECULAR_PROP.D2D1_POINTSPECULAR_PROP_SPECULAR_EXPONENT,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(specularConstant)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_POINTSPECULAR_PROP.D2D1_POINTSPECULAR_PROP_SPECULAR_CONSTANT,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                               bytData, bytData.Length) Then

                            bytData = ToByteArray(surfaceScale)
                            If D2D1Effect_SetValue(pID2D1Effect,
                                                   D2D1_POINTSPECULAR_PROP.D2D1_POINTSPECULAR_PROP_SURFACE_SCALE,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                   bytData, bytData.Length) Then

                                bytData = ToByteArray(color)
                                If D2D1Effect_SetValue(pID2D1Effect,
                                                       D2D1_POINTSPECULAR_PROP.D2D1_POINTSPECULAR_PROP_COLOR,
                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR3,
                                                       bytData, bytData.Length) Then

                                    bytData = ToByteArray(unitLength)
                                    If D2D1Effect_SetValue(pID2D1Effect,
                                                           D2D1_POINTSPECULAR_PROP.D2D1_POINTSPECULAR_PROP_KERNEL_UNIT_LENGTH,
                                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR2,
                                                           bytData, bytData.Length) Then

                                        bytData = ToByteArray(CInt(scaleMode))
                                        If D2D1Effect_SetValue(pID2D1Effect,
                                                               D2D1_POINTSPECULAR_PROP.D2D1_POINTSPECULAR_PROP_SCALE_MODE,
                                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                                               bytData, bytData.Length) Then

                                            D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1PointSpecularEffect)

                                        End If

                                    End If

                                End If

                            End If

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1LookupTable3DEffect(pID2D1Image As IntPtr,
                                                  Optional alphaMode As D2D1_ALPHA_MODE = D2D1_ALPHA_MODE.D2D1_ALPHA_MODE_PREMULTIPLIED) As IntPtr

        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr
        Dim pID2D1LookupTable3D As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            ' Nach dem Beispiel von:
            ' https://docs.microsoft.com/en-us/windows/win32/direct2d/3d-lookup-table-effect
            ' generiert eine Fake-3D-LUT zum testen!

            Dim intExtends As Integer() = New Integer(2) {16, 16, 16}

            Dim intElements As Integer = intExtends(0) * intExtends(1) * intExtends(2) * 4

            Dim sngSteps As Single() = New Single() {1.0F / CSng(intExtends(0) - 1),
                                                     1.0F / CSng(intExtends(1) - 1),
                                                     1.0F / CSng(intExtends(2) - 1)}

            Dim sngData() As Single = New Single() {}
            Array.Resize(sngData, intElements)

            Dim intItem As Integer = 0
            For intR As Integer = 0 To intExtends(2) - 1
                For intG As Integer = 0 To intExtends(1) - 1
                    For intB As Integer = 0 To intExtends(0) - 1
                        sngData(intItem + 0) = intB * sngSteps(0)
                        sngData(intItem + 1) = intG * sngSteps(1)
                        sngData(intItem + 2) = intR * sngSteps(2)
                        sngData(intItem + 3) = 1.0F
                        intItem += 4
                    Next
                Next
            Next

            Dim intStrides As Integer() = New Integer(1) {}
            intStrides(0) = IntMult(Marshal.SizeOf(Of Single) * 4, intExtends(0))
            intStrides(1) = IntMult(intStrides(0), intExtends(1))

            bytData = ToByteArray(sngData)
            If D2D1DeviceContext_CreateLookupTable3D(m_pID2D1DeviceContext2,
                                                     D2D1_BUFFER_PRECISION.D2D1_BUFFER_PRECISION_32BPC_FLOAT,
                                                     intExtends, bytData, bytData.Length, intStrides,
                                                     pID2D1LookupTable3D) Then

                If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                                  New Guid(CLSID_D2D1LookupTable3D),
                                                  pID2D1Effect) Then

                    D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                    bytData = ToByteArray(pID2D1LookupTable3D)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_LOOKUPTABLE3D_PROP.D2D1_LOOKUPTABLE3D_PROP_LUT,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_IUNKNOWN,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(CInt(alphaMode))
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_LOOKUPTABLE3D_PROP.D2D1_LOOKUPTABLE3D_PROP_ALPHA_MODE,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                               bytData, bytData.Length) Then

                            D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1LookupTable3DEffect)


                        End If

                    End If

                    Marshal.Release(pID2D1Effect)

                End If

                Marshal.Release(pID2D1LookupTable3D)

            End If

        End If

    End Function

    Public Function CreateD2D1OpacityMetadataEffect(pID2D1Image As IntPtr,
                                                    opaqueRect As D2D1_VECTOR_4F) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1OpacityMetadata),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(opaqueRect)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_OPACITYMETADATA_PROP.D2D1_OPACITYMETADATA_PROP_INPUT_OPAQUE_RECT,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR4,
                                       bytData, bytData.Length) Then

                    D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1OpacityMetadataEffect)


                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1LinearTransferEffect(pID2D1Image As IntPtr,
                                                   Optional redYIntercept As Single = 0F,
                                                   Optional greenYIntercept As Single = 0F,
                                                   Optional blueYIntercept As Single = 0F,
                                                   Optional alphaYIntercept As Single = 0F,
                                                   Optional redSlope As Single = 1.0F,
                                                   Optional greenSlope As Single = 1.0F,
                                                   Optional blueSlope As Single = 1.0F,
                                                   Optional alphaSlope As Single = 1.0F,
                                                   Optional redDisable As Boolean = False,
                                                   Optional greenDisable As Boolean = False,
                                                   Optional blueDisable As Boolean = False,
                                                   Optional alphaDisable As Boolean = False,
                                                   Optional clampOutput As Boolean = False) As IntPtr

        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1LinearTransfer),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(redYIntercept)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_LINEARTRANSFER_PROP.D2D1_LINEARTRANSFER_PROP_RED_Y_INTERCEPT,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(redSlope)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_LINEARTRANSFER_PROP.D2D1_LINEARTRANSFER_PROP_RED_SLOPE,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(redDisable)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_LINEARTRANSFER_PROP.D2D1_LINEARTRANSFER_PROP_RED_DISABLE,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                               bytData, bytData.Length) Then

                            bytData = ToByteArray(greenYIntercept)
                            If D2D1Effect_SetValue(pID2D1Effect,
                                                   D2D1_LINEARTRANSFER_PROP.D2D1_LINEARTRANSFER_PROP_GREEN_Y_INTERCEPT,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                   bytData, bytData.Length) Then

                                bytData = ToByteArray(greenSlope)
                                If D2D1Effect_SetValue(pID2D1Effect,
                                                       D2D1_LINEARTRANSFER_PROP.D2D1_LINEARTRANSFER_PROP_GREEN_SLOPE,
                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                       bytData, bytData.Length) Then

                                    bytData = ToByteArray(greenDisable)
                                    If D2D1Effect_SetValue(pID2D1Effect,
                                                           D2D1_LINEARTRANSFER_PROP.D2D1_LINEARTRANSFER_PROP_GREEN_DISABLE,
                                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                           bytData, bytData.Length) Then

                                        bytData = ToByteArray(blueYIntercept)
                                        If D2D1Effect_SetValue(pID2D1Effect,
                                                               D2D1_LINEARTRANSFER_PROP.D2D1_LINEARTRANSFER_PROP_BLUE_Y_INTERCEPT,
                                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                               bytData, bytData.Length) Then

                                            bytData = ToByteArray(blueSlope)
                                            If D2D1Effect_SetValue(pID2D1Effect,
                                                                   D2D1_LINEARTRANSFER_PROP.D2D1_LINEARTRANSFER_PROP_BLUE_SLOPE,
                                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                                   bytData, bytData.Length) Then

                                                bytData = ToByteArray(blueDisable)
                                                If D2D1Effect_SetValue(pID2D1Effect,
                                                                       D2D1_LINEARTRANSFER_PROP.D2D1_LINEARTRANSFER_PROP_BLUE_DISABLE,
                                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                                       bytData, bytData.Length) Then

                                                    bytData = ToByteArray(alphaYIntercept)
                                                    If D2D1Effect_SetValue(pID2D1Effect,
                                                                           D2D1_LINEARTRANSFER_PROP.D2D1_LINEARTRANSFER_PROP_ALPHA_Y_INTERCEPT,
                                                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                                           bytData, bytData.Length) Then

                                                        bytData = ToByteArray(alphaSlope)
                                                        If D2D1Effect_SetValue(pID2D1Effect,
                                                                               D2D1_LINEARTRANSFER_PROP.D2D1_LINEARTRANSFER_PROP_ALPHA_SLOPE,
                                                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                                               bytData, bytData.Length) Then

                                                            bytData = ToByteArray(alphaDisable)
                                                            If D2D1Effect_SetValue(pID2D1Effect,
                                                                                   D2D1_LINEARTRANSFER_PROP.D2D1_LINEARTRANSFER_PROP_ALPHA_DISABLE,
                                                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                                                   bytData, bytData.Length) Then

                                                                bytData = ToByteArray(clampOutput)
                                                                If D2D1Effect_SetValue(pID2D1Effect,
                                                                                       D2D1_LINEARTRANSFER_PROP.D2D1_LINEARTRANSFER_PROP_CLAMP_OUTPUT,
                                                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                                                       bytData, bytData.Length) Then

                                                                    D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1LinearTransferEffect)

                                                                End If

                                                            End If

                                                        End If

                                                    End If

                                                End If

                                            End If

                                        End If

                                    End If

                                End If

                            End If

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1ConvolveMatrixEffect(pID2D1Image As IntPtr,
                                                   Optional kernelMatrix As Single() = Nothing,
                                                   Optional kernelOffset As D2D1_VECTOR_2F? = Nothing,
                                                   Optional kernelUnitLength As D2D1_VECTOR_2F? = Nothing,
                                                   Optional divisor As Single = 1.0F,
                                                   Optional bias As Single = 0F,
                                                   Optional kernelSizeX As UInteger = 3UI,
                                                   Optional kernelSizeY As UInteger = 3UI,
                                                   Optional preserveAlpha As Boolean = False,
                                                   Optional clampOutput As Boolean = False,
                                                   Optional ByVal scaleMode As D2D1_CONVOLVEMATRIX_SCALE_MODE = D2D1_CONVOLVEMATRIX_SCALE_MODE.D2D1_CONVOLVEMATRIX_SCALE_MODE_LINEAR,
                                                   Optional ByVal borderMode As D2D1_BORDER_MODE = D2D1_BORDER_MODE.D2D1_BORDER_MODE_SOFT) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            Dim unitLength As D2D1_VECTOR_2F = If(kernelUnitLength.HasValue, CType(kernelUnitLength, D2D1_VECTOR_2F), New D2D1_VECTOR_2F(1, 1))
            Dim offset As D2D1_VECTOR_2F = If(kernelOffset.HasValue, CType(kernelOffset, D2D1_VECTOR_2F), New D2D1_VECTOR_2F(0, 0))

            If kernelSizeX < 3UI Then kernelSizeX = 3UI
            If kernelSizeY < 3UI Then kernelSizeX = 3UI

            If kernelMatrix Is Nothing Then kernelMatrix = New Single() {0F, 0F, 0F, 0F, 1.0F, 0F, 0F, 0F, 0F}

            If kernelMatrix.Length < 9 Then
                Array.Resize(kernelMatrix, 9)
            End If

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1ConvolveMatrix),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(unitLength)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_CONVOLVEMATRIX_PROP.D2D1_CONVOLVEMATRIX_PROP_KERNEL_UNIT_LENGTH,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR2,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(CInt(scaleMode))
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_CONVOLVEMATRIX_PROP.D2D1_CONVOLVEMATRIX_PROP_SCALE_MODE,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(kernelSizeX)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_CONVOLVEMATRIX_PROP.D2D1_CONVOLVEMATRIX_PROP_KERNEL_SIZE_X,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_UINT32,
                                               bytData, bytData.Length) Then

                            bytData = ToByteArray(kernelSizeY)
                            If D2D1Effect_SetValue(pID2D1Effect,
                                                   D2D1_CONVOLVEMATRIX_PROP.D2D1_CONVOLVEMATRIX_PROP_KERNEL_SIZE_Y,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_UINT32,
                                                   bytData, bytData.Length) Then

                                bytData = ToByteArray(kernelMatrix)
                                If D2D1Effect_SetValue(pID2D1Effect,
                                                       D2D1_CONVOLVEMATRIX_PROP.D2D1_CONVOLVEMATRIX_PROP_KERNEL_MATRIX,
                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BLOB,
                                                       bytData, bytData.Length) Then

                                    bytData = ToByteArray(divisor)
                                    If D2D1Effect_SetValue(pID2D1Effect,
                                                           D2D1_CONVOLVEMATRIX_PROP.D2D1_CONVOLVEMATRIX_PROP_DIVISOR,
                                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                            bytData, bytData.Length) Then

                                        bytData = ToByteArray(bias)
                                        If D2D1Effect_SetValue(pID2D1Effect,
                                                               D2D1_CONVOLVEMATRIX_PROP.D2D1_CONVOLVEMATRIX_PROP_BIAS,
                                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                               bytData, bytData.Length) Then

                                            bytData = ToByteArray(offset)
                                            If D2D1Effect_SetValue(pID2D1Effect,
                                                                   D2D1_CONVOLVEMATRIX_PROP.D2D1_CONVOLVEMATRIX_PROP_KERNEL_OFFSET,
                                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR2,
                                                                   bytData, bytData.Length) Then

                                                bytData = ToByteArray(preserveAlpha)
                                                If D2D1Effect_SetValue(pID2D1Effect,
                                                                       D2D1_CONVOLVEMATRIX_PROP.D2D1_CONVOLVEMATRIX_PROP_PRESERVE_ALPHA,
                                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                                       bytData, bytData.Length) Then

                                                    bytData = ToByteArray(CInt(borderMode))
                                                    If D2D1Effect_SetValue(pID2D1Effect,
                                                                       D2D1_CONVOLVEMATRIX_PROP.D2D1_CONVOLVEMATRIX_PROP_BORDER_MODE,
                                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                                                       bytData, bytData.Length) Then

                                                        bytData = ToByteArray(clampOutput)
                                                        If D2D1Effect_SetValue(pID2D1Effect,
                                                                       D2D1_CONVOLVEMATRIX_PROP.D2D1_CONVOLVEMATRIX_PROP_CLAMP_OUTPUT,
                                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                                       bytData, bytData.Length) Then

                                                            D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1ConvolveMatrixEffect)


                                                        End If

                                                    End If

                                                End If

                                            End If

                                        End If

                                    End If

                                End If

                            End If

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1DistantDiffuseEffect(pID2D1Image As IntPtr,
                                                   Optional diffuseColor As D2D1_VECTOR_3F? = Nothing,
                                                   Optional kernelUnitLength As D2D1_VECTOR_2F? = Nothing,
                                                   Optional azimuth As Single = 0F,
                                                   Optional elevation As Single = 0F,
                                                   Optional diffuseConstant As Single = 1.0F,
                                                   Optional surfaceScale As Single = 1.0F,
                                                   Optional scaleMode As D2D1_DISTANTDIFFUSE_SCALE_MODE = D2D1_DISTANTDIFFUSE_SCALE_MODE.D2D1_DISTANTDIFFUSE_SCALE_MODE_LINEAR) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            Dim color As D2D1_VECTOR_3F = If(diffuseColor.HasValue, CType(diffuseColor, D2D1_VECTOR_3F), New D2D1_VECTOR_3F(1, 1, 1))
            Dim unitLength As D2D1_VECTOR_2F = If(kernelUnitLength.HasValue, CType(kernelUnitLength, D2D1_VECTOR_2F), New D2D1_VECTOR_2F(1, 1))

            color.x = Clamp(color.x, 0F, 1.0F)
            color.y = Clamp(color.y, 0F, 1.0F)
            color.z = Clamp(color.z, 0F, 1.0F)

            azimuth = Clamp(azimuth, 0F, 360.0F)
            elevation = Clamp(elevation, 0F, 360.0F)
            diffuseConstant = Clamp(diffuseConstant, 0F, 10000.0F)
            surfaceScale = Clamp(surfaceScale, 0F, 10000.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1DistantDiffuse),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(azimuth)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_DISTANTDIFFUSE_PROP.D2D1_DISTANTDIFFUSE_PROP_AZIMUTH,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(elevation)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_DISTANTDIFFUSE_PROP.D2D1_DISTANTDIFFUSE_PROP_ELEVATION,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(diffuseConstant)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_DISTANTDIFFUSE_PROP.D2D1_DISTANTDIFFUSE_PROP_DIFFUSE_CONSTANT,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                               bytData, bytData.Length) Then

                            bytData = ToByteArray(surfaceScale)
                            If D2D1Effect_SetValue(pID2D1Effect,
                                                   D2D1_DISTANTDIFFUSE_PROP.D2D1_DISTANTDIFFUSE_PROP_SURFACE_SCALE,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                   bytData, bytData.Length) Then

                                bytData = ToByteArray(color)
                                If D2D1Effect_SetValue(pID2D1Effect,
                                                       D2D1_DISTANTDIFFUSE_PROP.D2D1_DISTANTDIFFUSE_PROP_COLOR,
                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR3,
                                                       bytData, bytData.Length) Then

                                    bytData = ToByteArray(unitLength)
                                    If D2D1Effect_SetValue(pID2D1Effect,
                                                           D2D1_DISTANTDIFFUSE_PROP.D2D1_DISTANTDIFFUSE_PROP_KERNEL_UNIT_LENGTH,
                                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR2,
                                                           bytData, bytData.Length) Then

                                        bytData = ToByteArray(CInt(scaleMode))
                                        If D2D1Effect_SetValue(pID2D1Effect,
                                                               D2D1_DISTANTDIFFUSE_PROP.D2D1_DISTANTDIFFUSE_PROP_SCALE_MODE,
                                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                                               bytData, bytData.Length) Then

                                            D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1DistantDiffuseEffect)


                                        End If


                                    End If

                                End If

                            End If

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1DpiCompensationEffect(pID2D1Image As IntPtr,
                                                    Optional inputDPI As D2D1_VECTOR_2F? = Nothing,
                                                    Optional interpolationMode As D2D1_DPICOMPENSATION_INTERPOLATION_MODE = D2D1_DPICOMPENSATION_INTERPOLATION_MODE.D2D1_DPICOMPENSATION_INTERPOLATION_MODE_LINEAR,
                                                    Optional borderMode As D2D1_BORDER_MODE = D2D1_BORDER_MODE.D2D1_BORDER_MODE_SOFT) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            Dim dpi As D2D1_VECTOR_2F = If(inputDPI.HasValue, CType(inputDPI, D2D1_VECTOR_2F), New D2D1_VECTOR_2F(96, 86))

            If dpi.x < 1.0F Then dpi.x = 1.0F
            If dpi.y < 1.0F Then dpi.y = 1.0F

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1DpiCompensation),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(dpi)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_DPICOMPENSATION_PROP.D2D1_DPICOMPENSATION_PROP_INPUT_DPI,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR2,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(CInt(interpolationMode))
                    If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_DPICOMPENSATION_PROP.D2D1_DPICOMPENSATION_PROP_INTERPOLATION_MODE,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                               bytData, bytData.Length) Then

                        bytData = ToByteArray(CInt(borderMode))
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_DPICOMPENSATION_PROP.D2D1_DPICOMPENSATION_PROP_BORDER_MODE,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                               bytData, bytData.Length) Then

                            D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1DpiCompensationEffect)

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1DirectionalBlurEffect(pID2D1Image As IntPtr,
                                                    Optional deviation As Single = 3.0F,
                                                    Optional angle As Single = 0F,
                                                    Optional borderMode As D2D1_BORDER_MODE = D2D1_BORDER_MODE.D2D1_BORDER_MODE_SOFT,
                                                    Optional optimization As D2D1_DIRECTIONALBLUR_OPTIMIZATION = D2D1_DIRECTIONALBLUR_OPTIMIZATION.D2D1_DIRECTIONALBLUR_OPTIMIZATION_BALANCED) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            angle = Clamp(angle, 0F, 360.0F)
            deviation = Clamp(deviation, 0F, 100.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1DirectionalBlur),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(deviation)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_DIRECTIONALBLUR_PROP.D2D1_DIRECTIONALBLUR_PROP_STANDARD_DEVIATION,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(angle)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_DIRECTIONALBLUR_PROP.D2D1_DIRECTIONALBLUR_PROP_ANGLE,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(CInt(optimization))
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_DIRECTIONALBLUR_PROP.D2D1_DIRECTIONALBLUR_PROP_OPTIMIZATION,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                               bytData, bytData.Length) Then

                            bytData = ToByteArray(CInt(borderMode))
                            If D2D1Effect_SetValue(pID2D1Effect,
                                                   D2D1_DIRECTIONALBLUR_PROP.D2D1_DIRECTIONALBLUR_PROP_BORDER_MODE,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                                   bytData, bytData.Length) Then

                                D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1DirectionalBlurEffect)

                            End If

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1TemperatureTintEffect(pID2D1Image As IntPtr,
                                                    Optional tint As Single = 0F,
                                                    Optional temperature As Single = 0F) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            tint = Clamp(tint, -1.0F, 1.0F)
            temperature = Clamp(temperature, -1.0F, 1.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1TemperatureTint),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(temperature)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_TEMPERATUREANDTINT_PROP.D2D1_TEMPERATUREANDTINT_PROP_TEMPERATURE,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(tint)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_TEMPERATUREANDTINT_PROP.D2D1_TEMPERATUREANDTINT_PROP_TINT,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                           bytData, bytData.Length) Then

                        D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1TemperatureTintEffect)

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1DisplacementMapEffect(pID2D1Image0 As IntPtr,
                                                    pID2D1Image1 As IntPtr,
                                                    Optional scale As Single = 0F,
                                                    Optional channelSelectX As D2D1_CHANNEL_SELECTOR = D2D1_CHANNEL_SELECTOR.D2D1_CHANNEL_SELECTOR_A,
                                                    Optional channelSelectY As D2D1_CHANNEL_SELECTOR = D2D1_CHANNEL_SELECTOR.D2D1_CHANNEL_SELECTOR_A) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image0 <> IntPtr.Zero AndAlso pID2D1Image1 <> IntPtr.Zero Then

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1DisplacementMap),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image0)
                D2D1Effect_SetInput(pID2D1Effect, 1, pID2D1Image1)

                bytData = ToByteArray(scale)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_DISPLACEMENTMAP_PROP.D2D1_DISPLACEMENTMAP_PROP_SCALE,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(CInt(channelSelectX))
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_DISPLACEMENTMAP_PROP.D2D1_DISPLACEMENTMAP_PROP_X_CHANNEL_SELECT,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(CInt(channelSelectY))
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_DISPLACEMENTMAP_PROP.D2D1_DISPLACEMENTMAP_PROP_Y_CHANNEL_SELECT,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                               bytData, bytData.Length) Then

                            D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1DisplacementMapEffect)

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1DistantSpecularEffect(pID2D1Image As IntPtr,
                                                    Optional specularColor As D2D1_VECTOR_3F? = Nothing,
                                                    Optional kernelUnitLength As D2D1_VECTOR_2F? = Nothing,
                                                    Optional azimuth As Single = 0F,
                                                    Optional elevation As Single = 0F,
                                                    Optional specularExponent As Single = 1.0F,
                                                    Optional specularConstant As Single = 1.0F,
                                                    Optional surfaceScale As Single = 1.0F,
                                                    Optional scaleMode As D2D1_DISTANTSPECULAR_SCALE_MODE = D2D1_DISTANTSPECULAR_SCALE_MODE.D2D1_DISTANTSPECULAR_SCALE_MODE_LINEAR) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            Dim color As D2D1_VECTOR_3F = If(specularColor.HasValue, CType(specularColor, D2D1_VECTOR_3F), New D2D1_VECTOR_3F(1, 1, 1))
            Dim unitLength As D2D1_VECTOR_2F = If(kernelUnitLength.HasValue, CType(kernelUnitLength, D2D1_VECTOR_2F), New D2D1_VECTOR_2F(1, 1))

            color.x = Clamp(color.x, 0F, 1.0F)
            color.y = Clamp(color.y, 0F, 1.0F)
            color.z = Clamp(color.z, 0F, 1.0F)

            azimuth = Clamp(azimuth, 0F, 360.0F)
            elevation = Clamp(elevation, 0F, 360.0F)
            specularExponent = Clamp(specularExponent, 1.0F, 128.0F)
            specularConstant = Clamp(specularConstant, 1.0F, 10000.0F)
            surfaceScale = Clamp(surfaceScale, 0F, 10000.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1DistantSpecular),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(azimuth)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_DISTANTSPECULAR_PROP.D2D1_DISTANTSPECULAR_PROP_AZIMUTH,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(elevation)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_DISTANTSPECULAR_PROP.D2D1_DISTANTSPECULAR_PROP_ELEVATION,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(specularExponent)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_DISTANTSPECULAR_PROP.D2D1_DISTANTSPECULAR_PROP_SPECULAR_EXPONENT,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                               bytData, bytData.Length) Then

                            bytData = ToByteArray(specularConstant)
                            If D2D1Effect_SetValue(pID2D1Effect,
                                                   D2D1_DISTANTSPECULAR_PROP.D2D1_DISTANTSPECULAR_PROP_SPECULAR_CONSTANT,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                   bytData, bytData.Length) Then

                                bytData = ToByteArray(surfaceScale)
                                If D2D1Effect_SetValue(pID2D1Effect,
                                                       D2D1_DISTANTSPECULAR_PROP.D2D1_DISTANTSPECULAR_PROP_SURFACE_SCALE,
                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                       bytData, bytData.Length) Then

                                    bytData = ToByteArray(color)
                                    If D2D1Effect_SetValue(pID2D1Effect,
                                                           D2D1_DISTANTSPECULAR_PROP.D2D1_DISTANTSPECULAR_PROP_COLOR,
                                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR3,
                                                           bytData, bytData.Length) Then

                                        bytData = ToByteArray(unitLength)
                                        If D2D1Effect_SetValue(pID2D1Effect,
                                                               D2D1_DISTANTSPECULAR_PROP.D2D1_DISTANTSPECULAR_PROP_KERNEL_UNIT_LENGTH,
                                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR2,
                                                               bytData, bytData.Length) Then

                                            bytData = ToByteArray(CInt(scaleMode))
                                            If D2D1Effect_SetValue(pID2D1Effect,
                                                                   D2D1_DISTANTSPECULAR_PROP.D2D1_DISTANTSPECULAR_PROP_SCALE_MODE,
                                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                                                   bytData, bytData.Length) Then

                                                D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1DistantSpecularEffect)

                                            End If

                                        End If

                                    End If

                                End If

                            End If

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1ColorManagementEffect(pID2D1Image As IntPtr,
                                                    Optional sourceColorSpace As D2D1_COLOR_SPACE = D2D1_COLOR_SPACE.D2D1_COLOR_SPACE_SRGB,
                                                    Optional sourceRenderingIntent As D2D1_COLORMANAGEMENT_RENDERING_INTENT = D2D1_COLORMANAGEMENT_RENDERING_INTENT.D2D1_COLORMANAGEMENT_RENDERING_INTENT_PERCEPTUAL,
                                                    Optional destinationColorSpace As D2D1_COLOR_SPACE = D2D1_COLOR_SPACE.D2D1_COLOR_SPACE_SRGB,
                                                    Optional destinationRenderingIntent As D2D1_COLORMANAGEMENT_RENDERING_INTENT = D2D1_COLORMANAGEMENT_RENDERING_INTENT.D2D1_COLORMANAGEMENT_RENDERING_INTENT_PERCEPTUAL,
                                                    Optional alphaMode As D2D1_COLORMANAGEMENT_ALPHA_MODE = D2D1_COLORMANAGEMENT_ALPHA_MODE.D2D1_COLORMANAGEMENT_ALPHA_MODE_PREMULTIPLIED,
                                                    Optional qualityMode As D2D1_COLORMANAGEMENT_QUALITY = D2D1_COLORMANAGEMENT_QUALITY.D2D1_COLORMANAGEMENT_QUALITY_NORMAL) As IntPtr

        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr
        Dim pID2D1ColorContextSrc As IntPtr
        Dim pID2D1ColorContextDst As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            If D2D1DeviceContext_CreateColorContext(m_pID2D1DeviceContext2, sourceColorSpace,
                                                    IntPtr.Zero, 0, pID2D1ColorContextSrc) Then

                If D2D1DeviceContext_CreateColorContext(m_pID2D1DeviceContext2, destinationColorSpace,
                                                        IntPtr.Zero, 0, pID2D1ColorContextDst) Then


                    If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                                      New Guid(CLSID_D2D1ColorManagement),
                                                      pID2D1Effect) Then


                        D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                        bytData = ToByteArray(pID2D1ColorContextSrc)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_COLORMANAGEMENT_PROP.D2D1_COLORMANAGEMENT_PROP_SOURCE_COLOR_CONTEXT,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_COLOR_CONTEXT,
                                               bytData, bytData.Length) Then

                            bytData = ToByteArray(CInt(sourceRenderingIntent))
                            If D2D1Effect_SetValue(pID2D1Effect,
                                                   D2D1_COLORMANAGEMENT_PROP.D2D1_COLORMANAGEMENT_PROP_SOURCE_RENDERING_INTENT,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                                   bytData, bytData.Length) Then

                                bytData = ToByteArray(pID2D1ColorContextDst)
                                If D2D1Effect_SetValue(pID2D1Effect,
                                                       D2D1_COLORMANAGEMENT_PROP.D2D1_COLORMANAGEMENT_PROP_DESTINATION_COLOR_CONTEXT,
                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_COLOR_CONTEXT,
                                                       bytData, bytData.Length) Then

                                    bytData = ToByteArray(CInt(destinationRenderingIntent))
                                    If D2D1Effect_SetValue(pID2D1Effect,
                                                           D2D1_COLORMANAGEMENT_PROP.D2D1_COLORMANAGEMENT_PROP_DESTINATION_RENDERING_INTENT,
                                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                                           bytData, bytData.Length) Then

                                        bytData = ToByteArray(CInt(alphaMode))
                                        If D2D1Effect_SetValue(pID2D1Effect,
                                                               D2D1_COLORMANAGEMENT_PROP.D2D1_COLORMANAGEMENT_PROP_ALPHA_MODE,
                                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                                               bytData, bytData.Length) Then

                                            bytData = ToByteArray(CInt(qualityMode))
                                            If D2D1Effect_SetValue(pID2D1Effect,
                                                                   D2D1_COLORMANAGEMENT_PROP.D2D1_COLORMANAGEMENT_PROP_QUALITY,
                                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                                                   bytData, bytData.Length) Then

                                                D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1ColorManagementEffect)

                                            End If

                                        End If

                                    End If

                                End If

                            End If

                        End If

                        Marshal.Release(pID2D1Effect)

                    End If

                    Marshal.Release(pID2D1ColorContextDst)

                End If

                Marshal.Release(pID2D1ColorContextSrc)

            End If

        End If

    End Function

    Public Function CreateD2D1DiscreteTransferEffect(pID2D1Image As IntPtr,
                                                     Optional redTable As Single() = Nothing,
                                                     Optional greenTable As Single() = Nothing,
                                                     Optional blueTable As Single() = Nothing,
                                                     Optional alphaTable As Single() = Nothing,
                                                     Optional redDisable As Boolean = False,
                                                     Optional greenDisable As Boolean = False,
                                                     Optional blueDisable As Boolean = False,
                                                     Optional alphaDisable As Boolean = False,
                                                     Optional clampOutput As Boolean = False) As IntPtr
        Dim bytData As Byte()
        Dim intCount As Integer
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            If redTable Is Nothing Then redTable = New Single() {0F, 1.0F}
            If greenTable Is Nothing Then greenTable = New Single() {0F, 1.0F}
            If blueTable Is Nothing Then blueTable = New Single() {0F, 1.0F}
            If alphaTable Is Nothing Then alphaTable = New Single() {0F, 1.0F}

            intCount = redTable.Length
            For intItem = 0 To intCount - 1
                redTable(intItem) = Clamp(redTable(intItem), 0F, 1.0F)
            Next

            intCount = greenTable.Length
            For intItem = 0 To intCount - 1
                greenTable(intItem) = Clamp(greenTable(intItem), 0F, 1.0F)
            Next

            intCount = blueTable.Length
            For intItem = 0 To intCount - 1
                blueTable(intItem) = Clamp(blueTable(intItem), 0F, 1.0F)
            Next

            intCount = alphaTable.Length
            For intItem = 0 To intCount - 1
                alphaTable(intItem) = Clamp(alphaTable(intItem), 0F, 1.0F)
            Next

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1DiscreteTransfer),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(redTable)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_DISCRETETRANSFER_PROP.D2D1_DISCRETETRANSFER_PROP_RED_TABLE,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BLOB,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(redDisable)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_DISCRETETRANSFER_PROP.D2D1_DISCRETETRANSFER_PROP_RED_DISABLE,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(greenTable)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_DISCRETETRANSFER_PROP.D2D1_DISCRETETRANSFER_PROP_GREEN_TABLE,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BLOB,
                                               bytData, bytData.Length) Then

                            bytData = ToByteArray(greenDisable)
                            If D2D1Effect_SetValue(pID2D1Effect,
                                                   D2D1_DISCRETETRANSFER_PROP.D2D1_DISCRETETRANSFER_PROP_GREEN_DISABLE,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                   bytData, bytData.Length) Then

                                bytData = ToByteArray(blueTable)
                                If D2D1Effect_SetValue(pID2D1Effect,
                                                       D2D1_DISCRETETRANSFER_PROP.D2D1_DISCRETETRANSFER_PROP_BLUE_TABLE,
                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BLOB,
                                                       bytData, bytData.Length) Then

                                    bytData = ToByteArray(blueDisable)
                                    If D2D1Effect_SetValue(pID2D1Effect,
                                                           D2D1_DISCRETETRANSFER_PROP.D2D1_DISCRETETRANSFER_PROP_BLUE_DISABLE,
                                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                            bytData, bytData.Length) Then

                                        bytData = ToByteArray(alphaTable)
                                        If D2D1Effect_SetValue(pID2D1Effect,
                                                               D2D1_DISCRETETRANSFER_PROP.D2D1_DISCRETETRANSFER_PROP_ALPHA_TABLE,
                                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BLOB,
                                                               bytData, bytData.Length) Then

                                            bytData = ToByteArray(alphaDisable)
                                            If D2D1Effect_SetValue(pID2D1Effect,
                                                                   D2D1_DISCRETETRANSFER_PROP.D2D1_DISCRETETRANSFER_PROP_ALPHA_DISABLE,
                                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                                   bytData, bytData.Length) Then

                                                bytData = ToByteArray(clampOutput)
                                                If D2D1Effect_SetValue(pID2D1Effect,
                                                                       D2D1_DISCRETETRANSFER_PROP.D2D1_DISCRETETRANSFER_PROP_CLAMP_OUTPUT,
                                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                                                       bytData, bytData.Length) Then

                                                    D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1DiscreteTransferEffect)


                                                End If

                                            End If

                                        End If

                                    End If

                                End If

                            End If

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1LuminanceToAlphaEffect(pID2D1Image As IntPtr) As IntPtr

        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1LuminanceToAlpha),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1LuminanceToAlphaEffect)

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1HighlightsShadowsEffect(pID2D1Image As IntPtr,
                                                      Optional highlights As Single = 0F,
                                                      Optional [shadows] As Single = 0F,
                                                      Optional clarity As Single = 0F,
                                                      Optional maskBlurRadius As Single = 1.25F,
                                                      Optional inputGamma As D2D1_HIGHLIGHTSANDSHADOWS_INPUT_GAMMA = D2D1_HIGHLIGHTSANDSHADOWS_INPUT_GAMMA.D2D1_HIGHLIGHTSANDSHADOWS_INPUT_GAMMA_LINEAR) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            highlights = Clamp(highlights, -1.0F, 1.0F)
            [shadows] = Clamp([shadows], -1.0F, 1.0F)
            clarity = Clamp(clarity, -1.0F, 1.0F)
            maskBlurRadius = Clamp(maskBlurRadius, 0F, 10.0F)

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1HighlightsShadows),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(highlights)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_HIGHLIGHTSANDSHADOWS_PROP.D2D1_HIGHLIGHTSANDSHADOWS_PROP_HIGHLIGHTS,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray([shadows])
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_HIGHLIGHTSANDSHADOWS_PROP.D2D1_HIGHLIGHTSANDSHADOWS_PROP_SHADOWS,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(clarity)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_HIGHLIGHTSANDSHADOWS_PROP.D2D1_HIGHLIGHTSANDSHADOWS_PROP_CLARITY,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                               bytData, bytData.Length) Then

                            bytData = ToByteArray(CInt(inputGamma))
                            If D2D1Effect_SetValue(pID2D1Effect,
                                                   D2D1_HIGHLIGHTSANDSHADOWS_PROP.D2D1_HIGHLIGHTSANDSHADOWS_PROP_INPUT_GAMMA,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                                   bytData, bytData.Length) Then

                                bytData = ToByteArray(maskBlurRadius)
                                If D2D1Effect_SetValue(pID2D1Effect,
                                                       D2D1_HIGHLIGHTSANDSHADOWS_PROP.D2D1_HIGHLIGHTSANDSHADOWS_PROP_MASK_BLUR_RADIUS,
                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                       bytData, bytData.Length) Then

                                    D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1HighlightsShadowsEffect)

                                End If

                            End If

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D12DAffineTransformEffect(pID2D1Image As IntPtr,
                                                      Optional tranformMatrix As D2D1_MATRIX_3X2_F? = Nothing,
                                                      Optional sharpness As Single = 1.0F,
                                                      Optional interpolationMode As D2D1_2DAFFINETRANSFORM_INTERPOLATION_MODE = D2D1_2DAFFINETRANSFORM_INTERPOLATION_MODE.D2D1_2DAFFINETRANSFORM_INTERPOLATION_MODE_LINEAR,
                                                      Optional borderMode As D2D1_BORDER_MODE = D2D1_BORDER_MODE.D2D1_BORDER_MODE_SOFT) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            sharpness = Clamp(sharpness, 0F, 1.0F)
            Dim matrix As D2D1_MATRIX_3X2_F = If(tranformMatrix.HasValue, CType(tranformMatrix, D2D1_MATRIX_3X2_F), New D2D1_MATRIX_3X2_F(1, 0, 0, 1, 0, 0))

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D12DAffineTransform),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(CInt(interpolationMode))
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_2DAFFINETRANSFORM_PROP.D2D1_2DAFFINETRANSFORM_PROP_INTERPOLATION_MODE,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(CInt(borderMode))
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_2DAFFINETRANSFORM_PROP.D2D1_2DAFFINETRANSFORM_PROP_BORDER_MODE,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(matrix)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_2DAFFINETRANSFORM_PROP.D2D1_2DAFFINETRANSFORM_PROP_TRANSFORM_MATRIX,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_MATRIX_3X2,
                                               bytData, bytData.Length) Then

                            bytData = ToByteArray(sharpness)
                            If D2D1Effect_SetValue(pID2D1Effect,
                                                   D2D1_2DAFFINETRANSFORM_PROP.D2D1_2DAFFINETRANSFORM_PROP_SHARPNESS,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                                   bytData, bytData.Length) Then

                                D2D1Effect_GetOutput(pID2D1Effect, CreateD2D12DAffineTransformEffect)

                            End If

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1ArithmeticCompositeEffect(pID2D1Image0 As IntPtr,
                                                        pID2D1Image1 As IntPtr,
                                                        Optional coefficients As D2D1_VECTOR_4F? = Nothing,
                                                        Optional clampOutput As Boolean = False) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image0 <> IntPtr.Zero AndAlso pID2D1Image1 <> IntPtr.Zero Then

            Dim coeff As D2D1_VECTOR_4F = If(coefficients.HasValue, CType(coefficients, D2D1_VECTOR_4F), New D2D1_VECTOR_4F(1, 0, 0, 0))

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1ArithmeticComposite),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image0)
                D2D1Effect_SetInput(pID2D1Effect, 1, pID2D1Image1)

                bytData = ToByteArray(coeff)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_ARITHMETICCOMPOSITE_PROP.D2D1_ARITHMETICCOMPOSITE_PROP_COEFFICIENTS,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR4,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(clampOutput)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_ARITHMETICCOMPOSITE_PROP.D2D1_ARITHMETICCOMPOSITE_PROP_CLAMP_OUTPUT,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_BOOL,
                                           bytData, bytData.Length) Then

                        D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1ArithmeticCompositeEffect)

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D1WhiteLevelAdjustmentEffect(pID2D1Image As IntPtr,
                                                         Optional inputWhiteLevel As Single = 80.0F,
                                                         Optional outputWhiteLevel As Single = 80.0F) As IntPtr
        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            If inputWhiteLevel < 1.0F Then inputWhiteLevel = 1.0F
            If outputWhiteLevel < 1.0F Then outputWhiteLevel = 1.0F

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D1WhiteLevelAdjustment),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(inputWhiteLevel)
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_WHITELEVELADJUSTMENT_PROP.D2D1_WHITELEVELADJUSTMENT_PROP_INPUT_WHITE_LEVEL,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(outputWhiteLevel)
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_WHITELEVELADJUSTMENT_PROP.D2D1_WHITELEVELADJUSTMENT_PROP_OUTPUT_WHITE_LEVEL,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                           bytData, bytData.Length) Then

                        D2D1Effect_GetOutput(pID2D1Effect, CreateD2D1WhiteLevelAdjustmentEffect)

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function

    Public Function CreateD2D13DPerspectiveTransformEffect(pID2D1Image As IntPtr,
                                                           Optional perspectiveOrigin As D2D1_VECTOR_2F? = Nothing,
                                                           Optional localOffset As D2D1_VECTOR_3F? = Nothing,
                                                           Optional globalOffset As D2D1_VECTOR_3F? = Nothing,
                                                           Optional rotationOrigin As D2D1_VECTOR_3F? = Nothing,
                                                           Optional rotation As D2D1_VECTOR_3F? = Nothing,
                                                           Optional depth As Single = 1000.0F,
                                                           Optional interpolationMode As D2D1_3DPERSPECTIVETRANSFORM_INTERPOLATION_MODE = D2D1_3DPERSPECTIVETRANSFORM_INTERPOLATION_MODE.D2D1_3DPERSPECTIVETRANSFORM_INTERPOLATION_MODE_LINEAR,
                                                           Optional borderMode As D2D1_BORDER_MODE = D2D1_BORDER_MODE.D2D1_BORDER_MODE_SOFT) As IntPtr

        Dim bytData As Byte()
        Dim pID2D1Effect As IntPtr

        If pID2D1Image <> IntPtr.Zero Then

            If depth < 1.0F Then depth = 1.0F

            Dim persOrigin As D2D1_VECTOR_2F = If(perspectiveOrigin.HasValue, CType(perspectiveOrigin, D2D1_VECTOR_2F), New D2D1_VECTOR_2F(0, 0))
            Dim locOffset As D2D1_VECTOR_3F = If(localOffset.HasValue, CType(localOffset, D2D1_VECTOR_3F), New D2D1_VECTOR_3F(0, 0, 0))
            Dim gloOffset As D2D1_VECTOR_3F = If(globalOffset.HasValue, CType(globalOffset, D2D1_VECTOR_3F), New D2D1_VECTOR_3F(0, 0, 0))
            Dim rotaOrigin As D2D1_VECTOR_3F = If(rotationOrigin.HasValue, CType(rotationOrigin, D2D1_VECTOR_3F), New D2D1_VECTOR_3F(0, 0, 0))
            Dim rota As D2D1_VECTOR_3F = If(rotation.HasValue, CType(rotation, D2D1_VECTOR_3F), New D2D1_VECTOR_3F(0, 0, 0))

            If D2D1DeviceContext_CreateEffect(m_pID2D1DeviceContext2,
                                              New Guid(CLSID_D2D13DPerspectiveTransform),
                                              pID2D1Effect) Then

                D2D1Effect_SetInput(pID2D1Effect, 0, pID2D1Image)

                bytData = ToByteArray(CInt(interpolationMode))
                If D2D1Effect_SetValue(pID2D1Effect,
                                       D2D1_3DPERSPECTIVETRANSFORM_PROP.D2D1_3DPERSPECTIVETRANSFORM_PROP_INTERPOLATION_MODE,
                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                       bytData, bytData.Length) Then

                    bytData = ToByteArray(CInt(borderMode))
                    If D2D1Effect_SetValue(pID2D1Effect,
                                           D2D1_3DPERSPECTIVETRANSFORM_PROP.D2D1_3DPERSPECTIVETRANSFORM_PROP_BORDER_MODE,
                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_ENUM,
                                           bytData, bytData.Length) Then

                        bytData = ToByteArray(depth)
                        If D2D1Effect_SetValue(pID2D1Effect,
                                               D2D1_3DPERSPECTIVETRANSFORM_PROP.D2D1_3DPERSPECTIVETRANSFORM_PROP_DEPTH,
                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_FLOAT,
                                               bytData, bytData.Length) Then

                            bytData = ToByteArray(persOrigin)
                            If D2D1Effect_SetValue(pID2D1Effect,
                                                   D2D1_3DPERSPECTIVETRANSFORM_PROP.D2D1_3DPERSPECTIVETRANSFORM_PROP_PERSPECTIVE_ORIGIN,
                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR2,
                                                   bytData, bytData.Length) Then

                                bytData = ToByteArray(locOffset)
                                If D2D1Effect_SetValue(pID2D1Effect,
                                                       D2D1_3DPERSPECTIVETRANSFORM_PROP.D2D1_3DPERSPECTIVETRANSFORM_PROP_LOCAL_OFFSET,
                                                       D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR3,
                                                       bytData, bytData.Length) Then

                                    bytData = ToByteArray(gloOffset)
                                    If D2D1Effect_SetValue(pID2D1Effect,
                                                           D2D1_3DPERSPECTIVETRANSFORM_PROP.D2D1_3DPERSPECTIVETRANSFORM_PROP_GLOBAL_OFFSET,
                                                           D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR3,
                                                           bytData, bytData.Length) Then

                                        bytData = ToByteArray(rotaOrigin)
                                        If D2D1Effect_SetValue(pID2D1Effect,
                                                               D2D1_3DPERSPECTIVETRANSFORM_PROP.D2D1_3DPERSPECTIVETRANSFORM_PROP_ROTATION_ORIGIN,
                                                               D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR3,
                                                               bytData, bytData.Length) Then

                                            bytData = ToByteArray(rota)
                                            If D2D1Effect_SetValue(pID2D1Effect,
                                                                   D2D1_3DPERSPECTIVETRANSFORM_PROP.D2D1_3DPERSPECTIVETRANSFORM_PROP_ROTATION,
                                                                   D2D1_PROPERTY_TYPE.D2D1_PROPERTY_TYPE_VECTOR3,
                                                                   bytData, bytData.Length) Then

                                                D2D1Effect_GetOutput(pID2D1Effect, CreateD2D13DPerspectiveTransformEffect)

                                            End If

                                        End If

                                    End If

                                End If

                            End If

                        End If

                    End If

                End If

                Marshal.Release(pID2D1Effect)

            End If

        End If

    End Function
#End Region

#Region "Private Functions"
    Private Function ToByteArray(value As Boolean) As Byte()
        Return BitConverter.GetBytes(Math.Abs(CInt(value)))
    End Function

    Private Function ToByteArray(value As Integer) As Byte()
        Return BitConverter.GetBytes(value)
    End Function

    Private Function ToByteArray(value As UInteger) As Byte()
        Return BitConverter.GetBytes(value)
    End Function

    Private Function ToByteArray(value As Single) As Byte()
        Return BitConverter.GetBytes(value)
    End Function

    Private Function ToByteArray(value As Single()) As Byte()
        Dim arrRet As Byte() = New Byte() {}
        Dim intCount As Integer = value.Length
        Dim intSize As Integer = Marshal.SizeOf(Of Single)
        Array.Resize(arrRet, intCount * intSize)
        For intItem = 0 To intCount - 1
            Array.Copy(BitConverter.GetBytes(value(intItem)), 0, arrRet, intItem * intSize, intSize)
        Next
        Return arrRet
    End Function

    Private Function ToByteArray(value As IntPtr) As Byte()
        Dim data As Byte() = New Byte() {}
        If IntPtr.Size = 4 Then '32bit
            data = BitConverter.GetBytes(value.ToInt32)
        ElseIf IntPtr.Size = 8 Then '64bit
            data = BitConverter.GetBytes(value.ToInt64)
        End If
        Return data
    End Function

    Private Function ToByteArray(Of T As Structure)([structure] As T) As Byte()
        Dim bufferSize = Marshal.SizeOf([structure])
        Dim byteArray = New Byte(bufferSize - 1) {}
        Dim handle As IntPtr = Marshal.AllocHGlobal(bufferSize)
        Marshal.StructureToPtr([structure], handle, False)
        Marshal.Copy(handle, byteArray, 0, bufferSize)
        Marshal.FreeHGlobal(handle)
        Return byteArray
    End Function

    Private Function Clamp(Value As UInteger, Min As UInteger, Max As UInteger) As UInteger
        If Value <= Min Then Return Min
        If Value >= Max Then Return Max
        Return Value
    End Function

    Private Function Clamp(Value As Single, Min As Single, Max As Single) As Single
        If Value <= Min Then Return Min
        If Value >= Max Then Return Max
        Return Value
    End Function

    Private Function IntMult(intMultiplicand As Integer, intMultiplier As Integer) As Integer
        IntMult = intMultiplicand * intMultiplier
    End Function
#End Region

#Region "Dispose"
    Protected Overridable Sub Dispose(disposing As Boolean)

        If Not m_bolDisposedValue Then

            ' ID2D1DeviceContext2 löschen
            If m_pID2D1DeviceContext2 <> IntPtr.Zero Then
                If Marshal.Release(m_pID2D1DeviceContext2) = 0 Then
                    m_pID2D1DeviceContext2 = IntPtr.Zero
                End If
            End If

            ' ID2D1RenderTarget löschen
            If m_pID2D1RenderTarget <> IntPtr.Zero Then
                If Marshal.Release(m_pID2D1RenderTarget) = 0 Then
                    m_pID2D1RenderTarget = IntPtr.Zero
                End If
            End If

            ' IWICBitmap löschen
            If m_pIWICBitmap <> IntPtr.Zero Then
                If Marshal.Release(m_pIWICBitmap) = 0 Then
                    m_pIWICBitmap = IntPtr.Zero
                End If
            End If

            ' ID2D1Factory löschen
            If m_pID2D1Factory <> IntPtr.Zero Then
                If Marshal.Release(m_pID2D1Factory) = 0 Then
                    m_pID2D1Factory = IntPtr.Zero
                End If
            End If

            ' IImagingFactory löschen
            If m_pIImagingFactory <> IntPtr.Zero Then
                If Marshal.Release(m_pIImagingFactory) = 0 Then
                    m_pIImagingFactory = IntPtr.Zero
                End If
            End If

            m_bolDisposedValue = True

        End If

    End Sub
    Public Sub Dispose() Implements IDisposable.Dispose
        Dispose(disposing:=True)
        GC.SuppressFinalize(Me)
    End Sub
#End Region

End Class
